#ifndef MBAA_ISO_INDEX_H
#define MBAA_ISO_INDEX_H

/*
I couldn't figure out the encryption used in the mepRofs.cvm's file index.
And mounting the iso on Linux leads to all zeroes for it anyway.

So what I did was run the game in PCSX2, save state, then manually uncompress
the state(simple gzopen/gzread) and extract the correct file index out of
the game's memory.

Search for "VOL0" in memory, there are several of these, one for each
part of the directory structure.

This could likely be used to crack the encryption, but I don't care enough.

This is not a complete index. I excluded a large number of files that were
not useful for framedisplay. You want them, you get them yourself.

Technically I should remove this from the code and have an external file
index, but this is good enough for now.

 */

struct mbaa_iso_index_entry_t {
	const char *filename;
	unsigned int size;
	unsigned int sector; // * 0x800 for location.
};

static const mbaa_iso_index_entry_t mbaa_iso_index[] = {
	{ "ARCADE_AKAAKIHA.ARX", 0x3cda40,	0x2032c },
	{ "ARCADE_AKAAKIHA.CG2", 0x85678,	0x20ac8 },
	{ "ARCADE_AKAAKIHA.CPF", 0x2f36c,	0x20bd3 },
	{ "ARCADE_AKAAKIHA.DOF", 0xd8a24,	0x20c32 },
	{ "ARCADE_AKAAKIHA.HA6", 0xa1684,	0x20de4 },
	{ "ARCADE_AKAAKIHA.PAT", 0x0,	0x20f27 },
	{ "ARCADE_AKAAKIHA.TXT", 0xda,	0x20f27 },
	{ "ARCADE_AKAAKIHA0.PAL", 0x400,	0x20f28 },
	{ "ARCADE_AKAAKIHA1.PAL", 0x400,	0x20f29 },
	{ "ARCADE_AKAAKIHA10.PAL", 0x400,	0x20f2a },
	{ "ARCADE_AKAAKIHA11.PAL", 0x400,	0x20f2b },
	{ "ARCADE_AKAAKIHA12.PAL", 0x400,	0x20f2c },
	{ "ARCADE_AKAAKIHA13.PAL", 0x400,	0x20f2d },
	{ "ARCADE_AKAAKIHA14.PAL", 0x400,	0x20f2e },
	{ "ARCADE_AKAAKIHA15.PAL", 0x400,	0x20f2f },
	{ "ARCADE_AKAAKIHA16.PAL", 0x400,	0x20f30 },
	{ "ARCADE_AKAAKIHA17.PAL", 0x400,	0x20f31 },
	{ "ARCADE_AKAAKIHA18.PAL", 0x400,	0x20f32 },
	{ "ARCADE_AKAAKIHA19.PAL", 0x400,	0x20f33 },
	{ "ARCADE_AKAAKIHA2.ARX", 0x3b61c0,	0x20f34 },
	{ "ARCADE_AKAAKIHA2.PAL", 0x400,	0x216a1 },
	{ "ARCADE_AKAAKIHA20.PAL", 0x400,	0x216a2 },
	{ "ARCADE_AKAAKIHA21.PAL", 0x400,	0x216a3 },
	{ "ARCADE_AKAAKIHA22.PAL", 0x400,	0x216a4 },
	{ "ARCADE_AKAAKIHA23.PAL", 0x400,	0x216a5 },
	{ "ARCADE_AKAAKIHA24.PAL", 0x400,	0x216a6 },
	{ "ARCADE_AKAAKIHA25.PAL", 0x400,	0x216a7 },
	{ "ARCADE_AKAAKIHA26.PAL", 0x400,	0x216a8 },
	{ "ARCADE_AKAAKIHA27.PAL", 0x400,	0x216a9 },
	{ "ARCADE_AKAAKIHA28.PAL", 0x400,	0x216aa },
	{ "ARCADE_AKAAKIHA29.PAL", 0x400,	0x216ab },
	{ "ARCADE_AKAAKIHA3.PAL", 0x400,	0x216ac },
	{ "ARCADE_AKAAKIHA30.PAL", 0x400,	0x216ad },
	{ "ARCADE_AKAAKIHA31.PAL", 0x400,	0x216ae },
	{ "ARCADE_AKAAKIHA32.PAL", 0x400,	0x216af },
	{ "ARCADE_AKAAKIHA33.PAL", 0x400,	0x216b0 },
	{ "ARCADE_AKAAKIHA34.PAL", 0x400,	0x216b1 },
	{ "ARCADE_AKAAKIHA35.PAL", 0x400,	0x216b2 },
	{ "ARCADE_AKAAKIHA4.PAL", 0x400,	0x216b3 },
	{ "ARCADE_AKAAKIHA5.PAL", 0x400,	0x216b4 },
	{ "ARCADE_AKAAKIHA6.PAL", 0x400,	0x216b5 },
	{ "ARCADE_AKAAKIHA7.PAL", 0x400,	0x216b6 },
	{ "ARCADE_AKAAKIHA8.PAL", 0x400,	0x216b7 },
	{ "ARCADE_AKAAKIHA9.PAL", 0x400,	0x216b8 },
	{ "ARCADE_AKAAKIHAPAL2.ARX", 0x22d40,	0x216b9 },
	{ "ARCADE_AKAAKIHA_0.HA6", 0x2f34,	0x216ff },
	{ "ARCADE_AKAAKIHA_0.TXT", 0xac,	0x21705 },
	{ "ARCADE_AKAAKIHA_0_C.TXT", 0x14ba,	0x21706 },
	{ "ARCADE_AKAAKIHA_1.HA6", 0xb9e0,	0x21709 },
	{ "ARCADE_AKAAKIHA_1.TXT", 0xac,	0x21721 },
	{ "ARCADE_AKAAKIHA_1_C.TXT", 0x1540,	0x21722 },
	{ "ARCADE_AKAAKIHA_2.HA6", 0x13f70,	0x21725 },
	{ "ARCADE_AKAAKIHA_2.TXT", 0xac,	0x2174d },
	{ "ARCADE_AKAAKIHA_2_C.TXT", 0x1396,	0x2174e },
	{ "ARCADE_AKAAKIHA_3.HA6", 0x1100c,	0x21751 },
	{ "ARCADE_AKAAKIHA_9.HA6", 0x13b68,	0x21774 },
	{ "ARCADE_AKAAKIHA_9.TXT", 0xac,	0x2179c },
	{ "ARCADE_AKAAKIHA_9_C.TXT", 0x125a,	0x2179d },
	{ "ARCADE_AKAAKIHA_SELIST.TXT", 0xf04,	0x217a0 },
	{ "ARCADE_AKIHA.ARX", 0x34d6c0,	0x217a2 },
	{ "ARCADE_AKIHA.CG2", 0x74e18,	0x21e3d },
	{ "ARCADE_AKIHA.CPF", 0x2f36c,	0x21f27 },
	{ "ARCADE_AKIHA.DOF", 0xdac40,	0x21f86 },
	{ "ARCADE_AKIHA.HA6", 0x7f390,	0x2213c },
	{ "ARCADE_AKIHA.PAT", 0x0,	0x2223b },
	{ "ARCADE_AKIHA.TXT", 0xc8,	0x2223b },
	{ "ARCADE_AKIHA0.PAL", 0x400,	0x2223c },
	{ "ARCADE_AKIHA1.PAL", 0x400,	0x2223d },
	{ "ARCADE_AKIHA10.PAL", 0x400,	0x2223e },
	{ "ARCADE_AKIHA11.PAL", 0x400,	0x2223f },
	{ "ARCADE_AKIHA12.PAL", 0x400,	0x22240 },
	{ "ARCADE_AKIHA13.PAL", 0x400,	0x22241 },
	{ "ARCADE_AKIHA14.PAL", 0x400,	0x22242 },
	{ "ARCADE_AKIHA15.PAL", 0x400,	0x22243 },
	{ "ARCADE_AKIHA16.PAL", 0x400,	0x22244 },
	{ "ARCADE_AKIHA17.PAL", 0x400,	0x22245 },
	{ "ARCADE_AKIHA18.PAL", 0x400,	0x22246 },
	{ "ARCADE_AKIHA19.PAL", 0x400,	0x22247 },
	{ "ARCADE_AKIHA2.ARX", 0x345f40,	0x22248 },
	{ "ARCADE_AKIHA2.PAL", 0x400,	0x228d4 },
	{ "ARCADE_AKIHA20.PAL", 0x400,	0x228d5 },
	{ "ARCADE_AKIHA21.PAL", 0x400,	0x228d6 },
	{ "ARCADE_AKIHA22.PAL", 0x400,	0x228d7 },
	{ "ARCADE_AKIHA23.PAL", 0x400,	0x228d8 },
	{ "ARCADE_AKIHA24.PAL", 0x400,	0x228d9 },
	{ "ARCADE_AKIHA25.PAL", 0x400,	0x228da },
	{ "ARCADE_AKIHA26.PAL", 0x400,	0x228db },
	{ "ARCADE_AKIHA27.PAL", 0x400,	0x228dc },
	{ "ARCADE_AKIHA28.PAL", 0x400,	0x228dd },
	{ "ARCADE_AKIHA29.PAL", 0x400,	0x228de },
	{ "ARCADE_AKIHA3.PAL", 0x400,	0x228df },
	{ "ARCADE_AKIHA30.PAL", 0x400,	0x228e0 },
	{ "ARCADE_AKIHA31.PAL", 0x400,	0x228e1 },
	{ "ARCADE_AKIHA32.PAL", 0x400,	0x228e2 },
	{ "ARCADE_AKIHA33.PAL", 0x400,	0x228e3 },
	{ "ARCADE_AKIHA34.PAL", 0x400,	0x228e4 },
	{ "ARCADE_AKIHA35.PAL", 0x400,	0x228e5 },
	{ "ARCADE_AKIHA4.PAL", 0x400,	0x228e6 },
	{ "ARCADE_AKIHA5.PAL", 0x400,	0x228e7 },
	{ "ARCADE_AKIHA6.PAL", 0x400,	0x228e8 },
	{ "ARCADE_AKIHA7.PAL", 0x400,	0x228e9 },
	{ "ARCADE_AKIHA8.PAL", 0x400,	0x228ea },
	{ "ARCADE_AKIHA9.PAL", 0x400,	0x228eb },
	{ "ARCADE_AKIHAPAL2.ARX", 0x1d1c0,	0x228ec },
	{ "ARCADE_AKIHA_0.HA6", 0x6e04,	0x22927 },
	{ "ARCADE_AKIHA_0.TXT", 0xa0,	0x22935 },
	{ "ARCADE_AKIHA_0_C.TXT", 0x15f7,	0x22936 },
	{ "ARCADE_AKIHA_1.HA6", 0x19390,	0x22939 },
	{ "ARCADE_AKIHA_1.TXT", 0xa0,	0x2296c },
	{ "ARCADE_AKIHA_1_C.TXT", 0x1494,	0x2296d },
	{ "ARCADE_AKIHA_2.HA6", 0x18a28,	0x22970 },
	{ "ARCADE_AKIHA_2.TXT", 0xa0,	0x229a2 },
	{ "ARCADE_AKIHA_2_C.TXT", 0x14aa,	0x229a3 },
	{ "ARCADE_AKIHA_3.HA6", 0x1100c,	0x229a6 },
	{ "ARCADE_AKIHA_3.TXT", 0xa0,	0x229c9 },
	{ "ARCADE_AKIHA_3_C.TXT", 0x12ba,	0x229ca },
	{ "ARCADE_AKIHA_9.HA6", 0x1bae8,	0x229cd },
	{ "ARCADE_AKIHA_9.TXT", 0xa0,	0x22a05 },
	{ "ARCADE_AKIHA_9_C.TXT", 0x1448,	0x22a06 },
	{ "ARCADE_AKIHA_SELIST.TXT", 0xc49,	0x22a09 },
	{ "ARCADE_AOKO.ARX", 0x3519c0,	0x22a0b },
	{ "ARCADE_AOKO.CG2", 0x6db28,	0x230af },
	{ "ARCADE_AOKO.CPF", 0x2f36c,	0x2318b },
	{ "ARCADE_AOKO.DOF", 0xbcbfc,	0x231ea },
	{ "ARCADE_AOKO.HA6", 0x6efe8,	0x23364 },
	{ "ARCADE_AOKO.PAT", 0x0,	0x23442 },
	{ "ARCADE_AOKO.TXT", 0xc2,	0x23442 },
	{ "ARCADE_AOKO0.PAL", 0x400,	0x23443 },
	{ "ARCADE_AOKO1.PAL", 0x400,	0x23444 },
	{ "ARCADE_AOKO10.PAL", 0x400,	0x23445 },
	{ "ARCADE_AOKO11.PAL", 0x400,	0x23446 },
	{ "ARCADE_AOKO12.PAL", 0x400,	0x23447 },
	{ "ARCADE_AOKO13.PAL", 0x400,	0x23448 },
	{ "ARCADE_AOKO14.PAL", 0x400,	0x23449 },
	{ "ARCADE_AOKO15.PAL", 0x400,	0x2344a },
	{ "ARCADE_AOKO16.PAL", 0x400,	0x2344b },
	{ "ARCADE_AOKO17.PAL", 0x400,	0x2344c },
	{ "ARCADE_AOKO18.PAL", 0x400,	0x2344d },
	{ "ARCADE_AOKO19.PAL", 0x400,	0x2344e },
	{ "ARCADE_AOKO2.ARX", 0x34ec80,	0x2344f },
	{ "ARCADE_AOKO2.PAL", 0x400,	0x23aed },
	{ "ARCADE_AOKO20.PAL", 0x400,	0x23aee },
	{ "ARCADE_AOKO21.PAL", 0x400,	0x23aef },
	{ "ARCADE_AOKO22.PAL", 0x400,	0x23af0 },
	{ "ARCADE_AOKO23.PAL", 0x400,	0x23af1 },
	{ "ARCADE_AOKO24.PAL", 0x400,	0x23af2 },
	{ "ARCADE_AOKO25.PAL", 0x400,	0x23af3 },
	{ "ARCADE_AOKO26.PAL", 0x400,	0x23af4 },
	{ "ARCADE_AOKO27.PAL", 0x400,	0x23af5 },
	{ "ARCADE_AOKO28.PAL", 0x400,	0x23af6 },
	{ "ARCADE_AOKO29.PAL", 0x400,	0x23af7 },
	{ "ARCADE_AOKO3.PAL", 0x400,	0x23af8 },
	{ "ARCADE_AOKO30.PAL", 0x400,	0x23af9 },
	{ "ARCADE_AOKO31.PAL", 0x400,	0x23afa },
	{ "ARCADE_AOKO32.PAL", 0x400,	0x23afb },
	{ "ARCADE_AOKO33.PAL", 0x400,	0x23afc },
	{ "ARCADE_AOKO34.PAL", 0x400,	0x23afd },
	{ "ARCADE_AOKO35.PAL", 0x400,	0x23afe },
	{ "ARCADE_AOKO4.PAL", 0x400,	0x23aff },
	{ "ARCADE_AOKO5.PAL", 0x400,	0x23b00 },
	{ "ARCADE_AOKO6.PAL", 0x400,	0x23b01 },
	{ "ARCADE_AOKO7.PAL", 0x400,	0x23b02 },
	{ "ARCADE_AOKO8.PAL", 0x400,	0x23b03 },
	{ "ARCADE_AOKO9.PAL", 0x400,	0x23b04 },
	{ "ARCADE_AOKOPAL2.ARX", 0x1c980,	0x23b05 },
	{ "ARCADE_AOKO_0.HA6", 0x1f628,	0x23b3f },
	{ "ARCADE_AOKO_0.TXT", 0xa1,	0x23b7e },
	{ "ARCADE_AOKO_0_C.TXT", 0x19b6,	0x23b7f },
	{ "ARCADE_AOKO_1.HA6", 0x225d0,	0x23b83 },
	{ "ARCADE_AOKO_1.TXT", 0xa4,	0x23bc8 },
	{ "ARCADE_AOKO_1_C.TXT", 0x18cc,	0x23bc9 },
	{ "ARCADE_AOKO_2.HA6", 0x1afa0,	0x23bcd },
	{ "ARCADE_AOKO_2.TXT", 0xa2,	0x23c03 },
	{ "ARCADE_AOKO_2_C.TXT", 0x1653,	0x23c04 },
	{ "ARCADE_AOKO_3.HA6", 0x1100c,	0x23c07 },
	{ "ARCADE_AOKO_3.TXT", 0xa2,	0x23c2a },
	{ "ARCADE_AOKO_3_C.TXT", 0x1388,	0x23c2b },
	{ "ARCADE_AOKO_9.HA6", 0x4a680,	0x23c2e },
	{ "ARCADE_AOKO_9.TXT", 0xa2,	0x23cc3 },
	{ "ARCADE_AOKO_9_C.TXT", 0x140b,	0x23cc4 },
	{ "ARCADE_AOKO_C.TXT", 0x1794,	0x23cc7 },
	{ "ARCADE_AOKO_SELIST.TXT", 0x11cc,	0x23cca },
	{ "ARCADE_AOKO_WIN.TXT", 0x7e1,	0x23ccd },
	{ "ARCADE_ARC.ARX", 0x39fa40,	0x23cce },
	{ "ARCADE_ARC.CG2", 0x805a8,	0x2440e },
	{ "ARCADE_ARC.CPF", 0x2f36c,	0x2450f },
	{ "ARCADE_ARC.DOF", 0xde248,	0x2456e },
	{ "ARCADE_ARC.HA6", 0x7fa9c,	0x2472b },
	{ "ARCADE_ARC.PAT", 0x144c8,	0x2482b },
	{ "ARCADE_ARC.TXT", 0xcc,	0x24854 },
	{ "ARCADE_ARC0.PAL", 0x400,	0x24855 },
	{ "ARCADE_ARC1.PAL", 0x400,	0x24856 },
	{ "ARCADE_ARC10.PAL", 0x400,	0x24857 },
	{ "ARCADE_ARC11.PAL", 0x400,	0x24858 },
	{ "ARCADE_ARC12.PAL", 0x400,	0x24859 },
	{ "ARCADE_ARC13.PAL", 0x400,	0x2485a },
	{ "ARCADE_ARC14.PAL", 0x400,	0x2485b },
	{ "ARCADE_ARC15.PAL", 0x400,	0x2485c },
	{ "ARCADE_ARC16.PAL", 0x400,	0x2485d },
	{ "ARCADE_ARC17.PAL", 0x400,	0x2485e },
	{ "ARCADE_ARC18.PAL", 0x400,	0x2485f },
	{ "ARCADE_ARC19.PAL", 0x400,	0x24860 },
	{ "ARCADE_ARC2.ARX", 0x397d40,	0x24861 },
	{ "ARCADE_ARC2.PAL", 0x400,	0x24f91 },
	{ "ARCADE_ARC20.PAL", 0x400,	0x24f92 },
	{ "ARCADE_ARC21.PAL", 0x400,	0x24f93 },
	{ "ARCADE_ARC22.PAL", 0x400,	0x24f94 },
	{ "ARCADE_ARC23.PAL", 0x400,	0x24f95 },
	{ "ARCADE_ARC24.PAL", 0x400,	0x24f96 },
	{ "ARCADE_ARC25.PAL", 0x400,	0x24f97 },
	{ "ARCADE_ARC26.PAL", 0x400,	0x24f98 },
	{ "ARCADE_ARC27.PAL", 0x400,	0x24f99 },
	{ "ARCADE_ARC28.PAL", 0x400,	0x24f9a },
	{ "ARCADE_ARC29.PAL", 0x400,	0x24f9b },
	{ "ARCADE_ARC3.PAL", 0x400,	0x24f9c },
	{ "ARCADE_ARC30.PAL", 0x400,	0x24f9d },
	{ "ARCADE_ARC31.PAL", 0x400,	0x24f9e },
	{ "ARCADE_ARC32.PAL", 0x400,	0x24f9f },
	{ "ARCADE_ARC33.PAL", 0x400,	0x24fa0 },
	{ "ARCADE_ARC34.PAL", 0x400,	0x24fa1 },
	{ "ARCADE_ARC35.PAL", 0x400,	0x24fa2 },
	{ "ARCADE_ARC4.PAL", 0x400,	0x24fa3 },
	{ "ARCADE_ARC5.PAL", 0x400,	0x24fa4 },
	{ "ARCADE_ARC6.PAL", 0x400,	0x24fa5 },
	{ "ARCADE_ARC7.PAL", 0x400,	0x24fa6 },
	{ "ARCADE_ARC8.PAL", 0x400,	0x24fa7 },
	{ "ARCADE_ARC9.PAL", 0x400,	0x24fa8 },
	{ "ARCADE_ARCPAL2.ARX", 0x1f300,	0x24fa9 },
	{ "ARCADE_ARC_0.HA6", 0x18eec,	0x24fe8 },
	{ "ARCADE_ARC_0.TXT", 0xb1,	0x2501a },
	{ "ARCADE_ARC_0_C.TXT", 0x1cec,	0x2501b },
	{ "ARCADE_ARC_1.HA6", 0x1a350,	0x2501f },
	{ "ARCADE_ARC_1.TXT", 0xb4,	0x25054 },
	{ "ARCADE_ARC_1_C.TXT", 0x1abe,	0x25055 },
	{ "ARCADE_ARC_2.HA6", 0xc5c4,	0x25059 },
	{ "ARCADE_ARC_2.TXT", 0xb4,	0x25072 },
	{ "ARCADE_ARC_2_C.TXT", 0x1b0f,	0x25073 },
	{ "ARCADE_ARC_3.HA6", 0x1100c,	0x25077 },
	{ "ARCADE_ARC_3.TXT", 0xb4,	0x2509a },
	{ "ARCADE_ARC_3_C.TXT", 0x196e,	0x2509b },
	{ "ARCADE_ARC_9.HA6", 0x29794,	0x2509f },
	{ "ARCADE_ARC_9.TXT", 0xb4,	0x250f2 },
	{ "ARCADE_ARC_9_C.TXT", 0x1a0a,	0x250f3 },
	{ "ARCADE_ARC_SELIST.TXT", 0x143c,	0x250f7 },
	{ "ARCADE_ARC_WIN.TXT", 0xaba,	0x250fa },
	{ "ARCADE_B_AKIHA.ARX", 0x34d6c0,	0x2517d },
	{ "ARCADE_B_AKIHA.CG2", 0x74e18,	0x25818 },
	{ "ARCADE_B_AKIHA.CPF", 0x2f36c,	0x25902 },
	{ "ARCADE_B_AKIHA.HA6", 0x88024,	0x25961 },
	{ "ARCADE_B_AKIHA0.PAL", 0x400,	0x25a72 },
	{ "ARCADE_B_AKIHA1.PAL", 0x400,	0x25a73 },
	{ "ARCADE_B_AKIHA10.PAL", 0x400,	0x25a74 },
	{ "ARCADE_B_AKIHA11.PAL", 0x400,	0x25a75 },
	{ "ARCADE_B_AKIHA12.PAL", 0x400,	0x25a76 },
	{ "ARCADE_B_AKIHA13.PAL", 0x400,	0x25a77 },
	{ "ARCADE_B_AKIHA14.PAL", 0x400,	0x25a78 },
	{ "ARCADE_B_AKIHA15.PAL", 0x400,	0x25a79 },
	{ "ARCADE_B_AKIHA16.PAL", 0x400,	0x25a7a },
	{ "ARCADE_B_AKIHA17.PAL", 0x400,	0x25a7b },
	{ "ARCADE_B_AKIHA18.PAL", 0x400,	0x25a7c },
	{ "ARCADE_B_AKIHA19.PAL", 0x400,	0x25a7d },
	{ "ARCADE_B_AKIHA2.ARX", 0x345f40,	0x25a7e },
	{ "ARCADE_B_AKIHA2.PAL", 0x400,	0x2610a },
	{ "ARCADE_B_AKIHA20.PAL", 0x400,	0x2610b },
	{ "ARCADE_B_AKIHA21.PAL", 0x400,	0x2610c },
	{ "ARCADE_B_AKIHA22.PAL", 0x400,	0x2610d },
	{ "ARCADE_B_AKIHA23.PAL", 0x400,	0x2610e },
	{ "ARCADE_B_AKIHA24.PAL", 0x400,	0x2610f },
	{ "ARCADE_B_AKIHA25.PAL", 0x400,	0x26110 },
	{ "ARCADE_B_AKIHA26.PAL", 0x400,	0x26111 },
	{ "ARCADE_B_AKIHA27.PAL", 0x400,	0x26112 },
	{ "ARCADE_B_AKIHA28.PAL", 0x400,	0x26113 },
	{ "ARCADE_B_AKIHA29.PAL", 0x400,	0x26114 },
	{ "ARCADE_B_AKIHA3.PAL", 0x400,	0x26115 },
	{ "ARCADE_B_AKIHA30.PAL", 0x400,	0x26116 },
	{ "ARCADE_B_AKIHA31.PAL", 0x400,	0x26117 },
	{ "ARCADE_B_AKIHA32.PAL", 0x400,	0x26118 },
	{ "ARCADE_B_AKIHA33.PAL", 0x400,	0x26119 },
	{ "ARCADE_B_AKIHA34.PAL", 0x400,	0x2611a },
	{ "ARCADE_B_AKIHA35.PAL", 0x400,	0x2611b },
	{ "ARCADE_B_AKIHA4.PAL", 0x400,	0x2611c },
	{ "ARCADE_B_AKIHA5.PAL", 0x400,	0x2611d },
	{ "ARCADE_B_AKIHA6.PAL", 0x400,	0x2611e },
	{ "ARCADE_B_AKIHA7.PAL", 0x400,	0x2611f },
	{ "ARCADE_B_AKIHA8.PAL", 0x400,	0x26120 },
	{ "ARCADE_B_AKIHA9.PAL", 0x400,	0x26121 },
	{ "ARCADE_B_AKIHAPAL2.ARX", 0x1d1c0,	0x26122 },
	{ "ARCADE_B_AKIHA_0.TXT", 0x90,	0x2615d },
	{ "ARCADE_B_AKIHA_0_C.TXT", 0x1498,	0x2615e },
	{ "ARCADE_B_AKIHA_8.TXT", 0x90,	0x26161 },
	{ "ARCADE_B_AKIHA_8_C.TXT", 0x14e5,	0x26162 },
	{ "ARCADE_B_AKIHA_9.TXT", 0x90,	0x26165 },
	{ "ARCADE_B_AKIHA_9_C.TXT", 0x14e5,	0x26166 },
	{ "ARCADE_B_AKIHA_SELIST.TXT", 0xc49,	0x26169 },
	{ "ARCADE_B_AOKO.ARX", 0x3519c0,	0x2616b },
	{ "ARCADE_B_AOKO.CG2", 0x6d978,	0x2680f },
	{ "ARCADE_B_AOKO.CPF", 0x2f36c,	0x268eb },
	{ "ARCADE_B_AOKO.HA6", 0x78cdc,	0x2694a },
	{ "ARCADE_B_AOKO0.PAL", 0x400,	0x26a3c },
	{ "ARCADE_B_AOKO1.PAL", 0x400,	0x26a3d },
	{ "ARCADE_B_AOKO10.PAL", 0x400,	0x26a3e },
	{ "ARCADE_B_AOKO11.PAL", 0x400,	0x26a3f },
	{ "ARCADE_B_AOKO12.PAL", 0x400,	0x26a40 },
	{ "ARCADE_B_AOKO13.PAL", 0x400,	0x26a41 },
	{ "ARCADE_B_AOKO14.PAL", 0x400,	0x26a42 },
	{ "ARCADE_B_AOKO15.PAL", 0x400,	0x26a43 },
	{ "ARCADE_B_AOKO16.PAL", 0x400,	0x26a44 },
	{ "ARCADE_B_AOKO17.PAL", 0x400,	0x26a45 },
	{ "ARCADE_B_AOKO18.PAL", 0x400,	0x26a46 },
	{ "ARCADE_B_AOKO19.PAL", 0x400,	0x26a47 },
	{ "ARCADE_B_AOKO2.ARX", 0x34ec00,	0x26a48 },
	{ "ARCADE_B_AOKO2.PAL", 0x400,	0x270e6 },
	{ "ARCADE_B_AOKO20.PAL", 0x400,	0x270e7 },
	{ "ARCADE_B_AOKO21.PAL", 0x400,	0x270e8 },
	{ "ARCADE_B_AOKO22.PAL", 0x400,	0x270e9 },
	{ "ARCADE_B_AOKO23.PAL", 0x400,	0x270ea },
	{ "ARCADE_B_AOKO24.PAL", 0x400,	0x270eb },
	{ "ARCADE_B_AOKO25.PAL", 0x400,	0x270ec },
	{ "ARCADE_B_AOKO26.PAL", 0x400,	0x270ed },
	{ "ARCADE_B_AOKO27.PAL", 0x400,	0x270ee },
	{ "ARCADE_B_AOKO28.PAL", 0x400,	0x270ef },
	{ "ARCADE_B_AOKO29.PAL", 0x400,	0x270f0 },
	{ "ARCADE_B_AOKO3.PAL", 0x400,	0x270f1 },
	{ "ARCADE_B_AOKO30.PAL", 0x400,	0x270f2 },
	{ "ARCADE_B_AOKO31.PAL", 0x400,	0x270f3 },
	{ "ARCADE_B_AOKO32.PAL", 0x400,	0x270f4 },
	{ "ARCADE_B_AOKO33.PAL", 0x400,	0x270f5 },
	{ "ARCADE_B_AOKO34.PAL", 0x400,	0x270f6 },
	{ "ARCADE_B_AOKO35.PAL", 0x400,	0x270f7 },
	{ "ARCADE_B_AOKO4.PAL", 0x400,	0x270f8 },
	{ "ARCADE_B_AOKO5.PAL", 0x400,	0x270f9 },
	{ "ARCADE_B_AOKO6.PAL", 0x400,	0x270fa },
	{ "ARCADE_B_AOKO7.PAL", 0x400,	0x270fb },
	{ "ARCADE_B_AOKO8.PAL", 0x400,	0x270fc },
	{ "ARCADE_B_AOKO9.PAL", 0x400,	0x270fd },
	{ "ARCADE_B_AOKOPAL2.ARX", 0x1c980,	0x270fe },
	{ "ARCADE_B_AOKO_0.TXT", 0x92,	0x27138 },
	{ "ARCADE_B_AOKO_0_C.TXT", 0x16c4,	0x27139 },
	{ "ARCADE_B_AOKO_9.TXT", 0x92,	0x2713c },
	{ "ARCADE_B_AOKO_9_C.TXT", 0x1724,	0x2713d },
	{ "ARCADE_B_AOKO_SELIST.TXT", 0x11cc,	0x27140 },
	{ "ARCADE_B_ARC.ARX", 0x3f5a00,	0x27143 },
	{ "ARCADE_B_ARC.CG2", 0x8e1b0,	0x2792f },
	{ "ARCADE_B_ARC.CPF", 0x2f36c,	0x27a4c },
	{ "ARCADE_B_ARC.HA6", 0xb12b4,	0x27aab },
	{ "ARCADE_B_ARC0.PAL", 0x400,	0x27c0e },
	{ "ARCADE_B_ARC1.PAL", 0x400,	0x27c0f },
	{ "ARCADE_B_ARC10.PAL", 0x400,	0x27c10 },
	{ "ARCADE_B_ARC11.PAL", 0x400,	0x27c11 },
	{ "ARCADE_B_ARC12.PAL", 0x400,	0x27c12 },
	{ "ARCADE_B_ARC13.PAL", 0x400,	0x27c13 },
	{ "ARCADE_B_ARC14.PAL", 0x400,	0x27c14 },
	{ "ARCADE_B_ARC15.PAL", 0x400,	0x27c15 },
	{ "ARCADE_B_ARC16.PAL", 0x400,	0x27c16 },
	{ "ARCADE_B_ARC17.PAL", 0x400,	0x27c17 },
	{ "ARCADE_B_ARC18.PAL", 0x400,	0x27c18 },
	{ "ARCADE_B_ARC19.PAL", 0x400,	0x27c19 },
	{ "ARCADE_B_ARC2.ARX", 0x4140c0,	0x27c1a },
	{ "ARCADE_B_ARC2.PAL", 0x400,	0x28443 },
	{ "ARCADE_B_ARC20.PAL", 0x400,	0x28444 },
	{ "ARCADE_B_ARC21.PAL", 0x400,	0x28445 },
	{ "ARCADE_B_ARC22.PAL", 0x400,	0x28446 },
	{ "ARCADE_B_ARC23.PAL", 0x400,	0x28447 },
	{ "ARCADE_B_ARC24.PAL", 0x400,	0x28448 },
	{ "ARCADE_B_ARC25.PAL", 0x400,	0x28449 },
	{ "ARCADE_B_ARC26.PAL", 0x400,	0x2844a },
	{ "ARCADE_B_ARC27.PAL", 0x400,	0x2844b },
	{ "ARCADE_B_ARC28.PAL", 0x400,	0x2844c },
	{ "ARCADE_B_ARC29.PAL", 0x400,	0x2844d },
	{ "ARCADE_B_ARC3.PAL", 0x400,	0x2844e },
	{ "ARCADE_B_ARC30.PAL", 0x400,	0x2844f },
	{ "ARCADE_B_ARC31.PAL", 0x400,	0x28450 },
	{ "ARCADE_B_ARC32.PAL", 0x400,	0x28451 },
	{ "ARCADE_B_ARC33.PAL", 0x400,	0x28452 },
	{ "ARCADE_B_ARC34.PAL", 0x400,	0x28453 },
	{ "ARCADE_B_ARC35.PAL", 0x400,	0x28454 },
	{ "ARCADE_B_ARC4.PAL", 0x400,	0x28455 },
	{ "ARCADE_B_ARC5.PAL", 0x400,	0x28456 },
	{ "ARCADE_B_ARC6.PAL", 0x400,	0x28457 },
	{ "ARCADE_B_ARC7.PAL", 0x400,	0x28458 },
	{ "ARCADE_B_ARC8.PAL", 0x400,	0x28459 },
	{ "ARCADE_B_ARC9.PAL", 0x400,	0x2845a },
	{ "ARCADE_B_ARCPAL2.ARX", 0x24200,	0x2845b },
	{ "ARCADE_B_ARC_0.TXT", 0xa2,	0x284a4 },
	{ "ARCADE_B_ARC_0_C.TXT", 0x1da4,	0x284a5 },
	{ "ARCADE_B_ARC_9.TXT", 0xa4,	0x284a9 },
	{ "ARCADE_B_ARC_9_C.TXT", 0x127b,	0x284aa },
	{ "ARCADE_B_ARC_SELIST.TXT", 0xa07,	0x284ad },
	{ "ARCADE_B_HERMES.ARX", 0x40,	0x284af },
	{ "ARCADE_B_HERMES.CPF", 0x2f36c,	0x284b0 },
	{ "ARCADE_B_HERMES0.PAL", 0x400,	0x2850f },
	{ "ARCADE_B_HERMES1.PAL", 0x400,	0x28510 },
	{ "ARCADE_B_HERMES10.PAL", 0x400,	0x28511 },
	{ "ARCADE_B_HERMES11.PAL", 0x400,	0x28512 },
	{ "ARCADE_B_HERMES12.PAL", 0x400,	0x28513 },
	{ "ARCADE_B_HERMES13.PAL", 0x400,	0x28514 },
	{ "ARCADE_B_HERMES14.PAL", 0x400,	0x28515 },
	{ "ARCADE_B_HERMES15.PAL", 0x400,	0x28516 },
	{ "ARCADE_B_HERMES16.PAL", 0x400,	0x28517 },
	{ "ARCADE_B_HERMES17.PAL", 0x400,	0x28518 },
	{ "ARCADE_B_HERMES18.PAL", 0x400,	0x28519 },
	{ "ARCADE_B_HERMES19.PAL", 0x400,	0x2851a },
	{ "ARCADE_B_HERMES2.ARX", 0x40,	0x2851b },
	{ "ARCADE_B_HERMES2.PAL", 0x400,	0x2851c },
	{ "ARCADE_B_HERMES20.PAL", 0x400,	0x2851d },
	{ "ARCADE_B_HERMES21.PAL", 0x400,	0x2851e },
	{ "ARCADE_B_HERMES22.PAL", 0x400,	0x2851f },
	{ "ARCADE_B_HERMES23.PAL", 0x400,	0x28520 },
	{ "ARCADE_B_HERMES24.PAL", 0x400,	0x28521 },
	{ "ARCADE_B_HERMES25.PAL", 0x400,	0x28522 },
	{ "ARCADE_B_HERMES26.PAL", 0x400,	0x28523 },
	{ "ARCADE_B_HERMES27.PAL", 0x400,	0x28524 },
	{ "ARCADE_B_HERMES28.PAL", 0x400,	0x28525 },
	{ "ARCADE_B_HERMES29.PAL", 0x400,	0x28526 },
	{ "ARCADE_B_HERMES3.PAL", 0x400,	0x28527 },
	{ "ARCADE_B_HERMES30.PAL", 0x400,	0x28528 },
	{ "ARCADE_B_HERMES31.PAL", 0x400,	0x28529 },
	{ "ARCADE_B_HERMES32.PAL", 0x400,	0x2852a },
	{ "ARCADE_B_HERMES33.PAL", 0x400,	0x2852b },
	{ "ARCADE_B_HERMES34.PAL", 0x400,	0x2852c },
	{ "ARCADE_B_HERMES35.PAL", 0x400,	0x2852d },
	{ "ARCADE_B_HERMES4.PAL", 0x400,	0x2852e },
	{ "ARCADE_B_HERMES5.PAL", 0x400,	0x2852f },
	{ "ARCADE_B_HERMES6.PAL", 0x400,	0x28530 },
	{ "ARCADE_B_HERMES7.PAL", 0x400,	0x28531 },
	{ "ARCADE_B_HERMES8.PAL", 0x400,	0x28532 },
	{ "ARCADE_B_HERMES9.PAL", 0x400,	0x28533 },
	{ "ARCADE_B_HERMESPAL2.ARX", 0x40,	0x28534 },
	{ "ARCADE_B_HERMES_9.TXT", 0xa8,	0x28535 },
	{ "ARCADE_B_HERMES_9_C.TXT", 0x5b9,	0x28536 },
	{ "ARCADE_B_HERMES_SELIST.TXT", 0x6fb,	0x28537 },
	{ "ARCADE_B_KOHAKU.CPF", 0x2f36c,	0x28538 },
	{ "ARCADE_B_KOHAKU_M.ARX", 0x17c680,	0x28597 },
	{ "ARCADE_B_KOHAKU_M.CG2", 0x45868,	0x28890 },
	{ "ARCADE_B_KOHAKU_M.CPF", 0x2f36c,	0x2891c },
	{ "ARCADE_B_KOHAKU_M.HA6", 0x42f20,	0x2897b },
	{ "ARCADE_B_KOHAKU_M.PAL", 0x10004,	0x28a01 },
	{ "ARCADE_B_KOHAKU_M0.PAL", 0x400,	0x28a22 },
	{ "ARCADE_B_KOHAKU_M1.PAL", 0x400,	0x28a23 },
	{ "ARCADE_B_KOHAKU_M10.PAL", 0x400,	0x28a24 },
	{ "ARCADE_B_KOHAKU_M11.PAL", 0x400,	0x28a25 },
	{ "ARCADE_B_KOHAKU_M12.PAL", 0x400,	0x28a26 },
	{ "ARCADE_B_KOHAKU_M13.PAL", 0x400,	0x28a27 },
	{ "ARCADE_B_KOHAKU_M14.PAL", 0x400,	0x28a28 },
	{ "ARCADE_B_KOHAKU_M15.PAL", 0x400,	0x28a29 },
	{ "ARCADE_B_KOHAKU_M16.PAL", 0x400,	0x28a2a },
	{ "ARCADE_B_KOHAKU_M17.PAL", 0x400,	0x28a2b },
	{ "ARCADE_B_KOHAKU_M18.PAL", 0x400,	0x28a2c },
	{ "ARCADE_B_KOHAKU_M19.PAL", 0x400,	0x28a2d },
	{ "ARCADE_B_KOHAKU_M2.ARX", 0x178500,	0x28a2e },
	{ "ARCADE_B_KOHAKU_M2.PAL", 0x400,	0x28d1f },
	{ "ARCADE_B_KOHAKU_M20.PAL", 0x400,	0x28d20 },
	{ "ARCADE_B_KOHAKU_M21.PAL", 0x400,	0x28d21 },
	{ "ARCADE_B_KOHAKU_M22.PAL", 0x400,	0x28d22 },
	{ "ARCADE_B_KOHAKU_M23.PAL", 0x400,	0x28d23 },
	{ "ARCADE_B_KOHAKU_M24.PAL", 0x400,	0x28d24 },
	{ "ARCADE_B_KOHAKU_M25.PAL", 0x400,	0x28d25 },
	{ "ARCADE_B_KOHAKU_M26.PAL", 0x400,	0x28d26 },
	{ "ARCADE_B_KOHAKU_M27.PAL", 0x400,	0x28d27 },
	{ "ARCADE_B_KOHAKU_M28.PAL", 0x400,	0x28d28 },
	{ "ARCADE_B_KOHAKU_M29.PAL", 0x400,	0x28d29 },
	{ "ARCADE_B_KOHAKU_M3.PAL", 0x400,	0x28d2a },
	{ "ARCADE_B_KOHAKU_M30.PAL", 0x400,	0x28d2b },
	{ "ARCADE_B_KOHAKU_M31.PAL", 0x400,	0x28d2c },
	{ "ARCADE_B_KOHAKU_M32.PAL", 0x400,	0x28d2d },
	{ "ARCADE_B_KOHAKU_M33.PAL", 0x400,	0x28d2e },
	{ "ARCADE_B_KOHAKU_M34.PAL", 0x400,	0x28d2f },
	{ "ARCADE_B_KOHAKU_M35.PAL", 0x400,	0x28d30 },
	{ "ARCADE_B_KOHAKU_M4.PAL", 0x400,	0x28d31 },
	{ "ARCADE_B_KOHAKU_M5.PAL", 0x400,	0x28d32 },
	{ "ARCADE_B_KOHAKU_M6.PAL", 0x400,	0x28d33 },
	{ "ARCADE_B_KOHAKU_M7.PAL", 0x400,	0x28d34 },
	{ "ARCADE_B_KOHAKU_M8.PAL", 0x400,	0x28d35 },
	{ "ARCADE_B_KOHAKU_M9.PAL", 0x400,	0x28d36 },
	{ "ARCADE_B_KOHAKU_MPAL2.ARX", 0x3e80,	0x28d37 },
	{ "ARCADE_B_KOHAKU_M_0.TXT", 0xac,	0x28d3f },
	{ "ARCADE_B_KOHAKU_M_0_C.TXT", 0x1280,	0x28d40 },
	{ "ARCADE_B_KOHAKU_M_9.TXT", 0xac,	0x28d43 },
	{ "ARCADE_B_KOHAKU_M_9_C.TXT", 0x12ba,	0x28d44 },
	{ "ARCADE_B_KOHAKU_M_SELIST.TXT", 0xc99,	0x28d47 },
	{ "ARCADE_B_MIYAKO.ARX", 0x176200,	0x28d49 },
	{ "ARCADE_B_MIYAKO.CG2", 0x3ca48,	0x29036 },
	{ "ARCADE_B_MIYAKO.CPF", 0x2f36c,	0x290b0 },
	{ "ARCADE_B_MIYAKO.HA6", 0x4d568,	0x2910f },
	{ "ARCADE_B_MIYAKO0.PAL", 0x400,	0x291aa },
	{ "ARCADE_B_MIYAKO1.PAL", 0x400,	0x291ab },
	{ "ARCADE_B_MIYAKO10.PAL", 0x400,	0x291ac },
	{ "ARCADE_B_MIYAKO11.PAL", 0x400,	0x291ad },
	{ "ARCADE_B_MIYAKO12.PAL", 0x400,	0x291ae },
	{ "ARCADE_B_MIYAKO13.PAL", 0x400,	0x291af },
	{ "ARCADE_B_MIYAKO14.PAL", 0x400,	0x291b0 },
	{ "ARCADE_B_MIYAKO15.PAL", 0x400,	0x291b1 },
	{ "ARCADE_B_MIYAKO16.PAL", 0x400,	0x291b2 },
	{ "ARCADE_B_MIYAKO17.PAL", 0x400,	0x291b3 },
	{ "ARCADE_B_MIYAKO18.PAL", 0x400,	0x291b4 },
	{ "ARCADE_B_MIYAKO19.PAL", 0x400,	0x291b5 },
	{ "ARCADE_B_MIYAKO2.ARX", 0x1746c0,	0x291b6 },
	{ "ARCADE_B_MIYAKO2.PAL", 0x400,	0x2949f },
	{ "ARCADE_B_MIYAKO20.PAL", 0x400,	0x294a0 },
	{ "ARCADE_B_MIYAKO21.PAL", 0x400,	0x294a1 },
	{ "ARCADE_B_MIYAKO22.PAL", 0x400,	0x294a2 },
	{ "ARCADE_B_MIYAKO23.PAL", 0x400,	0x294a3 },
	{ "ARCADE_B_MIYAKO24.PAL", 0x400,	0x294a4 },
	{ "ARCADE_B_MIYAKO25.PAL", 0x400,	0x294a5 },
	{ "ARCADE_B_MIYAKO26.PAL", 0x400,	0x294a6 },
	{ "ARCADE_B_MIYAKO27.PAL", 0x400,	0x294a7 },
	{ "ARCADE_B_MIYAKO28.PAL", 0x400,	0x294a8 },
	{ "ARCADE_B_MIYAKO29.PAL", 0x400,	0x294a9 },
	{ "ARCADE_B_MIYAKO3.PAL", 0x400,	0x294aa },
	{ "ARCADE_B_MIYAKO30.PAL", 0x400,	0x294ab },
	{ "ARCADE_B_MIYAKO31.PAL", 0x400,	0x294ac },
	{ "ARCADE_B_MIYAKO32.PAL", 0x400,	0x294ad },
	{ "ARCADE_B_MIYAKO33.PAL", 0x400,	0x294ae },
	{ "ARCADE_B_MIYAKO34.PAL", 0x400,	0x294af },
	{ "ARCADE_B_MIYAKO35.PAL", 0x400,	0x294b0 },
	{ "ARCADE_B_MIYAKO4.PAL", 0x400,	0x294b1 },
	{ "ARCADE_B_MIYAKO5.PAL", 0x400,	0x294b2 },
	{ "ARCADE_B_MIYAKO6.PAL", 0x400,	0x294b3 },
	{ "ARCADE_B_MIYAKO7.PAL", 0x400,	0x294b4 },
	{ "ARCADE_B_MIYAKO8.PAL", 0x400,	0x294b5 },
	{ "ARCADE_B_MIYAKO9.PAL", 0x400,	0x294b6 },
	{ "ARCADE_B_MIYAKOPAL2.ARX", 0x5bc0,	0x294b7 },
	{ "ARCADE_B_MIYAKO_0.TXT", 0xa6,	0x294c3 },
	{ "ARCADE_B_MIYAKO_0_C.TXT", 0x138f,	0x294c4 },
	{ "ARCADE_B_MIYAKO_9.TXT", 0xa6,	0x294c7 },
	{ "ARCADE_B_MIYAKO_9_C.TXT", 0x1411,	0x294c8 },
	{ "ARCADE_B_MIYAKO_SELIST.TXT", 0xbc3,	0x294cb },
	{ "ARCADE_B_M_HISUI_P.ARX", 0x8abc0,	0x294cd },
	{ "ARCADE_B_M_HISUI_P.CG2", 0x2e480,	0x295e3 },
	{ "ARCADE_B_M_HISUI_P.CPF", 0x2f36c,	0x29640 },
	{ "ARCADE_B_M_HISUI_P.HA6", 0x12fb4,	0x2969f },
	{ "ARCADE_B_M_HISUI_P.PAL", 0x10004,	0x296c5 },
	{ "ARCADE_B_M_HISUI_P0.PAL", 0x400,	0x296e6 },
	{ "ARCADE_B_M_HISUI_P1.PAL", 0x400,	0x296e7 },
	{ "ARCADE_B_M_HISUI_P10.PAL", 0x400,	0x296e8 },
	{ "ARCADE_B_M_HISUI_P11.PAL", 0x400,	0x296e9 },
	{ "ARCADE_B_M_HISUI_P12.PAL", 0x400,	0x296ea },
	{ "ARCADE_B_M_HISUI_P13.PAL", 0x400,	0x296eb },
	{ "ARCADE_B_M_HISUI_P14.PAL", 0x400,	0x296ec },
	{ "ARCADE_B_M_HISUI_P15.PAL", 0x400,	0x296ed },
	{ "ARCADE_B_M_HISUI_P16.PAL", 0x400,	0x296ee },
	{ "ARCADE_B_M_HISUI_P17.PAL", 0x400,	0x296ef },
	{ "ARCADE_B_M_HISUI_P18.PAL", 0x400,	0x296f0 },
	{ "ARCADE_B_M_HISUI_P19.PAL", 0x400,	0x296f1 },
	{ "ARCADE_B_M_HISUI_P2.ARX", 0x87500,	0x296f2 },
	{ "ARCADE_B_M_HISUI_P2.PAL", 0x400,	0x29801 },
	{ "ARCADE_B_M_HISUI_P20.PAL", 0x400,	0x29802 },
	{ "ARCADE_B_M_HISUI_P21.PAL", 0x400,	0x29803 },
	{ "ARCADE_B_M_HISUI_P22.PAL", 0x400,	0x29804 },
	{ "ARCADE_B_M_HISUI_P23.PAL", 0x400,	0x29805 },
	{ "ARCADE_B_M_HISUI_P24.PAL", 0x400,	0x29806 },
	{ "ARCADE_B_M_HISUI_P25.PAL", 0x400,	0x29807 },
	{ "ARCADE_B_M_HISUI_P26.PAL", 0x400,	0x29808 },
	{ "ARCADE_B_M_HISUI_P27.PAL", 0x400,	0x29809 },
	{ "ARCADE_B_M_HISUI_P28.PAL", 0x400,	0x2980a },
	{ "ARCADE_B_M_HISUI_P29.PAL", 0x400,	0x2980b },
	{ "ARCADE_B_M_HISUI_P3.PAL", 0x400,	0x2980c },
	{ "ARCADE_B_M_HISUI_P30.PAL", 0x400,	0x2980d },
	{ "ARCADE_B_M_HISUI_P31.PAL", 0x400,	0x2980e },
	{ "ARCADE_B_M_HISUI_P32.PAL", 0x400,	0x2980f },
	{ "ARCADE_B_M_HISUI_P33.PAL", 0x400,	0x29810 },
	{ "ARCADE_B_M_HISUI_P34.PAL", 0x400,	0x29811 },
	{ "ARCADE_B_M_HISUI_P35.PAL", 0x400,	0x29812 },
	{ "ARCADE_B_M_HISUI_P4.PAL", 0x400,	0x29813 },
	{ "ARCADE_B_M_HISUI_P5.PAL", 0x400,	0x29814 },
	{ "ARCADE_B_M_HISUI_P6.PAL", 0x400,	0x29815 },
	{ "ARCADE_B_M_HISUI_P7.PAL", 0x400,	0x29816 },
	{ "ARCADE_B_M_HISUI_P8.PAL", 0x400,	0x29817 },
	{ "ARCADE_B_M_HISUI_P9.PAL", 0x400,	0x29818 },
	{ "ARCADE_B_M_HISUI_PPAL2.ARX", 0x2180,	0x29819 },
	{ "ARCADE_B_M_HISUI_P_0.TXT", 0xb2,	0x2981e },
	{ "ARCADE_B_M_HISUI_P_0_C.TXT", 0x317,	0x2981f },
	{ "ARCADE_B_M_HISUI_P_9.TXT", 0xb2,	0x29820 },
	{ "ARCADE_B_M_HISUI_P_9_C.TXT", 0x317,	0x29821 },
	{ "ARCADE_B_M_HISUI_P_SELIST.TXT", 0x2ea,	0x29822 },
	{ "ARCADE_B_RYOUGI.CG2", 0x7ac08,	0x29823 },
	{ "ARCADE_B_RYOUGI.CPF", 0x2f36c,	0x29919 },
	{ "ARCADE_B_RYOUGI.HA6", 0xb18e8,	0x29978 },
	{ "ARCADE_B_RYOUGI0.PAL", 0x400,	0x29adc },
	{ "ARCADE_B_RYOUGI1.PAL", 0x400,	0x29add },
	{ "ARCADE_B_RYOUGI10.PAL", 0x400,	0x29ade },
	{ "ARCADE_B_RYOUGI11.PAL", 0x400,	0x29adf },
	{ "ARCADE_B_RYOUGI12.PAL", 0x400,	0x29ae0 },
	{ "ARCADE_B_RYOUGI13.PAL", 0x400,	0x29ae1 },
	{ "ARCADE_B_RYOUGI14.PAL", 0x400,	0x29ae2 },
	{ "ARCADE_B_RYOUGI15.PAL", 0x400,	0x29ae3 },
	{ "ARCADE_B_RYOUGI16.PAL", 0x400,	0x29ae4 },
	{ "ARCADE_B_RYOUGI17.PAL", 0x400,	0x29ae5 },
	{ "ARCADE_B_RYOUGI18.PAL", 0x400,	0x29ae6 },
	{ "ARCADE_B_RYOUGI19.PAL", 0x400,	0x29ae7 },
	{ "ARCADE_B_RYOUGI2.ARX", 0x2f8f40,	0x29ae8 },
	{ "ARCADE_B_RYOUGI2.PAL", 0x400,	0x2a0da },
	{ "ARCADE_B_RYOUGI20.PAL", 0x400,	0x2a0db },
	{ "ARCADE_B_RYOUGI21.PAL", 0x400,	0x2a0dc },
	{ "ARCADE_B_RYOUGI22.PAL", 0x400,	0x2a0dd },
	{ "ARCADE_B_RYOUGI23.PAL", 0x400,	0x2a0de },
	{ "ARCADE_B_RYOUGI24.PAL", 0x400,	0x2a0df },
	{ "ARCADE_B_RYOUGI25.PAL", 0x400,	0x2a0e0 },
	{ "ARCADE_B_RYOUGI26.PAL", 0x400,	0x2a0e1 },
	{ "ARCADE_B_RYOUGI27.PAL", 0x400,	0x2a0e2 },
	{ "ARCADE_B_RYOUGI28.PAL", 0x400,	0x2a0e3 },
	{ "ARCADE_B_RYOUGI29.PAL", 0x400,	0x2a0e4 },
	{ "ARCADE_B_RYOUGI3.PAL", 0x400,	0x2a0e5 },
	{ "ARCADE_B_RYOUGI30.PAL", 0x400,	0x2a0e6 },
	{ "ARCADE_B_RYOUGI31.PAL", 0x400,	0x2a0e7 },
	{ "ARCADE_B_RYOUGI32.PAL", 0x400,	0x2a0e8 },
	{ "ARCADE_B_RYOUGI33.PAL", 0x400,	0x2a0e9 },
	{ "ARCADE_B_RYOUGI34.PAL", 0x400,	0x2a0ea },
	{ "ARCADE_B_RYOUGI35.PAL", 0x400,	0x2a0eb },
	{ "ARCADE_B_RYOUGI4.PAL", 0x400,	0x2a0ec },
	{ "ARCADE_B_RYOUGI5.PAL", 0x400,	0x2a0ed },
	{ "ARCADE_B_RYOUGI6.PAL", 0x400,	0x2a0ee },
	{ "ARCADE_B_RYOUGI7.PAL", 0x400,	0x2a0ef },
	{ "ARCADE_B_RYOUGI8.PAL", 0x400,	0x2a0f0 },
	{ "ARCADE_B_RYOUGI9.PAL", 0x400,	0x2a0f1 },
	{ "ARCADE_B_RYOUGIPAL2.ARX", 0x13800,	0x2a0f2 },
	{ "ARCADE_B_RYOUGI_9.HA6", 0x1b680,	0x2a119 },
	{ "ARCADE_B_RYOUGI_9.TXT", 0xc8,	0x2a150 },
	{ "ARCADE_B_RYOUGI_9_C.TXT", 0x1f6f,	0x2a151 },
	{ "ARCADE_B_RYOUGI_SELIST.TXT", 0x1e59,	0x2a155 },
	{ "ARCADE_B_WARAKIA.ARX", 0x34d780,	0x2a159 },
	{ "ARCADE_B_WARAKIA.CG2", 0x72e38,	0x2a7f4 },
	{ "ARCADE_B_WARAKIA.CPF", 0x2f36c,	0x2a8da },
	{ "ARCADE_B_WARAKIA.HA6", 0x68fa0,	0x2a939 },
	{ "ARCADE_B_WARAKIA.PAL", 0x10004,	0x2aa0b },
	{ "ARCADE_B_WARAKIA0.PAL", 0x400,	0x2aa2c },
	{ "ARCADE_B_WARAKIA1.PAL", 0x400,	0x2aa2d },
	{ "ARCADE_B_WARAKIA10.PAL", 0x400,	0x2aa2e },
	{ "ARCADE_B_WARAKIA11.PAL", 0x400,	0x2aa2f },
	{ "ARCADE_B_WARAKIA12.PAL", 0x400,	0x2aa30 },
	{ "ARCADE_B_WARAKIA13.PAL", 0x400,	0x2aa31 },
	{ "ARCADE_B_WARAKIA14.PAL", 0x400,	0x2aa32 },
	{ "ARCADE_B_WARAKIA15.PAL", 0x400,	0x2aa33 },
	{ "ARCADE_B_WARAKIA16.PAL", 0x400,	0x2aa34 },
	{ "ARCADE_B_WARAKIA17.PAL", 0x400,	0x2aa35 },
	{ "ARCADE_B_WARAKIA18.PAL", 0x400,	0x2aa36 },
	{ "ARCADE_B_WARAKIA19.PAL", 0x400,	0x2aa37 },
	{ "ARCADE_B_WARAKIA2.ARX", 0x339cc0,	0x2aa38 },
	{ "ARCADE_B_WARAKIA2.PAL", 0x400,	0x2b0ac },
	{ "ARCADE_B_WARAKIA20.PAL", 0x400,	0x2b0ad },
	{ "ARCADE_B_WARAKIA21.PAL", 0x400,	0x2b0ae },
	{ "ARCADE_B_WARAKIA22.PAL", 0x400,	0x2b0af },
	{ "ARCADE_B_WARAKIA23.PAL", 0x400,	0x2b0b0 },
	{ "ARCADE_B_WARAKIA24.PAL", 0x400,	0x2b0b1 },
	{ "ARCADE_B_WARAKIA25.PAL", 0x400,	0x2b0b2 },
	{ "ARCADE_B_WARAKIA26.PAL", 0x400,	0x2b0b3 },
	{ "ARCADE_B_WARAKIA27.PAL", 0x400,	0x2b0b4 },
	{ "ARCADE_B_WARAKIA28.PAL", 0x400,	0x2b0b5 },
	{ "ARCADE_B_WARAKIA29.PAL", 0x400,	0x2b0b6 },
	{ "ARCADE_B_WARAKIA3.PAL", 0x400,	0x2b0b7 },
	{ "ARCADE_B_WARAKIA30.PAL", 0x400,	0x2b0b8 },
	{ "ARCADE_B_WARAKIA31.PAL", 0x400,	0x2b0b9 },
	{ "ARCADE_B_WARAKIA32.PAL", 0x400,	0x2b0ba },
	{ "ARCADE_B_WARAKIA33.PAL", 0x400,	0x2b0bb },
	{ "ARCADE_B_WARAKIA34.PAL", 0x400,	0x2b0bc },
	{ "ARCADE_B_WARAKIA35.PAL", 0x400,	0x2b0bd },
	{ "ARCADE_B_WARAKIA4.PAL", 0x400,	0x2b0be },
	{ "ARCADE_B_WARAKIA5.PAL", 0x400,	0x2b0bf },
	{ "ARCADE_B_WARAKIA6.PAL", 0x400,	0x2b0c0 },
	{ "ARCADE_B_WARAKIA7.PAL", 0x400,	0x2b0c1 },
	{ "ARCADE_B_WARAKIA8.PAL", 0x400,	0x2b0c2 },
	{ "ARCADE_B_WARAKIA9.PAL", 0x400,	0x2b0c3 },
	{ "ARCADE_B_WARAKIAPAL2.ARX", 0x1c140,	0x2b0c4 },
	{ "ARCADE_B_WARAKIA_0.TXT", 0xaa,	0x2b0fd },
	{ "ARCADE_B_WARAKIA_0_C.TXT", 0x10c2,	0x2b0fe },
	{ "ARCADE_B_WARAKIA_9.TXT", 0xaa,	0x2b101 },
	{ "ARCADE_B_WARAKIA_9_C.TXT", 0x1100,	0x2b102 },
	{ "ARCADE_B_WARAKIA_SELIST.TXT", 0xce0,	0x2b105 },
	{ "ARCADE_B_WLEN.ARX", 0x429300,	0x2b107 },
	{ "ARCADE_B_WLEN.CG2", 0x88000,	0x2b95a },
	{ "ARCADE_B_WLEN.CPF", 0x2f36c,	0x2ba6a },
	{ "ARCADE_B_WLEN.HA6", 0x7741c,	0x2bac9 },
	{ "ARCADE_B_WLEN0.PAL", 0x400,	0x2bbb8 },
	{ "ARCADE_B_WLEN1.PAL", 0x400,	0x2bbb9 },
	{ "ARCADE_B_WLEN10.PAL", 0x400,	0x2bbba },
	{ "ARCADE_B_WLEN11.PAL", 0x400,	0x2bbbb },
	{ "ARCADE_B_WLEN12.PAL", 0x400,	0x2bbbc },
	{ "ARCADE_B_WLEN13.PAL", 0x400,	0x2bbbd },
	{ "ARCADE_B_WLEN14.PAL", 0x400,	0x2bbbe },
	{ "ARCADE_B_WLEN15.PAL", 0x400,	0x2bbbf },
	{ "ARCADE_B_WLEN16.PAL", 0x400,	0x2bbc0 },
	{ "ARCADE_B_WLEN17.PAL", 0x400,	0x2bbc1 },
	{ "ARCADE_B_WLEN18.PAL", 0x400,	0x2bbc2 },
	{ "ARCADE_B_WLEN19.PAL", 0x400,	0x2bbc3 },
	{ "ARCADE_B_WLEN2.ARX", 0x423500,	0x2bbc4 },
	{ "ARCADE_B_WLEN2.PAL", 0x400,	0x2c40b },
	{ "ARCADE_B_WLEN20.PAL", 0x400,	0x2c40c },
	{ "ARCADE_B_WLEN21.PAL", 0x400,	0x2c40d },
	{ "ARCADE_B_WLEN22.PAL", 0x400,	0x2c40e },
	{ "ARCADE_B_WLEN23.PAL", 0x400,	0x2c40f },
	{ "ARCADE_B_WLEN24.PAL", 0x400,	0x2c410 },
	{ "ARCADE_B_WLEN25.PAL", 0x400,	0x2c411 },
	{ "ARCADE_B_WLEN26.PAL", 0x400,	0x2c412 },
	{ "ARCADE_B_WLEN27.PAL", 0x400,	0x2c413 },
	{ "ARCADE_B_WLEN28.PAL", 0x400,	0x2c414 },
	{ "ARCADE_B_WLEN29.PAL", 0x400,	0x2c415 },
	{ "ARCADE_B_WLEN3.PAL", 0x400,	0x2c416 },
	{ "ARCADE_B_WLEN30.PAL", 0x400,	0x2c417 },
	{ "ARCADE_B_WLEN31.PAL", 0x400,	0x2c418 },
	{ "ARCADE_B_WLEN32.PAL", 0x400,	0x2c419 },
	{ "ARCADE_B_WLEN33.PAL", 0x400,	0x2c41a },
	{ "ARCADE_B_WLEN34.PAL", 0x400,	0x2c41b },
	{ "ARCADE_B_WLEN35.PAL", 0x400,	0x2c41c },
	{ "ARCADE_B_WLEN4.PAL", 0x400,	0x2c41d },
	{ "ARCADE_B_WLEN5.PAL", 0x400,	0x2c41e },
	{ "ARCADE_B_WLEN6.PAL", 0x400,	0x2c41f },
	{ "ARCADE_B_WLEN7.PAL", 0x400,	0x2c420 },
	{ "ARCADE_B_WLEN8.PAL", 0x400,	0x2c421 },
	{ "ARCADE_B_WLEN9.PAL", 0x400,	0x2c422 },
	{ "ARCADE_B_WLENPAL2.ARX", 0x2bec0,	0x2c423 },
	{ "ARCADE_B_WLEN_0.TXT", 0x8d,	0x2c47b },
	{ "ARCADE_B_WLEN_0_C.TXT", 0x1341,	0x2c47c },
	{ "ARCADE_B_WLEN_9.TXT", 0x8d,	0x2c47f },
	{ "ARCADE_B_WLEN_9_C.TXT", 0x137b,	0x2c480 },
	{ "ARCADE_B_WLEN_SELIST.TXT", 0x950,	0x2c483 },
	{ "ARCADE_CHARASELECT.TXT", 0x19d2,	0x2c485 },
	{ "ARCADE_CIEL.ARX", 0x269d00,	0x2c489 },
	{ "ARCADE_CIEL.CG2", 0x70900,	0x2c95d },
	{ "ARCADE_CIEL.CPF", 0x2f36c,	0x2ca3f },
	{ "ARCADE_CIEL.DOF", 0xf0abc,	0x2ca9e },
	{ "ARCADE_CIEL.HA6", 0x84b6c,	0x2cc80 },
	{ "ARCADE_CIEL.PAT", 0x144c8,	0x2cd8a },
	{ "ARCADE_CIEL.TXT", 0xc2,	0x2cdb3 },
	{ "ARCADE_CIEL0.PAL", 0x400,	0x2cdb4 },
	{ "ARCADE_CIEL1.PAL", 0x400,	0x2cdb5 },
	{ "ARCADE_CIEL10.PAL", 0x400,	0x2cdb6 },
	{ "ARCADE_CIEL11.PAL", 0x400,	0x2cdb7 },
	{ "ARCADE_CIEL12.PAL", 0x400,	0x2cdb8 },
	{ "ARCADE_CIEL13.PAL", 0x400,	0x2cdb9 },
	{ "ARCADE_CIEL14.PAL", 0x400,	0x2cdba },
	{ "ARCADE_CIEL15.PAL", 0x400,	0x2cdbb },
	{ "ARCADE_CIEL16.PAL", 0x400,	0x2cdbc },
	{ "ARCADE_CIEL17.PAL", 0x400,	0x2cdbd },
	{ "ARCADE_CIEL18.PAL", 0x400,	0x2cdbe },
	{ "ARCADE_CIEL19.PAL", 0x400,	0x2cdbf },
	{ "ARCADE_CIEL2.ARX", 0x267680,	0x2cdc0 },
	{ "ARCADE_CIEL2.PAL", 0x400,	0x2d28f },
	{ "ARCADE_CIEL20.PAL", 0x400,	0x2d290 },
	{ "ARCADE_CIEL21.PAL", 0x400,	0x2d291 },
	{ "ARCADE_CIEL22.PAL", 0x400,	0x2d292 },
	{ "ARCADE_CIEL23.PAL", 0x400,	0x2d293 },
	{ "ARCADE_CIEL24.PAL", 0x400,	0x2d294 },
	{ "ARCADE_CIEL25.PAL", 0x400,	0x2d295 },
	{ "ARCADE_CIEL26.PAL", 0x400,	0x2d296 },
	{ "ARCADE_CIEL27.PAL", 0x400,	0x2d297 },
	{ "ARCADE_CIEL28.PAL", 0x400,	0x2d298 },
	{ "ARCADE_CIEL29.PAL", 0x400,	0x2d299 },
	{ "ARCADE_CIEL3.PAL", 0x400,	0x2d29a },
	{ "ARCADE_CIEL30.PAL", 0x400,	0x2d29b },
	{ "ARCADE_CIEL31.PAL", 0x400,	0x2d29c },
	{ "ARCADE_CIEL32.PAL", 0x400,	0x2d29d },
	{ "ARCADE_CIEL33.PAL", 0x400,	0x2d29e },
	{ "ARCADE_CIEL34.PAL", 0x400,	0x2d29f },
	{ "ARCADE_CIEL35.PAL", 0x400,	0x2d2a0 },
	{ "ARCADE_CIEL4.PAL", 0x400,	0x2d2a1 },
	{ "ARCADE_CIEL5.PAL", 0x400,	0x2d2a2 },
	{ "ARCADE_CIEL6.PAL", 0x400,	0x2d2a3 },
	{ "ARCADE_CIEL7.PAL", 0x400,	0x2d2a4 },
	{ "ARCADE_CIEL8.PAL", 0x400,	0x2d2a5 },
	{ "ARCADE_CIEL9.PAL", 0x400,	0x2d2a6 },
	{ "ARCADE_CIELPAL2.ARX", 0xc7c0,	0x2d2a7 },
	{ "ARCADE_CIEL_0.HA6", 0x17fe4,	0x2d2c0 },
	{ "ARCADE_CIEL_0.TXT", 0xa7,	0x2d2f0 },
	{ "ARCADE_CIEL_0_C.TXT", 0x232c,	0x2d2f1 },
	{ "ARCADE_CIEL_1.HA6", 0x1ab54,	0x2d2f6 },
	{ "ARCADE_CIEL_1.TXT", 0xa5,	0x2d32c },
	{ "ARCADE_CIEL_1_C.TXT", 0x163c,	0x2d32d },
	{ "ARCADE_CIEL_2.HA6", 0x1c2d0,	0x2d330 },
	{ "ARCADE_CIEL_2.TXT", 0x9a,	0x2d369 },
	{ "ARCADE_CIEL_2_C.TXT", 0x1c5f,	0x2d36a },
	{ "ARCADE_CIEL_3.HA6", 0x1100c,	0x2d36e },
	{ "ARCADE_CIEL_3.TXT", 0x9c,	0x2d391 },
	{ "ARCADE_CIEL_3_C.TXT", 0x1a84,	0x2d392 },
	{ "ARCADE_CIEL_9.HA6", 0x2fb40,	0x2d396 },
	{ "ARCADE_CIEL_9.TXT", 0x9c,	0x2d3f6 },
	{ "ARCADE_CIEL_9_C.TXT", 0x19e4,	0x2d3f7 },
	{ "ARCADE_CIEL_C.TXT", 0x1c49,	0x2d3fb },
	{ "ARCADE_CIEL_SELIST.TXT", 0x1020,	0x2d3ff },
	{ "ARCADE_CIEL_WIN.TXT", 0x910,	0x2d402 },
	{ "ARCADE_COMEDIT3.EXE", 0x16000,	0x2d404 },
	{ "ARCADE_CROUCH.CPF", 0x2f36c,	0x2d430 },
	{ "ARCADE_DAMIEN_C.TXT", 0x1484,	0x2d510 },
	{ "ARCADE_DEKU.CPF", 0x2f36c,	0x2d513 },
	{ "ARCADE_EFFECT.ARX", 0x157f40,	0x2d5b4 },
	{ "ARCADE_EFFECT.CG2", 0x2f560,	0x2d864 },
	{ "ARCADE_EFFECT.HA6", 0x1c60c,	0x2d8c3 },
	{ "ARCADE_EFFECT.PAT", 0x128e4,	0x2d8fc },
	{ "ARCADE_EFFECT.TXT", 0xba,	0x2d922 },
	{ "ARCADE_EFFECT2.ARX", 0x153880,	0x2d923 },
	{ "ARCADE_EFFECTPAL2.ARX", 0x13c40,	0x2dbcb },
	{ "ARCADE_F_CIEL.ARX", 0xff640,	0x2dbf3 },
	{ "ARCADE_F_CIEL.CG2", 0x20578,	0x2ddf2 },
	{ "ARCADE_F_CIEL.CPF", 0x2f36c,	0x2de33 },
	{ "ARCADE_F_CIEL_C.TXT", 0x1df,	0x2de92 },
	{ "ARCADE_GAKIHA.CPF", 0x2f36c,	0x2de93 },
	{ "ARCADE_GUARD.CPF", 0x2f36c,	0x2e1da },
	{ "ARCADE_G_AKIHA.CPF", 0x2f36c,	0x2e239 },
	{ "ARCADE_G_AKIHA.HA6", 0x12f8c,	0x2e298 },
	{ "ARCADE_G_AKIHA.PAT", 0x2640f,	0x2e2be },
	{ "ARCADE_G_AKIHA_0.HA6", 0x354c,	0x2e30b },
	{ "ARCADE_G_AKIHA_0.TXT", 0x9a,	0x2e312 },
	{ "ARCADE_G_AKIHA_0_C.TXT", 0x4f2,	0x2e313 },
	{ "ARCADE_G_AKIHA_9.TXT", 0x9a,	0x2e314 },
	{ "ARCADE_G_AKIHA_9_C.TXT", 0x7e6,	0x2e315 },
	{ "ARCADE_G_AKIHA_SELIST.TXT", 0x67c,	0x2e316 },
	{ "ARCADE_G_CHAOS_0.HA6", 0x2eb94,	0x2e317 },
	{ "ARCADE_G_CHAOS_0.TXT", 0x81,	0x2e375 },
	{ "ARCADE_G_CHAOS_0_C.TXT", 0xc23,	0x2e376 },
	{ "ARCADE_HERMES.ARX", 0xb2940,	0x2e378 },
	{ "ARCADE_HERMES.CG2", 0x1c618,	0x2e4de },
	{ "ARCADE_HERMES.CPF", 0x2f36c,	0x2e517 },
	{ "ARCADE_HERMES.HA6", 0x15d30,	0x2e576 },
	{ "ARCADE_HERMES.PAT", 0x5e269,	0x2e5a2 },
	{ "ARCADE_HERMES0.PAL", 0x400,	0x2e65f },
	{ "ARCADE_HERMES1.PAL", 0x400,	0x2e660 },
	{ "ARCADE_HERMES10.PAL", 0x400,	0x2e661 },
	{ "ARCADE_HERMES11.PAL", 0x400,	0x2e662 },
	{ "ARCADE_HERMES12.PAL", 0x400,	0x2e663 },
	{ "ARCADE_HERMES13.PAL", 0x400,	0x2e664 },
	{ "ARCADE_HERMES14.PAL", 0x400,	0x2e665 },
	{ "ARCADE_HERMES15.PAL", 0x400,	0x2e666 },
	{ "ARCADE_HERMES16.PAL", 0x400,	0x2e667 },
	{ "ARCADE_HERMES17.PAL", 0x400,	0x2e668 },
	{ "ARCADE_HERMES18.PAL", 0x400,	0x2e669 },
	{ "ARCADE_HERMES19.PAL", 0x400,	0x2e66a },
	{ "ARCADE_HERMES2.ARX", 0x146bc0,	0x2e66b },
	{ "ARCADE_HERMES2.PAL", 0x400,	0x2e8f9 },
	{ "ARCADE_HERMES20.PAL", 0x400,	0x2e8fa },
	{ "ARCADE_HERMES21.PAL", 0x400,	0x2e8fb },
	{ "ARCADE_HERMES22.PAL", 0x400,	0x2e8fc },
	{ "ARCADE_HERMES23.PAL", 0x400,	0x2e8fd },
	{ "ARCADE_HERMES24.PAL", 0x400,	0x2e8fe },
	{ "ARCADE_HERMES25.PAL", 0x400,	0x2e8ff },
	{ "ARCADE_HERMES26.PAL", 0x400,	0x2e900 },
	{ "ARCADE_HERMES27.PAL", 0x400,	0x2e901 },
	{ "ARCADE_HERMES28.PAL", 0x400,	0x2e902 },
	{ "ARCADE_HERMES29.PAL", 0x400,	0x2e903 },
	{ "ARCADE_HERMES3.PAL", 0x400,	0x2e904 },
	{ "ARCADE_HERMES30.PAL", 0x400,	0x2e905 },
	{ "ARCADE_HERMES31.PAL", 0x400,	0x2e906 },
	{ "ARCADE_HERMES32.PAL", 0x400,	0x2e907 },
	{ "ARCADE_HERMES33.PAL", 0x400,	0x2e908 },
	{ "ARCADE_HERMES34.PAL", 0x400,	0x2e909 },
	{ "ARCADE_HERMES35.PAL", 0x400,	0x2e90a },
	{ "ARCADE_HERMES4.PAL", 0x400,	0x2e90b },
	{ "ARCADE_HERMES5.PAL", 0x400,	0x2e90c },
	{ "ARCADE_HERMES6.PAL", 0x400,	0x2e90d },
	{ "ARCADE_HERMES7.PAL", 0x400,	0x2e90e },
	{ "ARCADE_HERMES8.PAL", 0x400,	0x2e90f },
	{ "ARCADE_HERMES9.PAL", 0x400,	0x2e910 },
	{ "ARCADE_HERMESPAL2.ARX", 0x133c0,	0x2e911 },
	{ "ARCADE_HERMES_0.TXT", 0xa6,	0x2e938 },
	{ "ARCADE_HERMES_0_C.TXT", 0x5f6,	0x2e939 },
	{ "ARCADE_HERMES_SELIST.TXT", 0x6fb,	0x2e93a },
	{ "ARCADE_HISUI.ARX", 0x1da440,	0x2e99e },
	{ "ARCADE_HISUI.CG2", 0x52b70,	0x2ed53 },
	{ "ARCADE_HISUI.CPF", 0x2f36c,	0x2edf9 },
	{ "ARCADE_HISUI.DOF", 0xcde08,	0x2ee58 },
	{ "ARCADE_HISUI.HA6", 0x9c4d8,	0x2eff4 },
	{ "ARCADE_HISUI.PAT", 0xd36a,	0x2f12d },
	{ "ARCADE_HISUI.TXT", 0xdc,	0x2f148 },
	{ "ARCADE_HISUI0.PAL", 0x400,	0x2f149 },
	{ "ARCADE_HISUI1.PAL", 0x400,	0x2f14a },
	{ "ARCADE_HISUI10.PAL", 0x400,	0x2f14b },
	{ "ARCADE_HISUI11.PAL", 0x400,	0x2f14c },
	{ "ARCADE_HISUI12.PAL", 0x400,	0x2f14d },
	{ "ARCADE_HISUI13.PAL", 0x400,	0x2f14e },
	{ "ARCADE_HISUI14.PAL", 0x400,	0x2f14f },
	{ "ARCADE_HISUI15.PAL", 0x400,	0x2f150 },
	{ "ARCADE_HISUI16.PAL", 0x400,	0x2f151 },
	{ "ARCADE_HISUI17.PAL", 0x400,	0x2f152 },
	{ "ARCADE_HISUI18.PAL", 0x400,	0x2f153 },
	{ "ARCADE_HISUI19.PAL", 0x400,	0x2f154 },
	{ "ARCADE_HISUI2.ARX", 0x1d8f80,	0x2f155 },
	{ "ARCADE_HISUI2.PAL", 0x400,	0x2f507 },
	{ "ARCADE_HISUI20.PAL", 0x400,	0x2f508 },
	{ "ARCADE_HISUI21.PAL", 0x400,	0x2f509 },
	{ "ARCADE_HISUI22.PAL", 0x400,	0x2f50a },
	{ "ARCADE_HISUI23.PAL", 0x400,	0x2f50b },
	{ "ARCADE_HISUI24.PAL", 0x400,	0x2f50c },
	{ "ARCADE_HISUI25.PAL", 0x400,	0x2f50d },
	{ "ARCADE_HISUI26.PAL", 0x400,	0x2f50e },
	{ "ARCADE_HISUI27.PAL", 0x400,	0x2f50f },
	{ "ARCADE_HISUI28.PAL", 0x400,	0x2f510 },
	{ "ARCADE_HISUI29.PAL", 0x400,	0x2f511 },
	{ "ARCADE_HISUI3.PAL", 0x400,	0x2f512 },
	{ "ARCADE_HISUI30.PAL", 0x400,	0x2f513 },
	{ "ARCADE_HISUI31.PAL", 0x400,	0x2f514 },
	{ "ARCADE_HISUI32.PAL", 0x400,	0x2f515 },
	{ "ARCADE_HISUI33.PAL", 0x400,	0x2f516 },
	{ "ARCADE_HISUI34.PAL", 0x400,	0x2f517 },
	{ "ARCADE_HISUI35.PAL", 0x400,	0x2f518 },
	{ "ARCADE_HISUI4.PAL", 0x400,	0x2f519 },
	{ "ARCADE_HISUI5.PAL", 0x400,	0x2f51a },
	{ "ARCADE_HISUI6.PAL", 0x400,	0x2f51b },
	{ "ARCADE_HISUI7.PAL", 0x400,	0x2f51c },
	{ "ARCADE_HISUI8.PAL", 0x400,	0x2f51d },
	{ "ARCADE_HISUI9.PAL", 0x400,	0x2f51e },
	{ "ARCADE_HISUIPAL2.ARX", 0x7080,	0x2f51f },
	{ "ARCADE_HISUI_0.HA6", 0x15ef0,	0x2f52e },
	{ "ARCADE_HISUI_0.TXT", 0xb8,	0x2f55a },
	{ "ARCADE_HISUI_0_C.TXT", 0x1945,	0x2f55b },
	{ "ARCADE_HISUI_1.HA6", 0x2f774,	0x2f55f },
	{ "ARCADE_HISUI_1.TXT", 0xba,	0x2f5be },
	{ "ARCADE_HISUI_1_C.TXT", 0x190d,	0x2f5bf },
	{ "ARCADE_HISUI_2.HA6", 0x1d6a0,	0x2f5c3 },
	{ "ARCADE_HISUI_2.TXT", 0xb2,	0x2f5fe },
	{ "ARCADE_HISUI_2_C.TXT", 0x1890,	0x2f5ff },
	{ "ARCADE_HISUI_3.HA6", 0x1100c,	0x2f603 },
	{ "ARCADE_HISUI_3.TXT", 0xb2,	0x2f626 },
	{ "ARCADE_HISUI_3_C.TXT", 0x14bf,	0x2f627 },
	{ "ARCADE_HISUI_9.HA6", 0x2d18c,	0x2f62a },
	{ "ARCADE_HISUI_9.TXT", 0xb2,	0x2f685 },
	{ "ARCADE_HISUI_9_C.TXT", 0x15f3,	0x2f686 },
	{ "ARCADE_HISUI_C.TXT", 0x1618,	0x2f689 },
	{ "ARCADE_HISUI_SELIST.TXT", 0x14c1,	0x2f68c },
	{ "ARCADE_HISUI_WIN.TXT", 0x6d8,	0x2f68f },
	{ "ARCADE_JUMP.CPF", 0x2f36c,	0x2f690 },
	{ "ARCADE_JUMPG.CPF", 0x2f36c,	0x2f6ef },
	{ "ARCADE_KISHIMA.ARX", 0x39ff40,	0x2f74e },
	{ "ARCADE_KISHIMA.CG2", 0x77da0,	0x2fe8e },
	{ "ARCADE_KISHIMA.CPF", 0x2f36c,	0x2ff7e },
	{ "ARCADE_KISHIMA.DOF", 0x7cbf8,	0x2ffdd },
	{ "ARCADE_KISHIMA.HA6", 0x55f88,	0x300d7 },
	{ "ARCADE_KISHIMA.PAT", 0x0,	0x30183 },
	{ "ARCADE_KISHIMA.TXT", 0xd4,	0x30183 },
	{ "ARCADE_KISHIMA0.PAL", 0x400,	0x30184 },
	{ "ARCADE_KISHIMA1.PAL", 0x400,	0x30185 },
	{ "ARCADE_KISHIMA10.PAL", 0x400,	0x30186 },
	{ "ARCADE_KISHIMA11.PAL", 0x400,	0x30187 },
	{ "ARCADE_KISHIMA12.PAL", 0x400,	0x30188 },
	{ "ARCADE_KISHIMA13.PAL", 0x400,	0x30189 },
	{ "ARCADE_KISHIMA14.PAL", 0x400,	0x3018a },
	{ "ARCADE_KISHIMA15.PAL", 0x400,	0x3018b },
	{ "ARCADE_KISHIMA16.PAL", 0x400,	0x3018c },
	{ "ARCADE_KISHIMA17.PAL", 0x400,	0x3018d },
	{ "ARCADE_KISHIMA18.PAL", 0x400,	0x3018e },
	{ "ARCADE_KISHIMA19.PAL", 0x400,	0x3018f },
	{ "ARCADE_KISHIMA2.ARX", 0x39a000,	0x30190 },
	{ "ARCADE_KISHIMA2.PAL", 0x400,	0x308c4 },
	{ "ARCADE_KISHIMA20.PAL", 0x400,	0x308c5 },
	{ "ARCADE_KISHIMA21.PAL", 0x400,	0x308c6 },
	{ "ARCADE_KISHIMA22.PAL", 0x400,	0x308c7 },
	{ "ARCADE_KISHIMA23.PAL", 0x400,	0x308c8 },
	{ "ARCADE_KISHIMA24.PAL", 0x400,	0x308c9 },
	{ "ARCADE_KISHIMA25.PAL", 0x400,	0x308ca },
	{ "ARCADE_KISHIMA26.PAL", 0x400,	0x308cb },
	{ "ARCADE_KISHIMA27.PAL", 0x400,	0x308cc },
	{ "ARCADE_KISHIMA28.PAL", 0x400,	0x308cd },
	{ "ARCADE_KISHIMA29.PAL", 0x400,	0x308ce },
	{ "ARCADE_KISHIMA3.PAL", 0x400,	0x308cf },
	{ "ARCADE_KISHIMA30.PAL", 0x400,	0x308d0 },
	{ "ARCADE_KISHIMA31.PAL", 0x400,	0x308d1 },
	{ "ARCADE_KISHIMA32.PAL", 0x400,	0x308d2 },
	{ "ARCADE_KISHIMA33.PAL", 0x400,	0x308d3 },
	{ "ARCADE_KISHIMA34.PAL", 0x400,	0x308d4 },
	{ "ARCADE_KISHIMA35.PAL", 0x400,	0x308d5 },
	{ "ARCADE_KISHIMA4.PAL", 0x400,	0x308d6 },
	{ "ARCADE_KISHIMA5.PAL", 0x400,	0x308d7 },
	{ "ARCADE_KISHIMA6.PAL", 0x400,	0x308d8 },
	{ "ARCADE_KISHIMA7.PAL", 0x400,	0x308d9 },
	{ "ARCADE_KISHIMA8.PAL", 0x400,	0x308da },
	{ "ARCADE_KISHIMA9.PAL", 0x400,	0x308db },
	{ "ARCADE_KISHIMAPAL2.ARX", 0x21c80,	0x308dc },
	{ "ARCADE_KISHIMA_0.HA6", 0x10824,	0x30920 },
	{ "ARCADE_KISHIMA_0.TXT", 0xac,	0x30942 },
	{ "ARCADE_KISHIMA_0_C.TXT", 0x1759,	0x30943 },
	{ "ARCADE_KISHIMA_1.HA6", 0x157f0,	0x30946 },
	{ "ARCADE_KISHIMA_1.TXT", 0xac,	0x30971 },
	{ "ARCADE_KISHIMA_1_C.TXT", 0x1675,	0x30972 },
	{ "ARCADE_KISHIMA_2.HA6", 0xe294,	0x30975 },
	{ "ARCADE_KISHIMA_2.TXT", 0xac,	0x30992 },
	{ "ARCADE_KISHIMA_2_C.TXT", 0x12a6,	0x30993 },
	{ "ARCADE_KISHIMA_3.HA6", 0x1100c,	0x30996 },
	{ "ARCADE_KISHIMA_3.TXT", 0xac,	0x309b9 },
	{ "ARCADE_KISHIMA_3_C.TXT", 0x1034,	0x309ba },
	{ "ARCADE_KISHIMA_9.HA6", 0x16f70,	0x309bd },
	{ "ARCADE_KISHIMA_9.TXT", 0xac,	0x309eb },
	{ "ARCADE_KISHIMA_9_C.TXT", 0x109a,	0x309ec },
	{ "ARCADE_KISHIMA_SELIST.TXT", 0x11ec,	0x309ef },
	{ "ARCADE_KISHIMA_WIN.TXT", 0x712,	0x309f2 },
	{ "ARCADE_KOHAKU.ARX", 0x22cec0,	0x309f3 },
	{ "ARCADE_KOHAKU.CG2", 0x62068,	0x30e4d },
	{ "ARCADE_KOHAKU.CPF", 0x2f36c,	0x30f12 },
	{ "ARCADE_KOHAKU.DOF", 0xe0d10,	0x30f71 },
	{ "ARCADE_KOHAKU.HA6", 0x97c5c,	0x31133 },
	{ "ARCADE_KOHAKU.PAT", 0xcb37,	0x31263 },
	{ "ARCADE_KOHAKU.TXT", 0xe1,	0x3127d },
	{ "ARCADE_KOHAKU0.PAL", 0x400,	0x3127e },
	{ "ARCADE_KOHAKU1.PAL", 0x400,	0x3127f },
	{ "ARCADE_KOHAKU10.PAL", 0x400,	0x31280 },
	{ "ARCADE_KOHAKU11.PAL", 0x400,	0x31281 },
	{ "ARCADE_KOHAKU12.PAL", 0x400,	0x31282 },
	{ "ARCADE_KOHAKU13.PAL", 0x400,	0x31283 },
	{ "ARCADE_KOHAKU14.PAL", 0x400,	0x31284 },
	{ "ARCADE_KOHAKU15.PAL", 0x400,	0x31285 },
	{ "ARCADE_KOHAKU16.PAL", 0x400,	0x31286 },
	{ "ARCADE_KOHAKU17.PAL", 0x400,	0x31287 },
	{ "ARCADE_KOHAKU18.PAL", 0x400,	0x31288 },
	{ "ARCADE_KOHAKU19.PAL", 0x400,	0x31289 },
	{ "ARCADE_KOHAKU2.ARX", 0x229f40,	0x3128a },
	{ "ARCADE_KOHAKU2.PAL", 0x400,	0x316de },
	{ "ARCADE_KOHAKU20.PAL", 0x400,	0x316df },
	{ "ARCADE_KOHAKU21.PAL", 0x400,	0x316e0 },
	{ "ARCADE_KOHAKU22.PAL", 0x400,	0x316e1 },
	{ "ARCADE_KOHAKU23.PAL", 0x400,	0x316e2 },
	{ "ARCADE_KOHAKU24.PAL", 0x400,	0x316e3 },
	{ "ARCADE_KOHAKU25.PAL", 0x400,	0x316e4 },
	{ "ARCADE_KOHAKU26.PAL", 0x400,	0x316e5 },
	{ "ARCADE_KOHAKU27.PAL", 0x400,	0x316e6 },
	{ "ARCADE_KOHAKU28.PAL", 0x400,	0x316e7 },
	{ "ARCADE_KOHAKU29.PAL", 0x400,	0x316e8 },
	{ "ARCADE_KOHAKU3.PAL", 0x400,	0x316e9 },
	{ "ARCADE_KOHAKU30.PAL", 0x400,	0x316ea },
	{ "ARCADE_KOHAKU31.PAL", 0x400,	0x316eb },
	{ "ARCADE_KOHAKU32.PAL", 0x400,	0x316ec },
	{ "ARCADE_KOHAKU33.PAL", 0x400,	0x316ed },
	{ "ARCADE_KOHAKU34.PAL", 0x400,	0x316ee },
	{ "ARCADE_KOHAKU35.PAL", 0x400,	0x316ef },
	{ "ARCADE_KOHAKU4.PAL", 0x400,	0x316f0 },
	{ "ARCADE_KOHAKU5.PAL", 0x400,	0x316f1 },
	{ "ARCADE_KOHAKU6.PAL", 0x400,	0x316f2 },
	{ "ARCADE_KOHAKU7.PAL", 0x400,	0x316f3 },
	{ "ARCADE_KOHAKU8.PAL", 0x400,	0x316f4 },
	{ "ARCADE_KOHAKU9.PAL", 0x400,	0x316f5 },
	{ "ARCADE_KOHAKUPAL2.ARX", 0x8540,	0x316f6 },
	{ "ARCADE_KOHAKU_0.HA6", 0x1f91c,	0x31707 },
	{ "ARCADE_KOHAKU_0.TXT", 0xb5,	0x31747 },
	{ "ARCADE_KOHAKU_0_C.TXT", 0x1495,	0x31748 },
	{ "ARCADE_KOHAKU_1.HA6", 0x3ca3c,	0x3174b },
	{ "ARCADE_KOHAKU_1.TXT", 0xb7,	0x317c5 },
	{ "ARCADE_KOHAKU_1_C.TXT", 0x1d14,	0x317c6 },
	{ "ARCADE_KOHAKU_2.HA6", 0x24d40,	0x317ca },
	{ "ARCADE_KOHAKU_2.TXT", 0xb7,	0x31814 },
	{ "ARCADE_KOHAKU_2_C.TXT", 0x162d,	0x31815 },
	{ "ARCADE_KOHAKU_3.HA6", 0x1100c,	0x31818 },
	{ "ARCADE_KOHAKU_3.TXT", 0xb7,	0x3183b },
	{ "ARCADE_KOHAKU_3_C.TXT", 0x1165,	0x3183c },
	{ "ARCADE_KOHAKU_9.HA6", 0x30430,	0x3183f },
	{ "ARCADE_KOHAKU_9.TXT", 0xb7,	0x318a0 },
	{ "ARCADE_KOHAKU_9_C.TXT", 0x1209,	0x318a1 },
	{ "ARCADE_KOHAKU_C.TXT", 0x12f8,	0x318a4 },
	{ "ARCADE_KOHAKU_M.ARX", 0x1cb980,	0x318a7 },
	{ "ARCADE_KOHAKU_M.CG2", 0x50ad0,	0x31c3f },
	{ "ARCADE_KOHAKU_M.CPF", 0x2f36c,	0x31ce1 },
	{ "ARCADE_KOHAKU_M.HA6", 0x6180c,	0x31d40 },
	{ "ARCADE_KOHAKU_M.TXT", 0xa6,	0x31e04 },
	{ "ARCADE_KOHAKU_M0.PAL", 0x400,	0x31e05 },
	{ "ARCADE_KOHAKU_M1.PAL", 0x400,	0x31e06 },
	{ "ARCADE_KOHAKU_M10.PAL", 0x400,	0x31e07 },
	{ "ARCADE_KOHAKU_M11.PAL", 0x400,	0x31e08 },
	{ "ARCADE_KOHAKU_M12.PAL", 0x400,	0x31e09 },
	{ "ARCADE_KOHAKU_M13.PAL", 0x400,	0x31e0a },
	{ "ARCADE_KOHAKU_M14.PAL", 0x400,	0x31e0b },
	{ "ARCADE_KOHAKU_M15.PAL", 0x400,	0x31e0c },
	{ "ARCADE_KOHAKU_M16.PAL", 0x400,	0x31e0d },
	{ "ARCADE_KOHAKU_M17.PAL", 0x400,	0x31e0e },
	{ "ARCADE_KOHAKU_M18.PAL", 0x400,	0x31e0f },
	{ "ARCADE_KOHAKU_M19.PAL", 0x400,	0x31e10 },
	{ "ARCADE_KOHAKU_M2.ARX", 0x1c91c0,	0x31e11 },
	{ "ARCADE_KOHAKU_M2.PAL", 0x400,	0x321a4 },
	{ "ARCADE_KOHAKU_M20.PAL", 0x400,	0x321a5 },
	{ "ARCADE_KOHAKU_M21.PAL", 0x400,	0x321a6 },
	{ "ARCADE_KOHAKU_M22.PAL", 0x400,	0x321a7 },
	{ "ARCADE_KOHAKU_M23.PAL", 0x400,	0x321a8 },
	{ "ARCADE_KOHAKU_M24.PAL", 0x400,	0x321a9 },
	{ "ARCADE_KOHAKU_M25.PAL", 0x400,	0x321aa },
	{ "ARCADE_KOHAKU_M26.PAL", 0x400,	0x321ab },
	{ "ARCADE_KOHAKU_M27.PAL", 0x400,	0x321ac },
	{ "ARCADE_KOHAKU_M28.PAL", 0x400,	0x321ad },
	{ "ARCADE_KOHAKU_M29.PAL", 0x400,	0x321ae },
	{ "ARCADE_KOHAKU_M3.PAL", 0x400,	0x321af },
	{ "ARCADE_KOHAKU_M30.PAL", 0x400,	0x321b0 },
	{ "ARCADE_KOHAKU_M31.PAL", 0x400,	0x321b1 },
	{ "ARCADE_KOHAKU_M32.PAL", 0x400,	0x321b2 },
	{ "ARCADE_KOHAKU_M33.PAL", 0x400,	0x321b3 },
	{ "ARCADE_KOHAKU_M34.PAL", 0x400,	0x321b4 },
	{ "ARCADE_KOHAKU_M35.PAL", 0x400,	0x321b5 },
	{ "ARCADE_KOHAKU_M4.PAL", 0x400,	0x321b6 },
	{ "ARCADE_KOHAKU_M5.PAL", 0x400,	0x321b7 },
	{ "ARCADE_KOHAKU_M6.PAL", 0x400,	0x321b8 },
	{ "ARCADE_KOHAKU_M7.PAL", 0x400,	0x321b9 },
	{ "ARCADE_KOHAKU_M8.PAL", 0x400,	0x321ba },
	{ "ARCADE_KOHAKU_M9.PAL", 0x400,	0x321bb },
	{ "ARCADE_KOHAKU_MPAL2.ARX", 0x5fc0,	0x321bc },
	{ "ARCADE_KOHAKU_M_0.HA6", 0x49a8,	0x321c8 },
	{ "ARCADE_KOHAKU_M_0.TXT", 0xbf,	0x321d2 },
	{ "ARCADE_KOHAKU_M_0_C.TXT", 0x1529,	0x321d3 },
	{ "ARCADE_KOHAKU_M_1.HA6", 0x12984,	0x321d6 },
	{ "ARCADE_KOHAKU_M_1.TXT", 0xbd,	0x321fc },
	{ "ARCADE_KOHAKU_M_1_C.TXT", 0x1487,	0x321fd },
	{ "ARCADE_KOHAKU_M_2.HA6", 0x155cc,	0x32200 },
	{ "ARCADE_KOHAKU_M_2.TXT", 0xbd,	0x3222b },
	{ "ARCADE_KOHAKU_M_2_C.TXT", 0x1006,	0x3222c },
	{ "ARCADE_KOHAKU_M_9_C.TXT", 0x1280,	0x3222f },
	{ "ARCADE_KOHAKU_M_C.TXT", 0x1600,	0x32232 },
	{ "ARCADE_KOHAKU_M_SELIST.TXT", 0xff2,	0x32235 },
	{ "ARCADE_KOHAKU_M_WIN.TXT", 0x75c,	0x32237 },
	{ "ARCADE_KOHAKU_SELIST.TXT", 0x1347,	0x32238 },
	{ "ARCADE_KOHAKU_WIN.TXT", 0x75c,	0x3223b },
	{ "ARCADE_LEN.ARX", 0x203000,	0x3227e },
	{ "ARCADE_LEN.CG2", 0x5a5b0,	0x32684 },
	{ "ARCADE_LEN.CPF", 0x2f36c,	0x32739 },
	{ "ARCADE_LEN.DOF", 0x8ba80,	0x32798 },
	{ "ARCADE_LEN.HA6", 0x5204c,	0x328b0 },
	{ "ARCADE_LEN.PAT", 0xc150,	0x32955 },
	{ "ARCADE_LEN.TXT", 0xcc,	0x3296e },
	{ "ARCADE_LEN0.PAL", 0x400,	0x3296f },
	{ "ARCADE_LEN1.PAL", 0x400,	0x32970 },
	{ "ARCADE_LEN10.PAL", 0x400,	0x32971 },
	{ "ARCADE_LEN11.PAL", 0x400,	0x32972 },
	{ "ARCADE_LEN12.PAL", 0x400,	0x32973 },
	{ "ARCADE_LEN13.PAL", 0x400,	0x32974 },
	{ "ARCADE_LEN14.PAL", 0x400,	0x32975 },
	{ "ARCADE_LEN15.PAL", 0x400,	0x32976 },
	{ "ARCADE_LEN16.PAL", 0x400,	0x32977 },
	{ "ARCADE_LEN17.PAL", 0x400,	0x32978 },
	{ "ARCADE_LEN18.PAL", 0x400,	0x32979 },
	{ "ARCADE_LEN19.PAL", 0x400,	0x3297a },
	{ "ARCADE_LEN2.ARX", 0x201000,	0x3297b },
	{ "ARCADE_LEN2.PAL", 0x400,	0x32d7d },
	{ "ARCADE_LEN20.PAL", 0x400,	0x32d7e },
	{ "ARCADE_LEN21.PAL", 0x400,	0x32d7f },
	{ "ARCADE_LEN22.PAL", 0x400,	0x32d80 },
	{ "ARCADE_LEN23.PAL", 0x400,	0x32d81 },
	{ "ARCADE_LEN24.PAL", 0x400,	0x32d82 },
	{ "ARCADE_LEN25.PAL", 0x400,	0x32d83 },
	{ "ARCADE_LEN26.PAL", 0x400,	0x32d84 },
	{ "ARCADE_LEN27.PAL", 0x400,	0x32d85 },
	{ "ARCADE_LEN28.PAL", 0x400,	0x32d86 },
	{ "ARCADE_LEN29.PAL", 0x400,	0x32d87 },
	{ "ARCADE_LEN3.PAL", 0x400,	0x32d88 },
	{ "ARCADE_LEN30.PAL", 0x400,	0x32d89 },
	{ "ARCADE_LEN31.PAL", 0x400,	0x32d8a },
	{ "ARCADE_LEN32.PAL", 0x400,	0x32d8b },
	{ "ARCADE_LEN33.PAL", 0x400,	0x32d8c },
	{ "ARCADE_LEN34.PAL", 0x400,	0x32d8d },
	{ "ARCADE_LEN35.PAL", 0x400,	0x32d8e },
	{ "ARCADE_LEN4.PAL", 0x400,	0x32d8f },
	{ "ARCADE_LEN5.PAL", 0x400,	0x32d90 },
	{ "ARCADE_LEN6.PAL", 0x400,	0x32d91 },
	{ "ARCADE_LEN7.PAL", 0x400,	0x32d92 },
	{ "ARCADE_LEN8.PAL", 0x400,	0x32d93 },
	{ "ARCADE_LEN9.PAL", 0x400,	0x32d94 },
	{ "ARCADE_LENPAL2.ARX", 0xf580,	0x32d95 },
	{ "ARCADE_LEN_0.HA6", 0x5468,	0x32db4 },
	{ "ARCADE_LEN_0.TXT", 0xa8,	0x32dbf },
	{ "ARCADE_LEN_0_C.TXT", 0x10cb,	0x32dc0 },
	{ "ARCADE_LEN_1.HA6", 0x173b4,	0x32dc3 },
	{ "ARCADE_LEN_1.TXT", 0xa8,	0x32df2 },
	{ "ARCADE_LEN_1_C.TXT", 0x124d,	0x32df3 },
	{ "ARCADE_LEN_2.HA6", 0xa288,	0x32df6 },
	{ "ARCADE_LEN_2.TXT", 0xa8,	0x32e0b },
	{ "ARCADE_LEN_2_C.TXT", 0x1114,	0x32e0c },
	{ "ARCADE_LEN_3.HA6", 0x1100c,	0x32e0f },
	{ "ARCADE_LEN_3.TXT", 0xa8,	0x32e32 },
	{ "ARCADE_LEN_3_C.TXT", 0xf34,	0x32e33 },
	{ "ARCADE_LEN_9.HA6", 0x15298,	0x32e35 },
	{ "ARCADE_LEN_9.TXT", 0xa8,	0x32e60 },
	{ "ARCADE_LEN_9_C.TXT", 0xfd0,	0x32e61 },
	{ "ARCADE_LEN_C.TXT", 0x142f,	0x32e63 },
	{ "ARCADE_LEN_SELIST.TXT", 0x380,	0x32e66 },
	{ "ARCADE_MIYAKO.ARX", 0x176200,	0x32e88 },
	{ "ARCADE_MIYAKO.CG2", 0x3d4e0,	0x33175 },
	{ "ARCADE_MIYAKO.CPF", 0x2f36c,	0x331f0 },
	{ "ARCADE_MIYAKO.DOF", 0x636a4,	0x3324f },
	{ "ARCADE_MIYAKO.HA6", 0x4bee0,	0x33316 },
	{ "ARCADE_MIYAKO.PAT", 0xbe94,	0x333ae },
	{ "ARCADE_MIYAKO.TXT", 0xe1,	0x333c6 },
	{ "ARCADE_MIYAKO0.PAL", 0x400,	0x333c7 },
	{ "ARCADE_MIYAKO1.PAL", 0x400,	0x333c8 },
	{ "ARCADE_MIYAKO10.PAL", 0x400,	0x333c9 },
	{ "ARCADE_MIYAKO11.PAL", 0x400,	0x333ca },
	{ "ARCADE_MIYAKO12.PAL", 0x400,	0x333cb },
	{ "ARCADE_MIYAKO13.PAL", 0x400,	0x333cc },
	{ "ARCADE_MIYAKO14.PAL", 0x400,	0x333cd },
	{ "ARCADE_MIYAKO15.PAL", 0x400,	0x333ce },
	{ "ARCADE_MIYAKO16.PAL", 0x400,	0x333cf },
	{ "ARCADE_MIYAKO17.PAL", 0x400,	0x333d0 },
	{ "ARCADE_MIYAKO18.PAL", 0x400,	0x333d1 },
	{ "ARCADE_MIYAKO19.PAL", 0x400,	0x333d2 },
	{ "ARCADE_MIYAKO2.ARX", 0x177d80,	0x333d3 },
	{ "ARCADE_MIYAKO2.PAL", 0x400,	0x336c3 },
	{ "ARCADE_MIYAKO20.PAL", 0x400,	0x336c4 },
	{ "ARCADE_MIYAKO21.PAL", 0x400,	0x336c5 },
	{ "ARCADE_MIYAKO22.PAL", 0x400,	0x336c6 },
	{ "ARCADE_MIYAKO23.PAL", 0x400,	0x336c7 },
	{ "ARCADE_MIYAKO24.PAL", 0x400,	0x336c8 },
	{ "ARCADE_MIYAKO25.PAL", 0x400,	0x336c9 },
	{ "ARCADE_MIYAKO26.PAL", 0x400,	0x336ca },
	{ "ARCADE_MIYAKO27.PAL", 0x400,	0x336cb },
	{ "ARCADE_MIYAKO28.PAL", 0x400,	0x336cc },
	{ "ARCADE_MIYAKO29.PAL", 0x400,	0x336cd },
	{ "ARCADE_MIYAKO3.PAL", 0x400,	0x336ce },
	{ "ARCADE_MIYAKO30.PAL", 0x400,	0x336cf },
	{ "ARCADE_MIYAKO31.PAL", 0x400,	0x336d0 },
	{ "ARCADE_MIYAKO32.PAL", 0x400,	0x336d1 },
	{ "ARCADE_MIYAKO33.PAL", 0x400,	0x336d2 },
	{ "ARCADE_MIYAKO34.PAL", 0x400,	0x336d3 },
	{ "ARCADE_MIYAKO35.PAL", 0x400,	0x336d4 },
	{ "ARCADE_MIYAKO4.PAL", 0x400,	0x336d5 },
	{ "ARCADE_MIYAKO5.PAL", 0x400,	0x336d6 },
	{ "ARCADE_MIYAKO6.PAL", 0x400,	0x336d7 },
	{ "ARCADE_MIYAKO7.PAL", 0x400,	0x336d8 },
	{ "ARCADE_MIYAKO8.PAL", 0x400,	0x336d9 },
	{ "ARCADE_MIYAKO9.PAL", 0x400,	0x336da },
	{ "ARCADE_MIYAKOPAL2.ARX", 0x5bc0,	0x336db },
	{ "ARCADE_MIYAKO_0.HA6", 0x723c,	0x336e7 },
	{ "ARCADE_MIYAKO_0.TXT", 0xb7,	0x336f6 },
	{ "ARCADE_MIYAKO_0_C.TXT", 0x14fe,	0x336f7 },
	{ "ARCADE_MIYAKO_1.HA6", 0x12aac,	0x336fa },
	{ "ARCADE_MIYAKO_1.TXT", 0xb7,	0x33720 },
	{ "ARCADE_MIYAKO_1_C.TXT", 0x16c1,	0x33721 },
	{ "ARCADE_MIYAKO_2.HA6", 0xcdfc,	0x33724 },
	{ "ARCADE_MIYAKO_2.TXT", 0xb7,	0x3373e },
	{ "ARCADE_MIYAKO_2_C.TXT", 0x1346,	0x3373f },
	{ "ARCADE_MIYAKO_3.HA6", 0x1100c,	0x33742 },
	{ "ARCADE_MIYAKO_3.TXT", 0xb7,	0x33765 },
	{ "ARCADE_MIYAKO_3_C.TXT", 0x10d9,	0x33766 },
	{ "ARCADE_MIYAKO_9.HA6", 0x193e4,	0x33769 },
	{ "ARCADE_MIYAKO_9.TXT", 0xb7,	0x3379c },
	{ "ARCADE_MIYAKO_9_C.TXT", 0xf7b,	0x3379d },
	{ "ARCADE_MIYAKO_SELIST.TXT", 0xd16,	0x3379f },
	{ "ARCADE_MIYAKO_WIN.TXT", 0x776,	0x337a1 },
	{ "ARCADE_M_HISUI.ARX", 0x2de5c0,	0x337a2 },
	{ "ARCADE_M_HISUI.CG2", 0x7d1e8,	0x33d5f },
	{ "ARCADE_M_HISUI.CPF", 0x2f36c,	0x33e5a },
	{ "ARCADE_M_HISUI.DOF", 0xb74e8,	0x33eb9 },
	{ "ARCADE_M_HISUI.HA6", 0xa852c,	0x34028 },
	{ "ARCADE_M_HISUI.PAT", 0x12030,	0x34179 },
	{ "ARCADE_M_HISUI.TXT", 0xe8,	0x3419e },
	{ "ARCADE_M_HISUI0.PAL", 0x400,	0x3419f },
	{ "ARCADE_M_HISUI1.PAL", 0x400,	0x341c1 },
	{ "ARCADE_M_HISUI10.PAL", 0x400,	0x341c2 },
	{ "ARCADE_M_HISUI11.PAL", 0x400,	0x341c3 },
	{ "ARCADE_M_HISUI12.PAL", 0x400,	0x341c4 },
	{ "ARCADE_M_HISUI13.PAL", 0x400,	0x341c5 },
	{ "ARCADE_M_HISUI14.PAL", 0x400,	0x341c6 },
	{ "ARCADE_M_HISUI15.PAL", 0x400,	0x341c7 },
	{ "ARCADE_M_HISUI16.PAL", 0x400,	0x341c8 },
	{ "ARCADE_M_HISUI17.PAL", 0x400,	0x341c9 },
	{ "ARCADE_M_HISUI18.PAL", 0x400,	0x341ca },
	{ "ARCADE_M_HISUI19.PAL", 0x400,	0x341cb },
	{ "ARCADE_M_HISUI2.ARX", 0x2d9bc0,	0x341cc },
	{ "ARCADE_M_HISUI2.PAL", 0x400,	0x34780 },
	{ "ARCADE_M_HISUI20.PAL", 0x400,	0x34781 },
	{ "ARCADE_M_HISUI21.PAL", 0x400,	0x34782 },
	{ "ARCADE_M_HISUI22.PAL", 0x400,	0x34783 },
	{ "ARCADE_M_HISUI23.PAL", 0x400,	0x34784 },
	{ "ARCADE_M_HISUI24.PAL", 0x400,	0x34785 },
	{ "ARCADE_M_HISUI25.PAL", 0x400,	0x34786 },
	{ "ARCADE_M_HISUI26.PAL", 0x400,	0x34787 },
	{ "ARCADE_M_HISUI27.PAL", 0x400,	0x34788 },
	{ "ARCADE_M_HISUI28.PAL", 0x400,	0x34789 },
	{ "ARCADE_M_HISUI29.PAL", 0x400,	0x3478a },
	{ "ARCADE_M_HISUI3.PAL", 0x400,	0x3478b },
	{ "ARCADE_M_HISUI30.PAL", 0x400,	0x3478c },
	{ "ARCADE_M_HISUI31.PAL", 0x400,	0x3478d },
	{ "ARCADE_M_HISUI32.PAL", 0x400,	0x3478e },
	{ "ARCADE_M_HISUI33.PAL", 0x400,	0x3478f },
	{ "ARCADE_M_HISUI34.PAL", 0x400,	0x34790 },
	{ "ARCADE_M_HISUI35.PAL", 0x400,	0x34791 },
	{ "ARCADE_M_HISUI4.PAL", 0x400,	0x34792 },
	{ "ARCADE_M_HISUI5.PAL", 0x400,	0x34793 },
	{ "ARCADE_M_HISUI6.PAL", 0x400,	0x34794 },
	{ "ARCADE_M_HISUI7.PAL", 0x400,	0x34795 },
	{ "ARCADE_M_HISUI8.PAL", 0x400,	0x34796 },
	{ "ARCADE_M_HISUI9.PAL", 0x400,	0x34797 },
	{ "ARCADE_M_HISUIPAL2.ARX", 0x165c0,	0x34798 },
	{ "ARCADE_M_HISUI_0.HA6", 0x34f4,	0x347c5 },
	{ "ARCADE_M_HISUI_0.TXT", 0xbe,	0x347cc },
	{ "ARCADE_M_HISUI_0_C.TXT", 0x1a46,	0x347cd },
	{ "ARCADE_M_HISUI_1.HA6", 0x6338,	0x347d1 },
	{ "ARCADE_M_HISUI_1.TXT", 0xbc,	0x347de },
	{ "ARCADE_M_HISUI_1_C.TXT", 0x1087,	0x347df },
	{ "ARCADE_M_HISUI_2.HA6", 0xe15c,	0x347e2 },
	{ "ARCADE_M_HISUI_2.TXT", 0xbc,	0x347ff },
	{ "ARCADE_M_HISUI_2_C.TXT", 0x13a6,	0x34800 },
	{ "ARCADE_M_HISUI_3.HA6", 0x1100c,	0x34803 },
	{ "ARCADE_M_HISUI_9.HA6", 0x17120,	0x34826 },
	{ "ARCADE_M_HISUI_9.TXT", 0xbc,	0x34855 },
	{ "ARCADE_M_HISUI_9_C.TXT", 0x1232,	0x34856 },
	{ "ARCADE_M_HISUI_M.ARX", 0x1d3fc0,	0x34859 },
	{ "ARCADE_M_HISUI_M.CG2", 0x53c68,	0x34c01 },
	{ "ARCADE_M_HISUI_M.CPF", 0x2f36c,	0x34ca9 },
	{ "ARCADE_M_HISUI_M.HA6", 0x4a0ac,	0x34d08 },
	{ "ARCADE_M_HISUI_M.TXT", 0xaa,	0x34d9d },
	{ "ARCADE_M_HISUI_M0.PAL", 0x400,	0x34d9e },
	{ "ARCADE_M_HISUI_M1.PAL", 0x400,	0x34d9f },
	{ "ARCADE_M_HISUI_M10.PAL", 0x400,	0x34da0 },
	{ "ARCADE_M_HISUI_M11.PAL", 0x400,	0x34da1 },
	{ "ARCADE_M_HISUI_M12.PAL", 0x400,	0x34da2 },
	{ "ARCADE_M_HISUI_M13.PAL", 0x400,	0x34da3 },
	{ "ARCADE_M_HISUI_M14.PAL", 0x400,	0x34da4 },
	{ "ARCADE_M_HISUI_M15.PAL", 0x400,	0x34da5 },
	{ "ARCADE_M_HISUI_M16.PAL", 0x400,	0x34da6 },
	{ "ARCADE_M_HISUI_M17.PAL", 0x400,	0x34da7 },
	{ "ARCADE_M_HISUI_M18.PAL", 0x400,	0x34da8 },
	{ "ARCADE_M_HISUI_M19.PAL", 0x400,	0x34da9 },
	{ "ARCADE_M_HISUI_M2.ARX", 0x1d1000,	0x34daa },
	{ "ARCADE_M_HISUI_M2.PAL", 0x400,	0x3514c },
	{ "ARCADE_M_HISUI_M20.PAL", 0x400,	0x3514d },
	{ "ARCADE_M_HISUI_M21.PAL", 0x400,	0x3514e },
	{ "ARCADE_M_HISUI_M22.PAL", 0x400,	0x3514f },
	{ "ARCADE_M_HISUI_M23.PAL", 0x400,	0x35150 },
	{ "ARCADE_M_HISUI_M24.PAL", 0x400,	0x35151 },
	{ "ARCADE_M_HISUI_M25.PAL", 0x400,	0x35152 },
	{ "ARCADE_M_HISUI_M26.PAL", 0x400,	0x35153 },
	{ "ARCADE_M_HISUI_M27.PAL", 0x400,	0x35154 },
	{ "ARCADE_M_HISUI_M28.PAL", 0x400,	0x35155 },
	{ "ARCADE_M_HISUI_M29.PAL", 0x400,	0x35156 },
	{ "ARCADE_M_HISUI_M3.PAL", 0x400,	0x35157 },
	{ "ARCADE_M_HISUI_M30.PAL", 0x400,	0x35158 },
	{ "ARCADE_M_HISUI_M31.PAL", 0x400,	0x35159 },
	{ "ARCADE_M_HISUI_M32.PAL", 0x400,	0x3515a },
	{ "ARCADE_M_HISUI_M33.PAL", 0x400,	0x3515b },
	{ "ARCADE_M_HISUI_M34.PAL", 0x400,	0x3515c },
	{ "ARCADE_M_HISUI_M35.PAL", 0x400,	0x3515d },
	{ "ARCADE_M_HISUI_M4.PAL", 0x400,	0x3515e },
	{ "ARCADE_M_HISUI_M5.PAL", 0x400,	0x3515f },
	{ "ARCADE_M_HISUI_M6.PAL", 0x400,	0x35160 },
	{ "ARCADE_M_HISUI_M7.PAL", 0x400,	0x35161 },
	{ "ARCADE_M_HISUI_M8.PAL", 0x400,	0x35162 },
	{ "ARCADE_M_HISUI_M9.PAL", 0x400,	0x35163 },
	{ "ARCADE_M_HISUI_MPAL2.ARX", 0xc380,	0x35164 },
	{ "ARCADE_M_HISUI_M_0.HA6", 0x4a84,	0x3517d },
	{ "ARCADE_M_HISUI_M_0.TXT", 0xc4,	0x35187 },
	{ "ARCADE_M_HISUI_M_0_C.TXT", 0xdf3,	0x35188 },
	{ "ARCADE_M_HISUI_M_1.HA6", 0xca14,	0x3518a },
	{ "ARCADE_M_HISUI_M_1.TXT", 0xc4,	0x351a4 },
	{ "ARCADE_M_HISUI_M_1_C.TXT", 0xd00,	0x351a5 },
	{ "ARCADE_M_HISUI_M_2.HA6", 0x115e4,	0x351a7 },
	{ "ARCADE_M_HISUI_M_2.TXT", 0xc4,	0x351ca },
	{ "ARCADE_M_HISUI_M_2_C.TXT", 0xddd,	0x351cb },
	{ "ARCADE_M_HISUI_M_9.HA6", 0x354c,	0x351cd },
	{ "ARCADE_M_HISUI_M_SELIST.TXT", 0x998,	0x351d4 },
	{ "ARCADE_M_HISUI_P.ARX", 0x22f8c0,	0x351d6 },
	{ "ARCADE_M_HISUI_P.CG2", 0x5c638,	0x35636 },
	{ "ARCADE_M_HISUI_P.CPF", 0x2f36c,	0x356ef },
	{ "ARCADE_M_HISUI_P.HA6", 0x73288,	0x3574e },
	{ "ARCADE_M_HISUI_P.TXT", 0xaa,	0x35835 },
	{ "ARCADE_M_HISUI_P0.PAL", 0x400,	0x35836 },
	{ "ARCADE_M_HISUI_P1.PAL", 0x400,	0x35837 },
	{ "ARCADE_M_HISUI_P10.PAL", 0x400,	0x35838 },
	{ "ARCADE_M_HISUI_P11.PAL", 0x400,	0x35839 },
	{ "ARCADE_M_HISUI_P12.PAL", 0x400,	0x3583a },
	{ "ARCADE_M_HISUI_P13.PAL", 0x400,	0x3583b },
	{ "ARCADE_M_HISUI_P14.PAL", 0x400,	0x3583c },
	{ "ARCADE_M_HISUI_P15.PAL", 0x400,	0x3583d },
	{ "ARCADE_M_HISUI_P16.PAL", 0x400,	0x3583e },
	{ "ARCADE_M_HISUI_P17.PAL", 0x400,	0x3583f },
	{ "ARCADE_M_HISUI_P18.PAL", 0x400,	0x35840 },
	{ "ARCADE_M_HISUI_P19.PAL", 0x400,	0x35841 },
	{ "ARCADE_M_HISUI_P2.ARX", 0x22c200,	0x35842 },
	{ "ARCADE_M_HISUI_P2.PAL", 0x400,	0x35c9b },
	{ "ARCADE_M_HISUI_P20.PAL", 0x400,	0x35c9c },
	{ "ARCADE_M_HISUI_P21.PAL", 0x400,	0x35c9d },
	{ "ARCADE_M_HISUI_P22.PAL", 0x400,	0x35c9e },
	{ "ARCADE_M_HISUI_P23.PAL", 0x400,	0x35c9f },
	{ "ARCADE_M_HISUI_P24.PAL", 0x400,	0x35ca0 },
	{ "ARCADE_M_HISUI_P25.PAL", 0x400,	0x35ca1 },
	{ "ARCADE_M_HISUI_P26.PAL", 0x400,	0x35ca2 },
	{ "ARCADE_M_HISUI_P27.PAL", 0x400,	0x35ca3 },
	{ "ARCADE_M_HISUI_P28.PAL", 0x400,	0x35ca4 },
	{ "ARCADE_M_HISUI_P29.PAL", 0x400,	0x35ca5 },
	{ "ARCADE_M_HISUI_P3.PAL", 0x400,	0x35ca6 },
	{ "ARCADE_M_HISUI_P30.PAL", 0x400,	0x35ca7 },
	{ "ARCADE_M_HISUI_P31.PAL", 0x400,	0x35ca8 },
	{ "ARCADE_M_HISUI_P32.PAL", 0x400,	0x35ca9 },
	{ "ARCADE_M_HISUI_P33.PAL", 0x400,	0x35caa },
	{ "ARCADE_M_HISUI_P34.PAL", 0x400,	0x35cab },
	{ "ARCADE_M_HISUI_P35.PAL", 0x400,	0x35cac },
	{ "ARCADE_M_HISUI_P4.PAL", 0x400,	0x35cad },
	{ "ARCADE_M_HISUI_P5.PAL", 0x400,	0x35cae },
	{ "ARCADE_M_HISUI_P6.PAL", 0x400,	0x35caf },
	{ "ARCADE_M_HISUI_P7.PAL", 0x400,	0x35cb0 },
	{ "ARCADE_M_HISUI_P8.PAL", 0x400,	0x35cb1 },
	{ "ARCADE_M_HISUI_P9.PAL", 0x400,	0x35cb2 },
	{ "ARCADE_M_HISUI_PPAL2.ARX", 0x11f00,	0x35cb3 },
	{ "ARCADE_M_HISUI_P_0.HA6", 0xd6f4,	0x35cd7 },
	{ "ARCADE_M_HISUI_P_0.TXT", 0xc8,	0x35cf2 },
	{ "ARCADE_M_HISUI_P_0_C.TXT", 0x31c,	0x35cf3 },
	{ "ARCADE_M_HISUI_P_1.HA6", 0xca2c,	0x35cf4 },
	{ "ARCADE_M_HISUI_P_1.TXT", 0xc4,	0x35d0e },
	{ "ARCADE_M_HISUI_P_1_C.TXT", 0x31c,	0x35d0f },
	{ "ARCADE_M_HISUI_P_2.HA6", 0xfe14,	0x35d10 },
	{ "ARCADE_M_HISUI_P_2.TXT", 0xc4,	0x35d30 },
	{ "ARCADE_M_HISUI_P_2_C.TXT", 0x31c,	0x35d31 },
	{ "ARCADE_M_HISUI_P_9.HA6", 0x354c,	0x35d32 },
	{ "ARCADE_M_HISUI_P_9_C.TXT", 0x160e,	0x35d39 },
	{ "ARCADE_M_HISUI_P_SELIST.TXT", 0x84c,	0x35d3c },
	{ "ARCADE_M_HISUI_SELIST.TXT", 0xd1e,	0x35d3e },
	{ "ARCADE_NANAYA.ARX", 0x1c5600,	0x35d40 },
	{ "ARCADE_NANAYA.CG2", 0x57eb0,	0x360cb },
	{ "ARCADE_NANAYA.CPF", 0x2f36c,	0x3617b },
	{ "ARCADE_NANAYA.DOF", 0x918d4,	0x361da },
	{ "ARCADE_NANAYA.HA6", 0x6b748,	0x362fe },
	{ "ARCADE_NANAYA.PAT", 0x0,	0x363d5 },
	{ "ARCADE_NANAYA.TXT", 0xce,	0x363d5 },
	{ "ARCADE_NANAYA0.PAL", 0x400,	0x363d6 },
	{ "ARCADE_NANAYA1.PAL", 0x400,	0x363d7 },
	{ "ARCADE_NANAYA10.PAL", 0x400,	0x363d8 },
	{ "ARCADE_NANAYA11.PAL", 0x400,	0x363d9 },
	{ "ARCADE_NANAYA12.PAL", 0x400,	0x363da },
	{ "ARCADE_NANAYA13.PAL", 0x400,	0x363db },
	{ "ARCADE_NANAYA14.PAL", 0x400,	0x363dc },
	{ "ARCADE_NANAYA15.PAL", 0x400,	0x363dd },
	{ "ARCADE_NANAYA16.PAL", 0x400,	0x363de },
	{ "ARCADE_NANAYA17.PAL", 0x400,	0x363df },
	{ "ARCADE_NANAYA18.PAL", 0x400,	0x363e0 },
	{ "ARCADE_NANAYA19.PAL", 0x400,	0x363e1 },
	{ "ARCADE_NANAYA2.ARX", 0x1c5080,	0x363e2 },
	{ "ARCADE_NANAYA2.PAL", 0x400,	0x3676d },
	{ "ARCADE_NANAYA20.PAL", 0x400,	0x3676e },
	{ "ARCADE_NANAYA21.PAL", 0x400,	0x3676f },
	{ "ARCADE_NANAYA22.PAL", 0x400,	0x36770 },
	{ "ARCADE_NANAYA23.PAL", 0x400,	0x36771 },
	{ "ARCADE_NANAYA24.PAL", 0x400,	0x36772 },
	{ "ARCADE_NANAYA25.PAL", 0x400,	0x36773 },
	{ "ARCADE_NANAYA26.PAL", 0x400,	0x36774 },
	{ "ARCADE_NANAYA27.PAL", 0x400,	0x36775 },
	{ "ARCADE_NANAYA28.PAL", 0x400,	0x36776 },
	{ "ARCADE_NANAYA29.PAL", 0x400,	0x36777 },
	{ "ARCADE_NANAYA3.PAL", 0x400,	0x36778 },
	{ "ARCADE_NANAYA30.PAL", 0x400,	0x36779 },
	{ "ARCADE_NANAYA31.PAL", 0x400,	0x3677a },
	{ "ARCADE_NANAYA32.PAL", 0x400,	0x3677b },
	{ "ARCADE_NANAYA33.PAL", 0x400,	0x3677c },
	{ "ARCADE_NANAYA34.PAL", 0x400,	0x3677d },
	{ "ARCADE_NANAYA35.PAL", 0x400,	0x3677e },
	{ "ARCADE_NANAYA4.PAL", 0x400,	0x3677f },
	{ "ARCADE_NANAYA5.PAL", 0x400,	0x36780 },
	{ "ARCADE_NANAYA6.PAL", 0x400,	0x36781 },
	{ "ARCADE_NANAYA7.PAL", 0x400,	0x36782 },
	{ "ARCADE_NANAYA8.PAL", 0x400,	0x36783 },
	{ "ARCADE_NANAYA9.PAL", 0x400,	0x36784 },
	{ "ARCADE_NANAYAPAL2.ARX", 0x91c0,	0x36785 },
	{ "ARCADE_NANAYA_0.HA6", 0xbab8,	0x36798 },
	{ "ARCADE_NANAYA_0.TXT", 0xa8,	0x367b0 },
	{ "ARCADE_NANAYA_0_C.TXT", 0x1427,	0x367b1 },
	{ "ARCADE_NANAYA_1.HA6", 0x175f8,	0x367b4 },
	{ "ARCADE_NANAYA_1.TXT", 0xa4,	0x367e3 },
	{ "ARCADE_NANAYA_1_C.TXT", 0x187d,	0x367e4 },
	{ "ARCADE_NANAYA_2.HA6", 0x15d78,	0x367e8 },
	{ "ARCADE_NANAYA_2.TXT", 0xa4,	0x36814 },
	{ "ARCADE_NANAYA_2_C.TXT", 0x1250,	0x36815 },
	{ "ARCADE_NANAYA_3.HA6", 0x1100c,	0x36818 },
	{ "ARCADE_NANAYA_3.TXT", 0xa4,	0x3683b },
	{ "ARCADE_NANAYA_3_C.TXT", 0xe14,	0x3683c },
	{ "ARCADE_NANAYA_9.HA6", 0x16a88,	0x3683e },
	{ "ARCADE_NANAYA_9.TXT", 0xa4,	0x3686c },
	{ "ARCADE_NANAYA_9_C.TXT", 0xd7e,	0x3686d },
	{ "ARCADE_NANAYA_C.TXT", 0xf9d,	0x3686f },
	{ "ARCADE_NANAYA_SELIST.TXT", 0xf36,	0x36871 },
	{ "ARCADE_NANAYA_WIN.TXT", 0x712,	0x36873 },
	{ "ARCADE_NECHAOS.ARX", 0x35ec40,	0x36874 },
	{ "ARCADE_NECHAOS.CG2", 0x5fde8,	0x36f32 },
	{ "ARCADE_NECHAOS.CPF", 0x2f36c,	0x36ff2 },
	{ "ARCADE_NECHAOS.DOF", 0x6f67c,	0x37051 },
	{ "ARCADE_NECHAOS.HA6", 0x796c0,	0x37130 },
	{ "ARCADE_NECHAOS.PAT", 0x310b6,	0x37223 },
	{ "ARCADE_NECHAOS.TXT", 0xec,	0x37286 },
	{ "ARCADE_NECHAOS0.PAL", 0x400,	0x37287 },
	{ "ARCADE_NECHAOS1.PAL", 0x400,	0x37288 },
	{ "ARCADE_NECHAOS10.PAL", 0x400,	0x37289 },
	{ "ARCADE_NECHAOS11.PAL", 0x400,	0x3728a },
	{ "ARCADE_NECHAOS12.PAL", 0x400,	0x3728b },
	{ "ARCADE_NECHAOS13.PAL", 0x400,	0x3728c },
	{ "ARCADE_NECHAOS14.PAL", 0x400,	0x3728d },
	{ "ARCADE_NECHAOS15.PAL", 0x400,	0x3728e },
	{ "ARCADE_NECHAOS16.PAL", 0x400,	0x3728f },
	{ "ARCADE_NECHAOS17.PAL", 0x400,	0x37290 },
	{ "ARCADE_NECHAOS18.PAL", 0x400,	0x37291 },
	{ "ARCADE_NECHAOS19.PAL", 0x400,	0x37292 },
	{ "ARCADE_NECHAOS2.ARX", 0x343bc0,	0x37293 },
	{ "ARCADE_NECHAOS2.PAL", 0x400,	0x3791b },
	{ "ARCADE_NECHAOS20.PAL", 0x400,	0x3791c },
	{ "ARCADE_NECHAOS21.PAL", 0x400,	0x3791d },
	{ "ARCADE_NECHAOS22.PAL", 0x400,	0x3791e },
	{ "ARCADE_NECHAOS23.PAL", 0x400,	0x3791f },
	{ "ARCADE_NECHAOS24.PAL", 0x400,	0x37920 },
	{ "ARCADE_NECHAOS25.PAL", 0x400,	0x37921 },
	{ "ARCADE_NECHAOS26.PAL", 0x400,	0x37922 },
	{ "ARCADE_NECHAOS27.PAL", 0x400,	0x37923 },
	{ "ARCADE_NECHAOS28.PAL", 0x400,	0x37924 },
	{ "ARCADE_NECHAOS29.PAL", 0x400,	0x37925 },
	{ "ARCADE_NECHAOS3.PAL", 0x400,	0x37926 },
	{ "ARCADE_NECHAOS30.PAL", 0x400,	0x37927 },
	{ "ARCADE_NECHAOS31.PAL", 0x400,	0x37928 },
	{ "ARCADE_NECHAOS32.PAL", 0x400,	0x37929 },
	{ "ARCADE_NECHAOS33.PAL", 0x400,	0x3792a },
	{ "ARCADE_NECHAOS34.PAL", 0x400,	0x3792b },
	{ "ARCADE_NECHAOS35.PAL", 0x400,	0x3792c },
	{ "ARCADE_NECHAOS4.PAL", 0x400,	0x3792d },
	{ "ARCADE_NECHAOS5.PAL", 0x400,	0x3792e },
	{ "ARCADE_NECHAOS6.PAL", 0x400,	0x3792f },
	{ "ARCADE_NECHAOS7.PAL", 0x400,	0x37930 },
	{ "ARCADE_NECHAOS8.PAL", 0x400,	0x37931 },
	{ "ARCADE_NECHAOS9.PAL", 0x400,	0x37932 },
	{ "ARCADE_NECHAOSPAL2.ARX", 0x26b80,	0x37933 },
	{ "ARCADE_NECHAOS_0.HA6", 0x3564,	0x37981 },
	{ "ARCADE_NECHAOS_0.TXT", 0xc2,	0x37988 },
	{ "ARCADE_NECHAOS_0_C.TXT", 0x1240,	0x37989 },
	{ "ARCADE_NECHAOS_1.HA6", 0x59a0,	0x3798c },
	{ "ARCADE_NECHAOS_1.TXT", 0xc2,	0x37998 },
	{ "ARCADE_NECHAOS_1_C.TXT", 0x1112,	0x37999 },
	{ "ARCADE_NECHAOS_2.HA6", 0x7bdc,	0x3799c },
	{ "ARCADE_NECHAOS_2.TXT", 0xc2,	0x379ac },
	{ "ARCADE_NECHAOS_2_C.TXT", 0x12e5,	0x379ad },
	{ "ARCADE_NECHAOS_SELIST.TXT", 0xf5b,	0x379b0 },
	{ "ARCADE_NECO.ARX", 0x2d6080,	0x379b2 },
	{ "ARCADE_NECO.CG2", 0x47c08,	0x37f5f },
	{ "ARCADE_NECO.CPF", 0x2f36c,	0x37fef },
	{ "ARCADE_NECO.DOF", 0x66304,	0x3804e },
	{ "ARCADE_NECO.HA6", 0x5cf28,	0x3811b },
	{ "ARCADE_NECO.PAT", 0x310b6,	0x381d5 },
	{ "ARCADE_NECO.TXT", 0xd3,	0x38238 },
	{ "ARCADE_NECO0.PAL", 0x400,	0x38239 },
	{ "ARCADE_NECO1.PAL", 0x400,	0x3823a },
	{ "ARCADE_NECO10.PAL", 0x400,	0x3823b },
	{ "ARCADE_NECO11.PAL", 0x400,	0x3823c },
	{ "ARCADE_NECO12.PAL", 0x400,	0x3823d },
	{ "ARCADE_NECO13.PAL", 0x400,	0x3823e },
	{ "ARCADE_NECO14.PAL", 0x400,	0x3823f },
	{ "ARCADE_NECO15.PAL", 0x400,	0x38240 },
	{ "ARCADE_NECO16.PAL", 0x400,	0x38241 },
	{ "ARCADE_NECO17.PAL", 0x400,	0x38242 },
	{ "ARCADE_NECO18.PAL", 0x400,	0x38243 },
	{ "ARCADE_NECO19.PAL", 0x400,	0x38244 },
	{ "ARCADE_NECO2.ARX", 0x2d0c80,	0x38245 },
	{ "ARCADE_NECO2.PAL", 0x400,	0x387e7 },
	{ "ARCADE_NECO20.PAL", 0x400,	0x387e8 },
	{ "ARCADE_NECO21.PAL", 0x400,	0x387e9 },
	{ "ARCADE_NECO22.PAL", 0x400,	0x387ea },
	{ "ARCADE_NECO23.PAL", 0x400,	0x387eb },
	{ "ARCADE_NECO24.PAL", 0x400,	0x387ec },
	{ "ARCADE_NECO25.PAL", 0x400,	0x387ed },
	{ "ARCADE_NECO26.PAL", 0x400,	0x387ee },
	{ "ARCADE_NECO27.PAL", 0x400,	0x387ef },
	{ "ARCADE_NECO28.PAL", 0x400,	0x387f0 },
	{ "ARCADE_NECO29.PAL", 0x400,	0x387f1 },
	{ "ARCADE_NECO3.PAL", 0x400,	0x387f2 },
	{ "ARCADE_NECO30.PAL", 0x400,	0x387f3 },
	{ "ARCADE_NECO31.PAL", 0x400,	0x387f4 },
	{ "ARCADE_NECO32.PAL", 0x400,	0x387f5 },
	{ "ARCADE_NECO33.PAL", 0x400,	0x387f6 },
	{ "ARCADE_NECO34.PAL", 0x400,	0x387f7 },
	{ "ARCADE_NECO35.PAL", 0x400,	0x387f8 },
	{ "ARCADE_NECO4.PAL", 0x400,	0x387f9 },
	{ "ARCADE_NECO5.PAL", 0x400,	0x387fa },
	{ "ARCADE_NECO6.PAL", 0x400,	0x387fb },
	{ "ARCADE_NECO7.PAL", 0x400,	0x387fc },
	{ "ARCADE_NECO8.PAL", 0x400,	0x387fd },
	{ "ARCADE_NECO9.PAL", 0x400,	0x387fe },
	{ "ARCADE_NECOPAL2.ARX", 0x25280,	0x387ff },
	{ "ARCADE_NECO_0.HA6", 0x34fc,	0x3884a },
	{ "ARCADE_NECO_0.TXT", 0xab,	0x38851 },
	{ "ARCADE_NECO_0_C.TXT", 0x1a9f,	0x38852 },
	{ "ARCADE_NECO_1.HA6", 0x90cc,	0x38856 },
	{ "ARCADE_NECO_1.TXT", 0xab,	0x38869 },
	{ "ARCADE_NECO_1_C.TXT", 0x1732,	0x3886a },
	{ "ARCADE_NECO_2.HA6", 0x5ee4,	0x3886d },
	{ "ARCADE_NECO_2.TXT", 0xab,	0x38879 },
	{ "ARCADE_NECO_2_C.TXT", 0x16d0,	0x3887a },
	{ "ARCADE_NECO_3.HA6", 0x1100c,	0x3887d },
	{ "ARCADE_NECO_9.HA6", 0x13480,	0x388a0 },
	{ "ARCADE_NECO_9.TXT", 0xab,	0x388c7 },
	{ "ARCADE_NECO_9_C.TXT", 0x165f,	0x388c8 },
	{ "ARCADE_NECO_P.ARX", 0xc2f00,	0x3890d },
	{ "ARCADE_NECO_P.CG2", 0x24838,	0x38a93 },
	{ "ARCADE_NECO_P.CPF", 0x2f36c,	0x38add },
	{ "ARCADE_NECO_P.HA6", 0x5a00c,	0x38b3c },
	{ "ARCADE_NECO_P.TXT", 0x9e,	0x38bf1 },
	{ "ARCADE_NECO_P0.PAL", 0x400,	0x38bf2 },
	{ "ARCADE_NECO_P1.PAL", 0x400,	0x38bf3 },
	{ "ARCADE_NECO_P10.PAL", 0x400,	0x38bf4 },
	{ "ARCADE_NECO_P11.PAL", 0x400,	0x38bf5 },
	{ "ARCADE_NECO_P12.PAL", 0x400,	0x38bf6 },
	{ "ARCADE_NECO_P13.PAL", 0x400,	0x38bf7 },
	{ "ARCADE_NECO_P14.PAL", 0x400,	0x38bf8 },
	{ "ARCADE_NECO_P15.PAL", 0x400,	0x38bf9 },
	{ "ARCADE_NECO_P16.PAL", 0x400,	0x38bfa },
	{ "ARCADE_NECO_P17.PAL", 0x400,	0x38bfb },
	{ "ARCADE_NECO_P18.PAL", 0x400,	0x38bfc },
	{ "ARCADE_NECO_P19.PAL", 0x400,	0x38bfd },
	{ "ARCADE_NECO_P2.ARX", 0xbe580,	0x38bfe },
	{ "ARCADE_NECO_P2.PAL", 0x400,	0x38d7b },
	{ "ARCADE_NECO_P20.PAL", 0x400,	0x38d7c },
	{ "ARCADE_NECO_P21.PAL", 0x400,	0x38d7d },
	{ "ARCADE_NECO_P22.PAL", 0x400,	0x38d7e },
	{ "ARCADE_NECO_P23.PAL", 0x400,	0x38d7f },
	{ "ARCADE_NECO_P24.PAL", 0x400,	0x38d80 },
	{ "ARCADE_NECO_P25.PAL", 0x400,	0x38d81 },
	{ "ARCADE_NECO_P26.PAL", 0x400,	0x38d82 },
	{ "ARCADE_NECO_P27.PAL", 0x400,	0x38d83 },
	{ "ARCADE_NECO_P28.PAL", 0x400,	0x38d84 },
	{ "ARCADE_NECO_P29.PAL", 0x400,	0x38d85 },
	{ "ARCADE_NECO_P3.PAL", 0x400,	0x38d86 },
	{ "ARCADE_NECO_P30.PAL", 0x400,	0x38d87 },
	{ "ARCADE_NECO_P31.PAL", 0x400,	0x38d88 },
	{ "ARCADE_NECO_P32.PAL", 0x400,	0x38d89 },
	{ "ARCADE_NECO_P33.PAL", 0x400,	0x38d8a },
	{ "ARCADE_NECO_P34.PAL", 0x400,	0x38d8b },
	{ "ARCADE_NECO_P35.PAL", 0x400,	0x38d8c },
	{ "ARCADE_NECO_P4.PAL", 0x400,	0x38d8d },
	{ "ARCADE_NECO_P5.PAL", 0x400,	0x38d8e },
	{ "ARCADE_NECO_P6.PAL", 0x400,	0x38d8f },
	{ "ARCADE_NECO_P7.PAL", 0x400,	0x38d90 },
	{ "ARCADE_NECO_P8.PAL", 0x400,	0x38d91 },
	{ "ARCADE_NECO_P9.PAL", 0x400,	0x38d92 },
	{ "ARCADE_NECO_PPAL2.ARX", 0x9600,	0x38d93 },
	{ "ARCADE_NECO_P_0.HA6", 0x354c,	0x38da6 },
	{ "ARCADE_NECO_P_0.TXT", 0xb3,	0x38dad },
	{ "ARCADE_NECO_P_0_C.TXT", 0x17b2,	0x38dae },
	{ "ARCADE_NECO_P_1.HA6", 0x94d8,	0x38db1 },
	{ "ARCADE_NECO_P_1.TXT", 0xb3,	0x38dc4 },
	{ "ARCADE_NECO_P_1_C.TXT", 0x18a2,	0x38dc5 },
	{ "ARCADE_NECO_P_2.HA6", 0x86f8,	0x38dc9 },
	{ "ARCADE_NECO_P_2.TXT", 0xb3,	0x38dda },
	{ "ARCADE_NECO_P_2_C.TXT", 0x17b4,	0x38ddb },
	{ "ARCADE_NECO_P_9.HA6", 0x2c,	0x38dde },
	{ "ARCADE_NECO_P_SELIST.TXT", 0xd39,	0x38ddf },
	{ "ARCADE_NECO_SELIST.TXT", 0xdaa,	0x38de1 },
	{ "ARCADE_NERO.ARX", 0x38e6c0,	0x38de3 },
	{ "ARCADE_NERO.CG2", 0x8c308,	0x39500 },
	{ "ARCADE_NERO.CPF", 0x2f36c,	0x39619 },
	{ "ARCADE_NERO.DOF", 0x9eaa0,	0x39678 },
	{ "ARCADE_NERO.HA6", 0x75bd8,	0x397b6 },
	{ "ARCADE_NERO.PAT", 0x0,	0x398a2 },
	{ "ARCADE_NERO.TXT", 0xc2,	0x398a2 },
	{ "ARCADE_NERO0.PAL", 0x400,	0x398a3 },
	{ "ARCADE_NERO1.PAL", 0x400,	0x398a4 },
	{ "ARCADE_NERO10.PAL", 0x400,	0x398a5 },
	{ "ARCADE_NERO11.PAL", 0x400,	0x398a6 },
	{ "ARCADE_NERO12.PAL", 0x400,	0x398a7 },
	{ "ARCADE_NERO13.PAL", 0x400,	0x398a8 },
	{ "ARCADE_NERO14.PAL", 0x400,	0x398a9 },
	{ "ARCADE_NERO15.PAL", 0x400,	0x398aa },
	{ "ARCADE_NERO16.PAL", 0x400,	0x398ab },
	{ "ARCADE_NERO17.PAL", 0x400,	0x398ac },
	{ "ARCADE_NERO18.PAL", 0x400,	0x398ad },
	{ "ARCADE_NERO19.PAL", 0x400,	0x398ae },
	{ "ARCADE_NERO2.ARX", 0x37f700,	0x398af },
	{ "ARCADE_NERO2.PAL", 0x400,	0x39fae },
	{ "ARCADE_NERO20.PAL", 0x400,	0x39faf },
	{ "ARCADE_NERO21.PAL", 0x400,	0x39fb0 },
	{ "ARCADE_NERO22.PAL", 0x400,	0x39fb1 },
	{ "ARCADE_NERO23.PAL", 0x400,	0x39fb2 },
	{ "ARCADE_NERO24.PAL", 0x400,	0x39fb3 },
	{ "ARCADE_NERO25.PAL", 0x400,	0x39fb4 },
	{ "ARCADE_NERO26.PAL", 0x400,	0x39fb5 },
	{ "ARCADE_NERO27.PAL", 0x400,	0x39fb6 },
	{ "ARCADE_NERO28.PAL", 0x400,	0x39fb7 },
	{ "ARCADE_NERO29.PAL", 0x400,	0x39fb8 },
	{ "ARCADE_NERO3.PAL", 0x400,	0x39fb9 },
	{ "ARCADE_NERO30.PAL", 0x400,	0x39fba },
	{ "ARCADE_NERO31.PAL", 0x400,	0x39fbb },
	{ "ARCADE_NERO32.PAL", 0x400,	0x39fbc },
	{ "ARCADE_NERO33.PAL", 0x400,	0x39fbd },
	{ "ARCADE_NERO34.PAL", 0x400,	0x39fbe },
	{ "ARCADE_NERO35.PAL", 0x400,	0x39fbf },
	{ "ARCADE_NERO4.PAL", 0x400,	0x39fc0 },
	{ "ARCADE_NERO5.PAL", 0x400,	0x39fc1 },
	{ "ARCADE_NERO6.PAL", 0x400,	0x39fc2 },
	{ "ARCADE_NERO7.PAL", 0x400,	0x39fc3 },
	{ "ARCADE_NERO8.PAL", 0x400,	0x39fc4 },
	{ "ARCADE_NERO9.PAL", 0x400,	0x39fc5 },
	{ "ARCADE_NEROPAL2.ARX", 0x116c0,	0x39fc6 },
	{ "ARCADE_NERO_0.HA6", 0xb8ac,	0x39fe9 },
	{ "ARCADE_NERO_0.TXT", 0x9c,	0x3a001 },
	{ "ARCADE_NERO_0_C.TXT", 0xfaf,	0x3a002 },
	{ "ARCADE_NERO_1.HA6", 0x17294,	0x3a004 },
	{ "ARCADE_NERO_1.TXT", 0x9c,	0x3a033 },
	{ "ARCADE_NERO_1_C.TXT", 0xe79,	0x3a034 },
	{ "ARCADE_NERO_2.HA6", 0x1365c,	0x3a036 },
	{ "ARCADE_NERO_2.TXT", 0x9c,	0x3a05d },
	{ "ARCADE_NERO_2_C.TXT", 0xfbe,	0x3a05e },
	{ "ARCADE_NERO_3.HA6", 0x1100c,	0x3a060 },
	{ "ARCADE_NERO_9.HA6", 0x1fdec,	0x3a083 },
	{ "ARCADE_NERO_9.TXT", 0x9c,	0x3a0c3 },
	{ "ARCADE_NERO_9_C.TXT", 0xdc4,	0x3a0c4 },
	{ "ARCADE_NERO_SELIST.TXT", 0xb41,	0x3a0c6 },
	{ "ARCADE_NERO_WIN.TXT", 0x72c,	0x3a0c8 },
	{ "ARCADE_RIES.ARX", 0x3a62c0,	0x3a28e },
	{ "ARCADE_RIES.CG2", 0x7d998,	0x3a9db },
	{ "ARCADE_RIES.CPF", 0x2f36c,	0x3aad7 },
	{ "ARCADE_RIES.HA6", 0x30e70,	0x3ab36 },
	{ "ARCADE_RIES.PAT", 0xd120,	0x3ab98 },
	{ "ARCADE_RIES.TXT", 0xd3,	0x3abb3 },
	{ "ARCADE_RIES0.PAL", 0x400,	0x3abb4 },
	{ "ARCADE_RIES1.PAL", 0x400,	0x3abb5 },
	{ "ARCADE_RIES10.PAL", 0x400,	0x3abb6 },
	{ "ARCADE_RIES11.PAL", 0x400,	0x3abb7 },
	{ "ARCADE_RIES12.PAL", 0x400,	0x3abb8 },
	{ "ARCADE_RIES13.PAL", 0x400,	0x3abb9 },
	{ "ARCADE_RIES14.PAL", 0x400,	0x3abba },
	{ "ARCADE_RIES15.PAL", 0x400,	0x3abbb },
	{ "ARCADE_RIES16.PAL", 0x400,	0x3abbc },
	{ "ARCADE_RIES17.PAL", 0x400,	0x3abbd },
	{ "ARCADE_RIES18.PAL", 0x400,	0x3abbe },
	{ "ARCADE_RIES19.PAL", 0x400,	0x3abbf },
	{ "ARCADE_RIES2.ARX", 0x39f980,	0x3abc0 },
	{ "ARCADE_RIES2.PAL", 0x400,	0x3b300 },
	{ "ARCADE_RIES20.PAL", 0x400,	0x3b301 },
	{ "ARCADE_RIES21.PAL", 0x400,	0x3b302 },
	{ "ARCADE_RIES22.PAL", 0x400,	0x3b303 },
	{ "ARCADE_RIES23.PAL", 0x400,	0x3b304 },
	{ "ARCADE_RIES24.PAL", 0x400,	0x3b305 },
	{ "ARCADE_RIES25.PAL", 0x400,	0x3b306 },
	{ "ARCADE_RIES26.PAL", 0x400,	0x3b307 },
	{ "ARCADE_RIES27.PAL", 0x400,	0x3b308 },
	{ "ARCADE_RIES28.PAL", 0x400,	0x3b309 },
	{ "ARCADE_RIES29.PAL", 0x400,	0x3b30a },
	{ "ARCADE_RIES3.PAL", 0x400,	0x3b30b },
	{ "ARCADE_RIES30.PAL", 0x400,	0x3b30c },
	{ "ARCADE_RIES31.PAL", 0x400,	0x3b30d },
	{ "ARCADE_RIES32.PAL", 0x400,	0x3b30e },
	{ "ARCADE_RIES33.PAL", 0x400,	0x3b30f },
	{ "ARCADE_RIES34.PAL", 0x400,	0x3b310 },
	{ "ARCADE_RIES35.PAL", 0x400,	0x3b311 },
	{ "ARCADE_RIES4.PAL", 0x400,	0x3b312 },
	{ "ARCADE_RIES5.PAL", 0x400,	0x3b313 },
	{ "ARCADE_RIES6.PAL", 0x400,	0x3b314 },
	{ "ARCADE_RIES7.PAL", 0x400,	0x3b315 },
	{ "ARCADE_RIES8.PAL", 0x400,	0x3b316 },
	{ "ARCADE_RIES9.PAL", 0x400,	0x3b317 },
	{ "ARCADE_RIESPAL2.ARX", 0x19780,	0x3b318 },
	{ "ARCADE_RIES_0.HA6", 0x12174,	0x3b34b },
	{ "ARCADE_RIES_0.TXT", 0xb5,	0x3b370 },
	{ "ARCADE_RIES_0_C.TXT", 0x15ab,	0x3b371 },
	{ "ARCADE_RIES_1.HA6", 0x14c9c,	0x3b374 },
	{ "ARCADE_RIES_1.TXT", 0xaf,	0x3b39e },
	{ "ARCADE_RIES_1_C.TXT", 0xf3c,	0x3b39f },
	{ "ARCADE_RIES_2.HA6", 0x13e70,	0x3b3a1 },
	{ "ARCADE_RIES_2.TXT", 0xad,	0x3b3c9 },
	{ "ARCADE_RIES_2_C.TXT", 0xf10,	0x3b3ca },
	{ "ARCADE_RIES_3.HA6", 0x1100c,	0x3b3cc },
	{ "ARCADE_RIES_3.TXT", 0xad,	0x3b3ef },
	{ "ARCADE_RIES_3_C.TXT", 0xb71,	0x3b3f0 },
	{ "ARCADE_RIES_9.HA6", 0x2b9c8,	0x3b3f2 },
	{ "ARCADE_RIES_9.TXT", 0xad,	0x3b44a },
	{ "ARCADE_RIES_9_C.TXT", 0xa18,	0x3b44b },
	{ "ARCADE_RIES_SELIST.TXT", 0xd9f,	0x3b44d },
	{ "ARCADE_ROA.ARX", 0x38e1c0,	0x3b470 },
	{ "ARCADE_ROA.CG2", 0x60550,	0x3bb8d },
	{ "ARCADE_ROA.CPF", 0x2f36c,	0x3bc4e },
	{ "ARCADE_ROA.HA6", 0x6ee8c,	0x3bcad },
	{ "ARCADE_ROA.PAT", 0x0,	0x3bd8b },
	{ "ARCADE_ROA.TXT", 0xbc,	0x3bd8b },
	{ "ARCADE_ROA0.PAL", 0x400,	0x3bd8c },
	{ "ARCADE_ROA1.PAL", 0x400,	0x3bd8d },
	{ "ARCADE_ROA10.PAL", 0x400,	0x3bd8e },
	{ "ARCADE_ROA11.PAL", 0x400,	0x3bd8f },
	{ "ARCADE_ROA12.PAL", 0x400,	0x3bd90 },
	{ "ARCADE_ROA13.PAL", 0x400,	0x3bd91 },
	{ "ARCADE_ROA14.PAL", 0x400,	0x3bd92 },
	{ "ARCADE_ROA15.PAL", 0x400,	0x3bd93 },
	{ "ARCADE_ROA16.PAL", 0x400,	0x3bd94 },
	{ "ARCADE_ROA17.PAL", 0x400,	0x3bd95 },
	{ "ARCADE_ROA18.PAL", 0x400,	0x3bd96 },
	{ "ARCADE_ROA19.PAL", 0x400,	0x3bd97 },
	{ "ARCADE_ROA2.ARX", 0x38acc0,	0x3bd98 },
	{ "ARCADE_ROA2.PAL", 0x400,	0x3c4ae },
	{ "ARCADE_ROA20.PAL", 0x400,	0x3c4af },
	{ "ARCADE_ROA21.PAL", 0x400,	0x3c4b0 },
	{ "ARCADE_ROA22.PAL", 0x400,	0x3c4b1 },
	{ "ARCADE_ROA23.PAL", 0x400,	0x3c4b2 },
	{ "ARCADE_ROA24.PAL", 0x400,	0x3c4b3 },
	{ "ARCADE_ROA25.PAL", 0x400,	0x3c4b4 },
	{ "ARCADE_ROA26.PAL", 0x400,	0x3c4b5 },
	{ "ARCADE_ROA27.PAL", 0x400,	0x3c4b6 },
	{ "ARCADE_ROA28.PAL", 0x400,	0x3c4b7 },
	{ "ARCADE_ROA29.PAL", 0x400,	0x3c4b8 },
	{ "ARCADE_ROA3.PAL", 0x400,	0x3c4b9 },
	{ "ARCADE_ROA30.PAL", 0x400,	0x3c4ba },
	{ "ARCADE_ROA31.PAL", 0x400,	0x3c4bb },
	{ "ARCADE_ROA32.PAL", 0x400,	0x3c4bc },
	{ "ARCADE_ROA33.PAL", 0x400,	0x3c4bd },
	{ "ARCADE_ROA34.PAL", 0x400,	0x3c4be },
	{ "ARCADE_ROA35.PAL", 0x400,	0x3c4bf },
	{ "ARCADE_ROA4.PAL", 0x400,	0x3c4c0 },
	{ "ARCADE_ROA5.PAL", 0x400,	0x3c4c1 },
	{ "ARCADE_ROA6.PAL", 0x400,	0x3c4c2 },
	{ "ARCADE_ROA7.PAL", 0x400,	0x3c4c3 },
	{ "ARCADE_ROA8.PAL", 0x400,	0x3c4c4 },
	{ "ARCADE_ROA9.PAL", 0x400,	0x3c4c5 },
	{ "ARCADE_ROAPAL2.ARX", 0x239c0,	0x3c4c6 },
	{ "ARCADE_ROA_0.HA6", 0x3e5c,	0x3c50e },
	{ "ARCADE_ROA_0.TXT", 0x98,	0x3c516 },
	{ "ARCADE_ROA_0_C.TXT", 0x1546,	0x3c517 },
	{ "ARCADE_ROA_1.HA6", 0x9320,	0x3c51a },
	{ "ARCADE_ROA_1.TXT", 0x98,	0x3c52d },
	{ "ARCADE_ROA_1_C.TXT", 0x109d,	0x3c52e },
	{ "ARCADE_ROA_2.HA6", 0x593c,	0x3c531 },
	{ "ARCADE_ROA_2.TXT", 0x98,	0x3c53d },
	{ "ARCADE_ROA_2_C.TXT", 0x1093,	0x3c53e },
	{ "ARCADE_ROA_3.HA6", 0x1100c,	0x3c541 },
	{ "ARCADE_ROA_9.HA6", 0x15548,	0x3c564 },
	{ "ARCADE_ROA_9.TXT", 0x98,	0x3c58f },
	{ "ARCADE_ROA_9_C.TXT", 0xf88,	0x3c590 },
	{ "ARCADE_ROA_SELIST.TXT", 0xc76,	0x3c592 },
	{ "ARCADE_RYOUGI.ARX", 0x2e4340,	0x3c594 },
	{ "ARCADE_RYOUGI.CG2", 0x7abc0,	0x3cb5d },
	{ "ARCADE_RYOUGI.CPF", 0x2f36c,	0x3cc53 },
	{ "ARCADE_RYOUGI.HA6", 0xb4fdc,	0x3ccb2 },
	{ "ARCADE_RYOUGI.PAT", 0xd506,	0x3ce1c },
	{ "ARCADE_RYOUGI0.PAL", 0x400,	0x3ce37 },
	{ "ARCADE_RYOUGI1.PAL", 0x400,	0x3ce38 },
	{ "ARCADE_RYOUGI10.PAL", 0x400,	0x3ce39 },
	{ "ARCADE_RYOUGI11.PAL", 0x400,	0x3ce3a },
	{ "ARCADE_RYOUGI12.PAL", 0x400,	0x3ce3b },
	{ "ARCADE_RYOUGI13.PAL", 0x400,	0x3ce3c },
	{ "ARCADE_RYOUGI14.PAL", 0x400,	0x3ce3d },
	{ "ARCADE_RYOUGI15.PAL", 0x400,	0x3ce3e },
	{ "ARCADE_RYOUGI16.PAL", 0x400,	0x3ce3f },
	{ "ARCADE_RYOUGI17.PAL", 0x400,	0x3ce40 },
	{ "ARCADE_RYOUGI18.PAL", 0x400,	0x3ce41 },
	{ "ARCADE_RYOUGI19.PAL", 0x400,	0x3ce42 },
	{ "ARCADE_RYOUGI2.ARX", 0x2f94c0,	0x3ce43 },
	{ "ARCADE_RYOUGI2.PAL", 0x400,	0x3d436 },
	{ "ARCADE_RYOUGI20.PAL", 0x400,	0x3d437 },
	{ "ARCADE_RYOUGI21.PAL", 0x400,	0x3d438 },
	{ "ARCADE_RYOUGI22.PAL", 0x400,	0x3d439 },
	{ "ARCADE_RYOUGI23.PAL", 0x400,	0x3d43a },
	{ "ARCADE_RYOUGI24.PAL", 0x400,	0x3d43b },
	{ "ARCADE_RYOUGI25.PAL", 0x400,	0x3d43c },
	{ "ARCADE_RYOUGI26.PAL", 0x400,	0x3d43d },
	{ "ARCADE_RYOUGI27.PAL", 0x400,	0x3d43e },
	{ "ARCADE_RYOUGI28.PAL", 0x400,	0x3d43f },
	{ "ARCADE_RYOUGI29.PAL", 0x400,	0x3d440 },
	{ "ARCADE_RYOUGI3.PAL", 0x400,	0x3d441 },
	{ "ARCADE_RYOUGI30.PAL", 0x400,	0x3d442 },
	{ "ARCADE_RYOUGI31.PAL", 0x400,	0x3d443 },
	{ "ARCADE_RYOUGI32.PAL", 0x400,	0x3d444 },
	{ "ARCADE_RYOUGI33.PAL", 0x400,	0x3d445 },
	{ "ARCADE_RYOUGI34.PAL", 0x400,	0x3d446 },
	{ "ARCADE_RYOUGI35.PAL", 0x400,	0x3d447 },
	{ "ARCADE_RYOUGI4.PAL", 0x400,	0x3d448 },
	{ "ARCADE_RYOUGI5.PAL", 0x400,	0x3d449 },
	{ "ARCADE_RYOUGI6.PAL", 0x400,	0x3d44a },
	{ "ARCADE_RYOUGI7.PAL", 0x400,	0x3d44b },
	{ "ARCADE_RYOUGI8.PAL", 0x400,	0x3d44c },
	{ "ARCADE_RYOUGI9.PAL", 0x400,	0x3d44d },
	{ "ARCADE_RYOUGIPAL2.ARX", 0x13800,	0x3d44e },
	{ "ARCADE_RYOUGI_0.HA6", 0x59a8,	0x3d475 },
	{ "ARCADE_RYOUGI_0.TXT", 0xc1,	0x3d481 },
	{ "ARCADE_RYOUGI_0_C.TXT", 0x2a6b,	0x3d482 },
	{ "ARCADE_RYOUGI_1.HA6", 0x7a08,	0x3d488 },
	{ "ARCADE_RYOUGI_1.TXT", 0xbf,	0x3d498 },
	{ "ARCADE_RYOUGI_1_C.TXT", 0x1f25,	0x3d499 },
	{ "ARCADE_RYOUGI_2.HA6", 0x96dc,	0x3d49d },
	{ "ARCADE_RYOUGI_2.TXT", 0xc1,	0x3d4b0 },
	{ "ARCADE_RYOUGI_2_C.TXT", 0x21cd,	0x3d4b1 },
	{ "ARCADE_RYOUGI_9.HA6", 0x19594,	0x3d4b6 },
	{ "ARCADE_RYOUGI_9.TXT", 0xc1,	0x3d4e9 },
	{ "ARCADE_RYOUGI_9_C.TXT", 0x1f6f,	0x3d4ea },
	{ "ARCADE_RYOUGI_SELIST.TXT", 0x1e59,	0x3d4ee },
	{ "ARCADE_SAMPLE_C.TXT", 0xa9c,	0x3d4f2 },
	{ "ARCADE_SATSUKI.ARX", 0x32b500,	0x3d4f4 },
	{ "ARCADE_SATSUKI.CG2", 0x5f4b8,	0x3db4b },
	{ "ARCADE_SATSUKI.CPF", 0x2f36c,	0x3dc0a },
	{ "ARCADE_SATSUKI.DOF", 0x6f4d8,	0x3dc69 },
	{ "ARCADE_SATSUKI.HA6", 0x646d8,	0x3dd48 },
	{ "ARCADE_SATSUKI.PAT", 0x0,	0x3de11 },
	{ "ARCADE_SATSUKI.TXT", 0xd4,	0x3de11 },
	{ "ARCADE_SATSUKI0.PAL", 0x400,	0x3de12 },
	{ "ARCADE_SATSUKI1.PAL", 0x400,	0x3de13 },
	{ "ARCADE_SATSUKI10.PAL", 0x400,	0x3de14 },
	{ "ARCADE_SATSUKI11.PAL", 0x400,	0x3de15 },
	{ "ARCADE_SATSUKI12.PAL", 0x400,	0x3de16 },
	{ "ARCADE_SATSUKI13.PAL", 0x400,	0x3de17 },
	{ "ARCADE_SATSUKI14.PAL", 0x400,	0x3de18 },
	{ "ARCADE_SATSUKI15.PAL", 0x400,	0x3de19 },
	{ "ARCADE_SATSUKI16.PAL", 0x400,	0x3de1a },
	{ "ARCADE_SATSUKI17.PAL", 0x400,	0x3de1b },
	{ "ARCADE_SATSUKI18.PAL", 0x400,	0x3de1c },
	{ "ARCADE_SATSUKI19.PAL", 0x400,	0x3de1d },
	{ "ARCADE_SATSUKI2.ARX", 0x3266c0,	0x3de1e },
	{ "ARCADE_SATSUKI2.PAL", 0x400,	0x3e46b },
	{ "ARCADE_SATSUKI20.PAL", 0x400,	0x3e46c },
	{ "ARCADE_SATSUKI21.PAL", 0x400,	0x3e46d },
	{ "ARCADE_SATSUKI22.PAL", 0x400,	0x3e46e },
	{ "ARCADE_SATSUKI23.PAL", 0x400,	0x3e46f },
	{ "ARCADE_SATSUKI24.PAL", 0x400,	0x3e470 },
	{ "ARCADE_SATSUKI25.PAL", 0x400,	0x3e471 },
	{ "ARCADE_SATSUKI26.PAL", 0x400,	0x3e472 },
	{ "ARCADE_SATSUKI27.PAL", 0x400,	0x3e473 },
	{ "ARCADE_SATSUKI28.PAL", 0x400,	0x3e474 },
	{ "ARCADE_SATSUKI29.PAL", 0x400,	0x3e475 },
	{ "ARCADE_SATSUKI3.PAL", 0x400,	0x3e476 },
	{ "ARCADE_SATSUKI30.PAL", 0x400,	0x3e477 },
	{ "ARCADE_SATSUKI31.PAL", 0x400,	0x3e478 },
	{ "ARCADE_SATSUKI32.PAL", 0x400,	0x3e479 },
	{ "ARCADE_SATSUKI33.PAL", 0x400,	0x3e47a },
	{ "ARCADE_SATSUKI34.PAL", 0x400,	0x3e47b },
	{ "ARCADE_SATSUKI35.PAL", 0x400,	0x3e47c },
	{ "ARCADE_SATSUKI4.PAL", 0x400,	0x3e47d },
	{ "ARCADE_SATSUKI5.PAL", 0x400,	0x3e47e },
	{ "ARCADE_SATSUKI6.PAL", 0x400,	0x3e47f },
	{ "ARCADE_SATSUKI7.PAL", 0x400,	0x3e480 },
	{ "ARCADE_SATSUKI8.PAL", 0x400,	0x3e481 },
	{ "ARCADE_SATSUKI9.PAL", 0x400,	0x3e482 },
	{ "ARCADE_SATSUKIPAL2.ARX", 0x21440,	0x3e483 },
	{ "ARCADE_SATSUKI_0.HA6", 0x7240,	0x3e4c6 },
	{ "ARCADE_SATSUKI_0.TXT", 0xaa,	0x3e4d5 },
	{ "ARCADE_SATSUKI_0_C.TXT", 0x1299,	0x3e4d6 },
	{ "ARCADE_SATSUKI_1.HA6", 0xae7c,	0x3e4d9 },
	{ "ARCADE_SATSUKI_1.TXT", 0xaa,	0x3e4ef },
	{ "ARCADE_SATSUKI_1_C.TXT", 0x107d,	0x3e4f0 },
	{ "ARCADE_SATSUKI_2.HA6", 0xe7bc,	0x3e4f3 },
	{ "ARCADE_SATSUKI_2.TXT", 0xa8,	0x3e510 },
	{ "ARCADE_SATSUKI_2_C.TXT", 0xfdc,	0x3e511 },
	{ "ARCADE_SATSUKI_3.HA6", 0x1100c,	0x3e513 },
	{ "ARCADE_SATSUKI_3.TXT", 0xa8,	0x3e536 },
	{ "ARCADE_SATSUKI_3_C.TXT", 0xd8c,	0x3e537 },
	{ "ARCADE_SATSUKI_9.HA6", 0x1a160,	0x3e539 },
	{ "ARCADE_SATSUKI_9.TXT", 0xa8,	0x3e56e },
	{ "ARCADE_SATSUKI_9_C.TXT", 0xe91,	0x3e56f },
	{ "ARCADE_SATSUKI_C.TXT", 0xfc9,	0x3e571 },
	{ "ARCADE_SATSUKI_SELIST.TXT", 0xf14,	0x3e573 },
	{ "ARCADE_SATSUKI_WIN.TXT", 0x92c,	0x3e575 },
	{ "ARCADE_SHIKI.ARX", 0x1cee80,	0x3e577 },
	{ "ARCADE_SHIKI.CG2", 0x5a7a8,	0x3e915 },
	{ "ARCADE_SHIKI.CPF", 0x2f36c,	0x3e9ca },
	{ "ARCADE_SHIKI.HA6", 0x793dc,	0x3ea29 },
	{ "ARCADE_SHIKI.PAT", 0x0,	0x3eb1c },
	{ "ARCADE_SHIKI.TXT", 0xca,	0x3eb1c },
	{ "ARCADE_SHIKI0.PAL", 0x400,	0x3eb1d },
	{ "ARCADE_SHIKI1.PAL", 0x400,	0x3eb1e },
	{ "ARCADE_SHIKI10.PAL", 0x400,	0x3eb1f },
	{ "ARCADE_SHIKI11.PAL", 0x400,	0x3eb20 },
	{ "ARCADE_SHIKI12.PAL", 0x400,	0x3eb21 },
	{ "ARCADE_SHIKI13.PAL", 0x400,	0x3eb22 },
	{ "ARCADE_SHIKI14.PAL", 0x400,	0x3eb23 },
	{ "ARCADE_SHIKI15.PAL", 0x400,	0x3eb24 },
	{ "ARCADE_SHIKI16.PAL", 0x400,	0x3eb25 },
	{ "ARCADE_SHIKI17.PAL", 0x400,	0x3eb26 },
	{ "ARCADE_SHIKI18.PAL", 0x400,	0x3eb27 },
	{ "ARCADE_SHIKI19.PAL", 0x400,	0x3eb28 },
	{ "ARCADE_SHIKI2.ARX", 0x1d4640,	0x3eb29 },
	{ "ARCADE_SHIKI2.PAL", 0x400,	0x3eed2 },
	{ "ARCADE_SHIKI20.PAL", 0x400,	0x3eed3 },
	{ "ARCADE_SHIKI21.PAL", 0x400,	0x3eed4 },
	{ "ARCADE_SHIKI22.PAL", 0x400,	0x3eed5 },
	{ "ARCADE_SHIKI23.PAL", 0x400,	0x3eed6 },
	{ "ARCADE_SHIKI24.PAL", 0x400,	0x3eed7 },
	{ "ARCADE_SHIKI25.PAL", 0x400,	0x3eed8 },
	{ "ARCADE_SHIKI26.PAL", 0x400,	0x3eed9 },
	{ "ARCADE_SHIKI27.PAL", 0x400,	0x3eeda },
	{ "ARCADE_SHIKI28.PAL", 0x400,	0x3eedb },
	{ "ARCADE_SHIKI29.PAL", 0x400,	0x3eedc },
	{ "ARCADE_SHIKI3.PAL", 0x400,	0x3eedd },
	{ "ARCADE_SHIKI30.PAL", 0x400,	0x3eede },
	{ "ARCADE_SHIKI31.PAL", 0x400,	0x3eedf },
	{ "ARCADE_SHIKI32.PAL", 0x400,	0x3eee0 },
	{ "ARCADE_SHIKI33.PAL", 0x400,	0x3eee1 },
	{ "ARCADE_SHIKI34.PAL", 0x400,	0x3eee2 },
	{ "ARCADE_SHIKI35.PAL", 0x400,	0x3eee3 },
	{ "ARCADE_SHIKI4.PAL", 0x400,	0x3eee4 },
	{ "ARCADE_SHIKI5.PAL", 0x400,	0x3eee5 },
	{ "ARCADE_SHIKI6.PAL", 0x400,	0x3eee6 },
	{ "ARCADE_SHIKI7.PAL", 0x400,	0x3eee7 },
	{ "ARCADE_SHIKI8.PAL", 0x400,	0x3eee8 },
	{ "ARCADE_SHIKI9.PAL", 0x400,	0x3eee9 },
	{ "ARCADE_SHIKIPAL2.ARX", 0x8d80,	0x3eeea },
	{ "ARCADE_SHIKI_0.HA6", 0x28d60,	0x3eefc },
	{ "ARCADE_SHIKI_0.TXT", 0xa6,	0x3ef4e },
	{ "ARCADE_SHIKI_0_C.TXT", 0x1396,	0x3ef4f },
	{ "ARCADE_SHIKI_1.HA6", 0xea50,	0x3ef52 },
	{ "ARCADE_SHIKI_1.TXT", 0xa6,	0x3ef70 },
	{ "ARCADE_SHIKI_1_C.TXT", 0x10cf,	0x3ef71 },
	{ "ARCADE_SHIKI_2.HA6", 0x16bf8,	0x3ef74 },
	{ "ARCADE_SHIKI_2.TXT", 0xa6,	0x3efa2 },
	{ "ARCADE_SHIKI_2_C.TXT", 0xfed,	0x3efa3 },
	{ "ARCADE_SHIKI_3.HA6", 0x1100c,	0x3efa5 },
	{ "ARCADE_SHIKI_3.TXT", 0xa6,	0x3efc8 },
	{ "ARCADE_SHIKI_3_C.TXT", 0xe65,	0x3efc9 },
	{ "ARCADE_SHIKI_9.HA6", 0x2425c,	0x3efcb },
	{ "ARCADE_SHIKI_9.TXT", 0xa6,	0x3f014 },
	{ "ARCADE_SHIKI_9_C.TXT", 0xe81,	0x3f015 },
	{ "ARCADE_SHIKI_C.TXT", 0x1017,	0x3f017 },
	{ "ARCADE_SHIKI_SELIST.TXT", 0xfed,	0x3f01a },
	{ "ARCADE_SION.ARX", 0x264b00,	0x3f01c },
	{ "ARCADE_SION.CG2", 0x5f278,	0x3f4e6 },
	{ "ARCADE_SION.CPF", 0x2f36c,	0x3f5a5 },
	{ "ARCADE_SION.HA6", 0x57a7c,	0x3f604 },
	{ "ARCADE_SION.PAT", 0x0,	0x3f6b4 },
	{ "ARCADE_SION.TXT", 0xd8,	0x3f6b4 },
	{ "ARCADE_SION0.PAL", 0x400,	0x3f6b5 },
	{ "ARCADE_SION1.PAL", 0x400,	0x3f6b6 },
	{ "ARCADE_SION10.PAL", 0x400,	0x3f6b7 },
	{ "ARCADE_SION11.PAL", 0x400,	0x3f6b8 },
	{ "ARCADE_SION12.PAL", 0x400,	0x3f6b9 },
	{ "ARCADE_SION13.PAL", 0x400,	0x3f6ba },
	{ "ARCADE_SION14.PAL", 0x400,	0x3f6bb },
	{ "ARCADE_SION15.PAL", 0x400,	0x3f6bc },
	{ "ARCADE_SION16.PAL", 0x400,	0x3f6bd },
	{ "ARCADE_SION17.PAL", 0x400,	0x3f6be },
	{ "ARCADE_SION18.PAL", 0x400,	0x3f6bf },
	{ "ARCADE_SION19.PAL", 0x400,	0x3f6c0 },
	{ "ARCADE_SION2.ARX", 0x263300,	0x3f6c1 },
	{ "ARCADE_SION2.PAL", 0x400,	0x3fb88 },
	{ "ARCADE_SION20.PAL", 0x400,	0x3fb89 },
	{ "ARCADE_SION21.PAL", 0x400,	0x3fb8a },
	{ "ARCADE_SION22.PAL", 0x400,	0x3fb8b },
	{ "ARCADE_SION23.PAL", 0x400,	0x3fb8c },
	{ "ARCADE_SION24.PAL", 0x400,	0x3fb8d },
	{ "ARCADE_SION25.PAL", 0x400,	0x3fb8e },
	{ "ARCADE_SION26.PAL", 0x400,	0x3fb8f },
	{ "ARCADE_SION27.PAL", 0x400,	0x3fb90 },
	{ "ARCADE_SION28.PAL", 0x400,	0x3fb91 },
	{ "ARCADE_SION29.PAL", 0x400,	0x3fb92 },
	{ "ARCADE_SION3.PAL", 0x400,	0x3fb93 },
	{ "ARCADE_SION30.PAL", 0x400,	0x3fb94 },
	{ "ARCADE_SION31.PAL", 0x400,	0x3fb95 },
	{ "ARCADE_SION32.PAL", 0x400,	0x3fb96 },
	{ "ARCADE_SION33.PAL", 0x400,	0x3fb97 },
	{ "ARCADE_SION34.PAL", 0x400,	0x3fb98 },
	{ "ARCADE_SION35.PAL", 0x400,	0x3fb99 },
	{ "ARCADE_SION4.PAL", 0x400,	0x3fb9a },
	{ "ARCADE_SION5.PAL", 0x400,	0x3fb9b },
	{ "ARCADE_SION6.PAL", 0x400,	0x3fb9c },
	{ "ARCADE_SION7.PAL", 0x400,	0x3fb9d },
	{ "ARCADE_SION8.PAL", 0x400,	0x3fb9e },
	{ "ARCADE_SION9.PAL", 0x400,	0x3fb9f },
	{ "ARCADE_SIONPAL2.ARX", 0x10a40,	0x3fba0 },
	{ "ARCADE_SION_0.HA6", 0xdacc,	0x3fbc2 },
	{ "ARCADE_SION_0.TXT", 0x98,	0x3fbde },
	{ "ARCADE_SION_0_C.TXT", 0x1571,	0x3fbdf },
	{ "ARCADE_SION_1.HA6", 0x121d4,	0x3fbe2 },
	{ "ARCADE_SION_1.TXT", 0x98,	0x3fc07 },
	{ "ARCADE_SION_1_C.TXT", 0x1339,	0x3fc08 },
	{ "ARCADE_SION_2.HA6", 0xbf28,	0x3fc0b },
	{ "ARCADE_SION_2.TXT", 0x9c,	0x3fc23 },
	{ "ARCADE_SION_2_C.TXT", 0x1316,	0x3fc24 },
	{ "ARCADE_SION_3.HA6", 0x16678,	0x3fc27 },
	{ "ARCADE_SION_3.TXT", 0x9c,	0x3fc54 },
	{ "ARCADE_SION_3_C.TXT", 0x1018,	0x3fc55 },
	{ "ARCADE_SION_9.HA6", 0x2b634,	0x3fc58 },
	{ "ARCADE_SION_9.TXT", 0x9c,	0x3fcaf },
	{ "ARCADE_SION_9_C.TXT", 0x1118,	0x3fcb0 },
	{ "ARCADE_SION_C.TXT", 0x118a,	0x3fcb3 },
	{ "ARCADE_SION_SELIST.TXT", 0xe7d,	0x3fcb6 },
	{ "ARCADE_SION_WIN.TXT", 0x942,	0x3fcb8 },
	{ "ARCADE_S_AKIHA.ARX", 0x368600,	0x3fcba },
	{ "ARCADE_S_AKIHA.CG2", 0x72a18,	0x4038b },
	{ "ARCADE_S_AKIHA.CPF", 0x2f36c,	0x40471 },
	{ "ARCADE_S_AKIHA.HA6", 0x7ae54,	0x404d0 },
	{ "ARCADE_S_AKIHA.PAT", 0x0,	0x405c6 },
	{ "ARCADE_S_AKIHA0.PAL", 0x400,	0x405c6 },
	{ "ARCADE_S_AKIHA1.PAL", 0x400,	0x405c7 },
	{ "ARCADE_S_AKIHA10.PAL", 0x400,	0x405c8 },
	{ "ARCADE_S_AKIHA11.PAL", 0x400,	0x405c9 },
	{ "ARCADE_S_AKIHA12.PAL", 0x400,	0x405ca },
	{ "ARCADE_S_AKIHA13.PAL", 0x400,	0x405cb },
	{ "ARCADE_S_AKIHA14.PAL", 0x400,	0x405cc },
	{ "ARCADE_S_AKIHA15.PAL", 0x400,	0x405cd },
	{ "ARCADE_S_AKIHA16.PAL", 0x400,	0x405ce },
	{ "ARCADE_S_AKIHA17.PAL", 0x400,	0x405cf },
	{ "ARCADE_S_AKIHA18.PAL", 0x400,	0x405d0 },
	{ "ARCADE_S_AKIHA19.PAL", 0x400,	0x405d1 },
	{ "ARCADE_S_AKIHA2.ARX", 0x3868c0,	0x405d2 },
	{ "ARCADE_S_AKIHA2.PAL", 0x400,	0x40ce0 },
	{ "ARCADE_S_AKIHA20.PAL", 0x400,	0x40ce1 },
	{ "ARCADE_S_AKIHA21.PAL", 0x400,	0x40ce2 },
	{ "ARCADE_S_AKIHA22.PAL", 0x400,	0x40ce3 },
	{ "ARCADE_S_AKIHA23.PAL", 0x400,	0x40ce4 },
	{ "ARCADE_S_AKIHA24.PAL", 0x400,	0x40ce5 },
	{ "ARCADE_S_AKIHA25.PAL", 0x400,	0x40ce6 },
	{ "ARCADE_S_AKIHA26.PAL", 0x400,	0x40ce7 },
	{ "ARCADE_S_AKIHA27.PAL", 0x400,	0x40ce8 },
	{ "ARCADE_S_AKIHA28.PAL", 0x400,	0x40ce9 },
	{ "ARCADE_S_AKIHA29.PAL", 0x400,	0x40cea },
	{ "ARCADE_S_AKIHA3.PAL", 0x400,	0x40ceb },
	{ "ARCADE_S_AKIHA30.PAL", 0x400,	0x40cec },
	{ "ARCADE_S_AKIHA31.PAL", 0x400,	0x40ced },
	{ "ARCADE_S_AKIHA32.PAL", 0x400,	0x40cee },
	{ "ARCADE_S_AKIHA33.PAL", 0x400,	0x40cef },
	{ "ARCADE_S_AKIHA34.PAL", 0x400,	0x40cf0 },
	{ "ARCADE_S_AKIHA35.PAL", 0x400,	0x40cf1 },
	{ "ARCADE_S_AKIHA4.PAL", 0x400,	0x40cf2 },
	{ "ARCADE_S_AKIHA5.PAL", 0x400,	0x40cf3 },
	{ "ARCADE_S_AKIHA6.PAL", 0x400,	0x40cf4 },
	{ "ARCADE_S_AKIHA7.PAL", 0x400,	0x40cf5 },
	{ "ARCADE_S_AKIHA8.PAL", 0x400,	0x40cf6 },
	{ "ARCADE_S_AKIHA9.PAL", 0x400,	0x40cf7 },
	{ "ARCADE_S_AKIHAPAL2.ARX", 0x20c00,	0x40cf8 },
	{ "ARCADE_S_AKIHA_0.HA6", 0x4b2c,	0x40d3a },
	{ "ARCADE_S_AKIHA_0.TXT", 0xb6,	0x40d44 },
	{ "ARCADE_S_AKIHA_0_C.TXT", 0x1a88,	0x40d45 },
	{ "ARCADE_S_AKIHA_1.HA6", 0x9d54,	0x40d49 },
	{ "ARCADE_S_AKIHA_1.TXT", 0xb6,	0x40d5d },
	{ "ARCADE_S_AKIHA_1_C.TXT", 0x1452,	0x40d5e },
	{ "ARCADE_S_AKIHA_2.HA6", 0x9bd4,	0x40d61 },
	{ "ARCADE_S_AKIHA_2.TXT", 0xb6,	0x40d75 },
	{ "ARCADE_S_AKIHA_2_C.TXT", 0x13f5,	0x40d76 },
	{ "ARCADE_S_AKIHA_SELIST.TXT", 0x1338,	0x40d79 },
	{ "ARCADE_TEMPLATE.CPF", 0x2f36c,	0x40d7c },
	{ "ARCADE_VECTOR.TXT", 0x196f,	0x40dfc },
	{ "ARCADE_VECTORLIST.TXT", 0x30d5,	0x40e00 },
	{ "ARCADE_VECTORSAMPLE.TXT", 0x9c0,	0x40e07 },
	{ "ARCADE_V_SION.ARX", 0x3bea40,	0x40e09 },
	{ "ARCADE_V_SION.CG2", 0x7a0e0,	0x41587 },
	{ "ARCADE_V_SION.CPF", 0x2f36c,	0x4167c },
	{ "ARCADE_V_SION.HA6", 0x68bd8,	0x416db },
	{ "ARCADE_V_SION.PAT", 0xbe38,	0x417ad },
	{ "ARCADE_V_SION.TXT", 0xe1,	0x417c5 },
	{ "ARCADE_V_SION0.PAL", 0x400,	0x417c6 },
	{ "ARCADE_V_SION1.PAL", 0x400,	0x417c7 },
	{ "ARCADE_V_SION10.PAL", 0x400,	0x417c8 },
	{ "ARCADE_V_SION11.PAL", 0x400,	0x417c9 },
	{ "ARCADE_V_SION12.PAL", 0x400,	0x417ca },
	{ "ARCADE_V_SION13.PAL", 0x400,	0x417cb },
	{ "ARCADE_V_SION14.PAL", 0x400,	0x417cc },
	{ "ARCADE_V_SION15.PAL", 0x400,	0x417cd },
	{ "ARCADE_V_SION16.PAL", 0x400,	0x417ce },
	{ "ARCADE_V_SION17.PAL", 0x400,	0x417cf },
	{ "ARCADE_V_SION18.PAL", 0x400,	0x417d0 },
	{ "ARCADE_V_SION19.PAL", 0x400,	0x417d1 },
	{ "ARCADE_V_SION2.ARX", 0x3b9d00,	0x417d2 },
	{ "ARCADE_V_SION2.PAL", 0x400,	0x41f46 },
	{ "ARCADE_V_SION20.PAL", 0x400,	0x41f47 },
	{ "ARCADE_V_SION21.PAL", 0x400,	0x41f48 },
	{ "ARCADE_V_SION22.PAL", 0x400,	0x41f49 },
	{ "ARCADE_V_SION23.PAL", 0x400,	0x41f4a },
	{ "ARCADE_V_SION24.PAL", 0x400,	0x41f4b },
	{ "ARCADE_V_SION25.PAL", 0x400,	0x41f4c },
	{ "ARCADE_V_SION26.PAL", 0x400,	0x41f4d },
	{ "ARCADE_V_SION27.PAL", 0x400,	0x41f4e },
	{ "ARCADE_V_SION28.PAL", 0x400,	0x41f4f },
	{ "ARCADE_V_SION29.PAL", 0x400,	0x41f50 },
	{ "ARCADE_V_SION3.PAL", 0x400,	0x41f51 },
	{ "ARCADE_V_SION30.PAL", 0x400,	0x41f52 },
	{ "ARCADE_V_SION31.PAL", 0x400,	0x41f53 },
	{ "ARCADE_V_SION32.PAL", 0x400,	0x41f54 },
	{ "ARCADE_V_SION33.PAL", 0x400,	0x41f55 },
	{ "ARCADE_V_SION34.PAL", 0x400,	0x41f56 },
	{ "ARCADE_V_SION35.PAL", 0x400,	0x41f57 },
	{ "ARCADE_V_SION4.PAL", 0x400,	0x41f58 },
	{ "ARCADE_V_SION5.PAL", 0x400,	0x41f59 },
	{ "ARCADE_V_SION6.PAL", 0x400,	0x41f5a },
	{ "ARCADE_V_SION7.PAL", 0x400,	0x41f5b },
	{ "ARCADE_V_SION8.PAL", 0x400,	0x41f5c },
	{ "ARCADE_V_SION9.PAL", 0x400,	0x41f5d },
	{ "ARCADE_V_SIONPAL2.ARX", 0x19bc0,	0x41f5e },
	{ "ARCADE_V_SION_0.HA6", 0xa5ec,	0x41f92 },
	{ "ARCADE_V_SION_0.TXT", 0xb5,	0x41fa7 },
	{ "ARCADE_V_SION_0_C.TXT", 0x1692,	0x41fa8 },
	{ "ARCADE_V_SION_1.HA6", 0x13c8c,	0x41fab },
	{ "ARCADE_V_SION_1.TXT", 0xc2,	0x41fd3 },
	{ "ARCADE_V_SION_1_C.TXT", 0x174a,	0x41fd4 },
	{ "ARCADE_V_SION_2.HA6", 0x8dac,	0x41fd7 },
	{ "ARCADE_V_SION_2.TXT", 0xb7,	0x41fe9 },
	{ "ARCADE_V_SION_2_C.TXT", 0x1305,	0x41fea },
	{ "ARCADE_V_SION_3.HA6", 0x1100c,	0x41fed },
	{ "ARCADE_V_SION_3.TXT", 0xb7,	0x42010 },
	{ "ARCADE_V_SION_3_C.TXT", 0x1017,	0x42011 },
	{ "ARCADE_V_SION_9.HA6", 0x214d8,	0x42014 },
	{ "ARCADE_V_SION_9.TXT", 0xb7,	0x42057 },
	{ "ARCADE_V_SION_9_C.TXT", 0x1042,	0x42058 },
	{ "ARCADE_V_SION_C.TXT", 0x1200,	0x4205b },
	{ "ARCADE_V_SION_SELIST.TXT", 0xf25,	0x4205e },
	{ "ARCADE_V_SION_WIN.TXT", 0x556,	0x42060 },
	{ "ARCADE_WARAKIA.ARX", 0x34d780,	0x42061 },
	{ "ARCADE_WARAKIA.CG2", 0x72e38,	0x426fc },
	{ "ARCADE_WARAKIA.CPF", 0x2f36c,	0x427e2 },
	{ "ARCADE_WARAKIA.HA6", 0x67d60,	0x42841 },
	{ "ARCADE_WARAKIA.PAT", 0xc39a,	0x42911 },
	{ "ARCADE_WARAKIA.TXT", 0xe8,	0x4292a },
	{ "ARCADE_WARAKIA0.PAL", 0x400,	0x4292b },
	{ "ARCADE_WARAKIA1.PAL", 0x400,	0x4292c },
	{ "ARCADE_WARAKIA10.PAL", 0x400,	0x4292d },
	{ "ARCADE_WARAKIA11.PAL", 0x400,	0x4292e },
	{ "ARCADE_WARAKIA12.PAL", 0x400,	0x4292f },
	{ "ARCADE_WARAKIA13.PAL", 0x400,	0x42930 },
	{ "ARCADE_WARAKIA14.PAL", 0x400,	0x42931 },
	{ "ARCADE_WARAKIA15.PAL", 0x400,	0x42932 },
	{ "ARCADE_WARAKIA16.PAL", 0x400,	0x42933 },
	{ "ARCADE_WARAKIA17.PAL", 0x400,	0x42934 },
	{ "ARCADE_WARAKIA18.PAL", 0x400,	0x42935 },
	{ "ARCADE_WARAKIA19.PAL", 0x400,	0x42936 },
	{ "ARCADE_WARAKIA2.ARX", 0x339cc0,	0x42937 },
	{ "ARCADE_WARAKIA2.PAL", 0x400,	0x42fab },
	{ "ARCADE_WARAKIA20.PAL", 0x400,	0x42fac },
	{ "ARCADE_WARAKIA21.PAL", 0x400,	0x42fad },
	{ "ARCADE_WARAKIA22.PAL", 0x400,	0x42fae },
	{ "ARCADE_WARAKIA23.PAL", 0x400,	0x42faf },
	{ "ARCADE_WARAKIA24.PAL", 0x400,	0x42fb0 },
	{ "ARCADE_WARAKIA25.PAL", 0x400,	0x42fb1 },
	{ "ARCADE_WARAKIA26.PAL", 0x400,	0x42fb2 },
	{ "ARCADE_WARAKIA27.PAL", 0x400,	0x42fb3 },
	{ "ARCADE_WARAKIA28.PAL", 0x400,	0x42fb4 },
	{ "ARCADE_WARAKIA29.PAL", 0x400,	0x42fb5 },
	{ "ARCADE_WARAKIA3.PAL", 0x400,	0x42fb6 },
	{ "ARCADE_WARAKIA30.PAL", 0x400,	0x42fb7 },
	{ "ARCADE_WARAKIA31.PAL", 0x400,	0x42fb8 },
	{ "ARCADE_WARAKIA32.PAL", 0x400,	0x42fb9 },
	{ "ARCADE_WARAKIA33.PAL", 0x400,	0x42fba },
	{ "ARCADE_WARAKIA34.PAL", 0x400,	0x42fbb },
	{ "ARCADE_WARAKIA35.PAL", 0x400,	0x42fbc },
	{ "ARCADE_WARAKIA4.PAL", 0x400,	0x42fbd },
	{ "ARCADE_WARAKIA5.PAL", 0x400,	0x42fbe },
	{ "ARCADE_WARAKIA6.PAL", 0x400,	0x42fbf },
	{ "ARCADE_WARAKIA7.PAL", 0x400,	0x42fc0 },
	{ "ARCADE_WARAKIA8.PAL", 0x400,	0x42fc1 },
	{ "ARCADE_WARAKIA9.PAL", 0x400,	0x42fc2 },
	{ "ARCADE_WARAKIAPAL2.ARX", 0x1c140,	0x42fc3 },
	{ "ARCADE_WARAKIA_0.HA6", 0x9358,	0x42ffc },
	{ "ARCADE_WARAKIA_0.TXT", 0xbc,	0x4300f },
	{ "ARCADE_WARAKIA_0_C.TXT", 0x1636,	0x43010 },
	{ "ARCADE_WARAKIA_1.HA6", 0x163c8,	0x43013 },
	{ "ARCADE_WARAKIA_1.TXT", 0xc0,	0x43040 },
	{ "ARCADE_WARAKIA_1_C.TXT", 0x1256,	0x43041 },
	{ "ARCADE_WARAKIA_2.HA6", 0xa86c,	0x43044 },
	{ "ARCADE_WARAKIA_2.TXT", 0xbc,	0x4305a },
	{ "ARCADE_WARAKIA_2_C.TXT", 0x11d3,	0x4305b },
	{ "ARCADE_WARAKIA_3.HA6", 0x1100c,	0x4305e },
	{ "ARCADE_WARAKIA_3.TXT", 0xbc,	0x43081 },
	{ "ARCADE_WARAKIA_3_C.TXT", 0xeed,	0x43082 },
	{ "ARCADE_WARAKIA_9.HA6", 0x16588,	0x43084 },
	{ "ARCADE_WARAKIA_9.TXT", 0xbc,	0x430b1 },
	{ "ARCADE_WARAKIA_9_C.TXT", 0x1014,	0x430b2 },
	{ "ARCADE_WARAKIA_C.TXT", 0x10d9,	0x430b5 },
	{ "ARCADE_WARAKIA_SELIST.TXT", 0xdaf,	0x430b8 },
	{ "ARCADE_WARC.ARX", 0x36fb40,	0x430db },
	{ "ARCADE_WARC.CG2", 0x76b28,	0x437bb },
	{ "ARCADE_WARC.CPF", 0x2f36c,	0x438a9 },
	{ "ARCADE_WARC.HA6", 0x7a4c4,	0x43908 },
	{ "ARCADE_WARC.PAT", 0xef12,	0x439fd },
	{ "ARCADE_WARC.TXT", 0xd3,	0x43a1b },
	{ "ARCADE_WARC0.PAL", 0x400,	0x43a1c },
	{ "ARCADE_WARC1.PAL", 0x400,	0x43a1d },
	{ "ARCADE_WARC10.PAL", 0x400,	0x43a1e },
	{ "ARCADE_WARC11.PAL", 0x400,	0x43a1f },
	{ "ARCADE_WARC12.PAL", 0x400,	0x43a20 },
	{ "ARCADE_WARC13.PAL", 0x400,	0x43a21 },
	{ "ARCADE_WARC14.PAL", 0x400,	0x43a22 },
	{ "ARCADE_WARC15.PAL", 0x400,	0x43a23 },
	{ "ARCADE_WARC16.PAL", 0x400,	0x43a24 },
	{ "ARCADE_WARC17.PAL", 0x400,	0x43a25 },
	{ "ARCADE_WARC18.PAL", 0x400,	0x43a26 },
	{ "ARCADE_WARC19.PAL", 0x400,	0x43a27 },
	{ "ARCADE_WARC2.ARX", 0x361800,	0x43a28 },
	{ "ARCADE_WARC2.PAL", 0x400,	0x440eb },
	{ "ARCADE_WARC20.PAL", 0x400,	0x440ec },
	{ "ARCADE_WARC21.PAL", 0x400,	0x440ed },
	{ "ARCADE_WARC22.PAL", 0x400,	0x440ee },
	{ "ARCADE_WARC23.PAL", 0x400,	0x440ef },
	{ "ARCADE_WARC24.PAL", 0x400,	0x440f0 },
	{ "ARCADE_WARC25.PAL", 0x400,	0x440f1 },
	{ "ARCADE_WARC26.PAL", 0x400,	0x440f2 },
	{ "ARCADE_WARC27.PAL", 0x400,	0x440f3 },
	{ "ARCADE_WARC28.PAL", 0x400,	0x440f4 },
	{ "ARCADE_WARC29.PAL", 0x400,	0x440f5 },
	{ "ARCADE_WARC3.PAL", 0x400,	0x440f6 },
	{ "ARCADE_WARC30.PAL", 0x400,	0x440f7 },
	{ "ARCADE_WARC31.PAL", 0x400,	0x440f8 },
	{ "ARCADE_WARC32.PAL", 0x400,	0x440f9 },
	{ "ARCADE_WARC33.PAL", 0x400,	0x440fa },
	{ "ARCADE_WARC34.PAL", 0x400,	0x440fb },
	{ "ARCADE_WARC35.PAL", 0x400,	0x440fc },
	{ "ARCADE_WARC4.PAL", 0x400,	0x440fd },
	{ "ARCADE_WARC5.PAL", 0x400,	0x440fe },
	{ "ARCADE_WARC6.PAL", 0x400,	0x440ff },
	{ "ARCADE_WARC7.PAL", 0x400,	0x44100 },
	{ "ARCADE_WARC8.PAL", 0x400,	0x44101 },
	{ "ARCADE_WARC9.PAL", 0x400,	0x44102 },
	{ "ARCADE_WARCPAL2.ARX", 0x20c00,	0x44103 },
	{ "ARCADE_WARC_0.HA6", 0xaee0,	0x44145 },
	{ "ARCADE_WARC_0.TXT", 0xab,	0x4415b },
	{ "ARCADE_WARC_0_C.TXT", 0x15d1,	0x4415c },
	{ "ARCADE_WARC_1.HA6", 0x11410,	0x4415f },
	{ "ARCADE_WARC_1.TXT", 0xab,	0x44182 },
	{ "ARCADE_WARC_1_C.TXT", 0x12b7,	0x44183 },
	{ "ARCADE_WARC_2.HA6", 0xb6dc,	0x44186 },
	{ "ARCADE_WARC_2.TXT", 0xad,	0x4419d },
	{ "ARCADE_WARC_2_C.TXT", 0x1480,	0x4419e },
	{ "ARCADE_WARC_3.HA6", 0x1100c,	0x441a1 },
	{ "ARCADE_WARC_3.TXT", 0xad,	0x441c4 },
	{ "ARCADE_WARC_3_C.TXT", 0x1292,	0x441c5 },
	{ "ARCADE_WARC_9.HA6", 0x46fb8,	0x441c8 },
	{ "ARCADE_WARC_9.TXT", 0xad,	0x44256 },
	{ "ARCADE_WARC_9_C.TXT", 0x1279,	0x44257 },
	{ "ARCADE_WARC_C.TXT", 0x1706,	0x4425a },
	{ "ARCADE_WARC_SELIST.TXT", 0xc14,	0x4425d },
	{ "ARCADE_WLEN.ARX", 0x429300,	0x44280 },
	{ "ARCADE_WLEN.CG2", 0x88000,	0x44ad3 },
	{ "ARCADE_WLEN.CPF", 0x2f36c,	0x44be3 },
	{ "ARCADE_WLEN.HA6", 0x739d4,	0x44c42 },
	{ "ARCADE_WLEN.PAL", 0x10004,	0x44d2a },
	{ "ARCADE_WLEN.PAT", 0x0,	0x44d4b },
	{ "ARCADE_WLEN.TXT", 0xc2,	0x44d4b },
	{ "ARCADE_WLEN0.PAL", 0x400,	0x44d4c },
	{ "ARCADE_WLEN1.PAL", 0x400,	0x44d4d },
	{ "ARCADE_WLEN10.PAL", 0x400,	0x44d4e },
	{ "ARCADE_WLEN11.PAL", 0x400,	0x44d4f },
	{ "ARCADE_WLEN12.PAL", 0x400,	0x44d50 },
	{ "ARCADE_WLEN13.PAL", 0x400,	0x44d51 },
	{ "ARCADE_WLEN14.PAL", 0x400,	0x44d52 },
	{ "ARCADE_WLEN15.PAL", 0x400,	0x44d53 },
	{ "ARCADE_WLEN16.PAL", 0x400,	0x44d54 },
	{ "ARCADE_WLEN17.PAL", 0x400,	0x44d55 },
	{ "ARCADE_WLEN18.PAL", 0x400,	0x44d56 },
	{ "ARCADE_WLEN19.PAL", 0x400,	0x44d57 },
	{ "ARCADE_WLEN2.ARX", 0x423500,	0x44d58 },
	{ "ARCADE_WLEN2.PAL", 0x400,	0x4559f },
	{ "ARCADE_WLEN20.PAL", 0x400,	0x455a0 },
	{ "ARCADE_WLEN21.PAL", 0x400,	0x455a1 },
	{ "ARCADE_WLEN22.PAL", 0x400,	0x455a2 },
	{ "ARCADE_WLEN23.PAL", 0x400,	0x455a3 },
	{ "ARCADE_WLEN24.PAL", 0x400,	0x455a4 },
	{ "ARCADE_WLEN25.PAL", 0x400,	0x455a5 },
	{ "ARCADE_WLEN26.PAL", 0x400,	0x455a6 },
	{ "ARCADE_WLEN27.PAL", 0x400,	0x455a7 },
	{ "ARCADE_WLEN28.PAL", 0x400,	0x455a8 },
	{ "ARCADE_WLEN29.PAL", 0x400,	0x455a9 },
	{ "ARCADE_WLEN3.PAL", 0x400,	0x455aa },
	{ "ARCADE_WLEN30.PAL", 0x400,	0x455ab },
	{ "ARCADE_WLEN31.PAL", 0x400,	0x455ac },
	{ "ARCADE_WLEN32.PAL", 0x400,	0x455ad },
	{ "ARCADE_WLEN33.PAL", 0x400,	0x455ae },
	{ "ARCADE_WLEN34.PAL", 0x400,	0x455af },
	{ "ARCADE_WLEN35.PAL", 0x400,	0x455b0 },
	{ "ARCADE_WLEN4.PAL", 0x400,	0x455b1 },
	{ "ARCADE_WLEN5.PAL", 0x400,	0x455b2 },
	{ "ARCADE_WLEN6.PAL", 0x400,	0x455b3 },
	{ "ARCADE_WLEN7.PAL", 0x400,	0x455b4 },
	{ "ARCADE_WLEN8.PAL", 0x400,	0x455b5 },
	{ "ARCADE_WLEN9.PAL", 0x400,	0x455b6 },
	{ "ARCADE_WLENPAL2.ARX", 0x2bec0,	0x455b7 },
	{ "ARCADE_WLEN_0.HA6", 0xdbfc,	0x4560f },
	{ "ARCADE_WLEN_0.TXT", 0x9c,	0x4562b },
	{ "ARCADE_WLEN_0_C.TXT", 0x1788,	0x4562c },
	{ "ARCADE_WLEN_1.HA6", 0x14200,	0x4562f },
	{ "ARCADE_WLEN_1.TXT", 0x9c,	0x45658 },
	{ "ARCADE_WLEN_1_C.TXT", 0x133b,	0x45659 },
	{ "ARCADE_WLEN_2.HA6", 0x14914,	0x4565c },
	{ "ARCADE_WLEN_2.TXT", 0x9c,	0x45686 },
	{ "ARCADE_WLEN_2_C.TXT", 0x1510,	0x45687 },
	{ "ARCADE_WLEN_3.HA6", 0x1100c,	0x4568a },
	{ "ARCADE_WLEN_3.TXT", 0x9c,	0x456ad },
	{ "ARCADE_WLEN_3_C.TXT", 0x1238,	0x456ae },
	{ "ARCADE_WLEN_9.HA6", 0x15f68,	0x456b1 },
	{ "ARCADE_WLEN_9.TXT", 0x9c,	0x456dd },
	{ "ARCADE_WLEN_9_C.TXT", 0x12f2,	0x456de },
	{ "ARCADE_WLEN_C.TXT", 0x16d8,	0x456e1 },
	{ "ARCADE_WLEN_SELIST.TXT", 0x1377,	0x456e4 },
	{ "ARCADE__EFFECT.TXT", 0xba,	0x456e7 },
	{ "ARCADE__TYPE.TXT", 0x26f5,	0x456e8 },
	{ "ARCADE___AKAAKIHA.TXT", 0xf1,	0x456ed },
	{ "ARCADE___AKIHA.TXT", 0xdc,	0x456ee },
	{ "ARCADE___AOKO.TXT", 0xd5,	0x456ef },
	{ "ARCADE___ARC.TXT", 0xde,	0x456f0 },
	{ "ARCADE___CIEL.TXT", 0xd6,	0x456f1 },
	{ "ARCADE___HISUI.TXT", 0xf2,	0x456f2 },
	{ "ARCADE___KISHIMA.TXT", 0xea,	0x456f3 },
	{ "ARCADE___KOHAKU.TXT", 0xf6,	0x456f4 },
	{ "ARCADE___LEN.TXT", 0xdc,	0x456f5 },
	{ "ARCADE___MIYAKO.TXT", 0xf4,	0x456f6 },
	{ "ARCADE___M_HISUI.TXT", 0xfe,	0x456f7 },
	{ "ARCADE___NANAYA.TXT", 0xe1,	0x456f8 },
	{ "ARCADE___NECO.TXT", 0xe6,	0x456f9 },
	{ "ARCADE___NERO.TXT", 0xd7,	0x456fa },
	{ "ARCADE___RIES.TXT", 0xe6,	0x456fb },
	{ "ARCADE___ROA.TXT", 0xcc,	0x456fc },
	{ "ARCADE___SATSUKI.TXT", 0xec,	0x456fd },
	{ "ARCADE___SHIKI.TXT", 0xe0,	0x456fe },
	{ "ARCADE___SION.TXT", 0xed,	0x456ff },
	{ "ARCADE___V_SION.TXT", 0xf6,	0x45700 },
	{ "ARCADE___WARAKIA.TXT", 0x100,	0x45701 },
	{ "ARCADE___WARC.TXT", 0xe6,	0x45702 },
	{ "ARCADE___WLEN.TXT", 0xd5,	0x45703 },
	{ "ACTION01.CPF", 0x2f36c,	0x45704 },
	{ "ACTION02.CPF", 0x2f36c,	0x45763 },
	{ "AKAAKIHA.CG2", 0x85678,	0x457c2 },
	{ "AKAAKIHA.CPF", 0x2f36c,	0x458cd },
	{ "AKAAKIHA.DOF", 0xd8a24,	0x4592c },
	{ "AKAAKIHA.HA6", 0xa16e8,	0x45ade },
	{ "AKAAKIHA.PAT", 0x0,	0x45c21 },
	{ "AKAAKIHA.TXT", 0xda,	0x45c21 },
	{ "AKAAKIHA0.PAL", 0x400,	0x45c22 },
	{ "AKAAKIHA1.PAL", 0x400,	0x45c23 },
	{ "AKAAKIHA10.PAL", 0x400,	0x45c24 },
	{ "AKAAKIHA11.PAL", 0x400,	0x45c25 },
	{ "AKAAKIHA12.PAL", 0x400,	0x45c26 },
	{ "AKAAKIHA13.PAL", 0x400,	0x45c27 },
	{ "AKAAKIHA14.PAL", 0x400,	0x45c28 },
	{ "AKAAKIHA15.PAL", 0x400,	0x45c29 },
	{ "AKAAKIHA16.PAL", 0x400,	0x45c2a },
	{ "AKAAKIHA17.PAL", 0x400,	0x45c2b },
	{ "AKAAKIHA18.PAL", 0x400,	0x45c2c },
	{ "AKAAKIHA19.PAL", 0x400,	0x45c2d },
	{ "AKAAKIHA2.ARX", 0x3b61c0,	0x45c2e },
	{ "AKAAKIHA2.PAL", 0x400,	0x4639b },
	{ "AKAAKIHA20.PAL", 0x400,	0x4639c },
	{ "AKAAKIHA21.PAL", 0x400,	0x4639d },
	{ "AKAAKIHA22.PAL", 0x400,	0x4639e },
	{ "AKAAKIHA23.PAL", 0x400,	0x4639f },
	{ "AKAAKIHA24.PAL", 0x400,	0x463a0 },
	{ "AKAAKIHA25.PAL", 0x400,	0x463a1 },
	{ "AKAAKIHA26.PAL", 0x400,	0x463a2 },
	{ "AKAAKIHA27.PAL", 0x400,	0x463a3 },
	{ "AKAAKIHA28.PAL", 0x400,	0x463a4 },
	{ "AKAAKIHA29.PAL", 0x400,	0x463a5 },
	{ "AKAAKIHA3.PAL", 0x400,	0x463a6 },
	{ "AKAAKIHA30.PAL", 0x400,	0x463a7 },
	{ "AKAAKIHA31.PAL", 0x400,	0x463a8 },
	{ "AKAAKIHA32.PAL", 0x400,	0x463a9 },
	{ "AKAAKIHA33.PAL", 0x400,	0x463aa },
	{ "AKAAKIHA34.PAL", 0x400,	0x463ab },
	{ "AKAAKIHA35.PAL", 0x400,	0x463ac },
	{ "AKAAKIHA4.PAL", 0x400,	0x463ad },
	{ "AKAAKIHA5.PAL", 0x400,	0x463ae },
	{ "AKAAKIHA6.PAL", 0x400,	0x463af },
	{ "AKAAKIHA7.PAL", 0x400,	0x463b0 },
	{ "AKAAKIHA8.PAL", 0x400,	0x463b1 },
	{ "AKAAKIHA9.PAL", 0x400,	0x463b2 },
	{ "AKAAKIHAPAL2.ARX", 0x22d40,	0x463b3 },
	{ "AKAAKIHA_0.HA6", 0x2f34,	0x463f9 },
	{ "AKAAKIHA_0.TXT", 0xac,	0x463ff },
	{ "AKAAKIHA_0_C.TXT", 0x14ba,	0x46400 },
	{ "AKAAKIHA_1.HA6", 0xb9f8,	0x46403 },
	{ "AKAAKIHA_1.TXT", 0xac,	0x4641b },
	{ "AKAAKIHA_1_C.TXT", 0x1540,	0x4641c },
	{ "AKAAKIHA_2.HA6", 0x148a8,	0x4641f },
	{ "AKAAKIHA_2.TXT", 0xac,	0x46449 },
	{ "AKAAKIHA_2_C.TXT", 0x1396,	0x4644a },
	{ "AKAAKIHA_3.HA6", 0x1100c,	0x4644d },
	{ "AKAAKIHA_9.HA6", 0x13b68,	0x46470 },
	{ "AKAAKIHA_9.TXT", 0xac,	0x46498 },
	{ "AKAAKIHA_9_C.TXT", 0x125a,	0x46499 },
	{ "AKAAKIHA_SELIST.TXT", 0xf04,	0x4649c },
	{ "AKIHA.CG2", 0x74e18,	0x4649e },
	{ "AKIHA.CPF", 0x2f36c,	0x46588 },
	{ "AKIHA.DOF", 0xdac40,	0x465e7 },
	{ "AKIHA.HA6", 0x7f3b0,	0x4679d },
	{ "AKIHA.LST", 0xb208,	0x4689c },
	{ "AKIHA.PAT", 0x0,	0x468b3 },
	{ "AKIHA.TXT", 0xc8,	0x468b3 },
	{ "AKIHA0.PAL", 0x400,	0x468b4 },
	{ "AKIHA1.PAL", 0x400,	0x468b5 },
	{ "AKIHA10.PAL", 0x400,	0x468b6 },
	{ "AKIHA11.PAL", 0x400,	0x468b7 },
	{ "AKIHA12.PAL", 0x400,	0x468b8 },
	{ "AKIHA13.PAL", 0x400,	0x468b9 },
	{ "AKIHA14.PAL", 0x400,	0x468ba },
	{ "AKIHA15.PAL", 0x400,	0x468bb },
	{ "AKIHA16.PAL", 0x400,	0x468bc },
	{ "AKIHA17.PAL", 0x400,	0x468bd },
	{ "AKIHA18.PAL", 0x400,	0x468be },
	{ "AKIHA19.PAL", 0x400,	0x468bf },
	{ "AKIHA2.ARX", 0x345f40,	0x468c0 },
	{ "AKIHA2.PAL", 0x400,	0x46f4c },
	{ "AKIHA20.PAL", 0x400,	0x46f4d },
	{ "AKIHA21.PAL", 0x400,	0x46f4e },
	{ "AKIHA22.PAL", 0x400,	0x46f4f },
	{ "AKIHA23.PAL", 0x400,	0x46f50 },
	{ "AKIHA24.PAL", 0x400,	0x46f51 },
	{ "AKIHA25.PAL", 0x400,	0x46f52 },
	{ "AKIHA26.PAL", 0x400,	0x46f53 },
	{ "AKIHA27.PAL", 0x400,	0x46f54 },
	{ "AKIHA28.PAL", 0x400,	0x46f55 },
	{ "AKIHA29.PAL", 0x400,	0x46f56 },
	{ "AKIHA3.PAL", 0x400,	0x46f57 },
	{ "AKIHA30.PAL", 0x400,	0x46f58 },
	{ "AKIHA31.PAL", 0x400,	0x46f59 },
	{ "AKIHA32.PAL", 0x400,	0x46f5a },
	{ "AKIHA33.PAL", 0x400,	0x46f5b },
	{ "AKIHA34.PAL", 0x400,	0x46f5c },
	{ "AKIHA35.PAL", 0x400,	0x46f5d },
	{ "AKIHA4.PAL", 0x400,	0x46f5e },
	{ "AKIHA5.PAL", 0x400,	0x46f5f },
	{ "AKIHA6.PAL", 0x400,	0x46f60 },
	{ "AKIHA7.PAL", 0x400,	0x46f61 },
	{ "AKIHA8.PAL", 0x400,	0x46f62 },
	{ "AKIHA9.PAL", 0x400,	0x46f63 },
	{ "AKIHAPAL2.ARX", 0x1d1c0,	0x46f64 },
	{ "AKIHA_0.HA6", 0x6e0c,	0x46f9f },
	{ "AKIHA_0.LST", 0xb208,	0x46fad },
	{ "AKIHA_0.TXT", 0xa0,	0x46fc4 },
	{ "AKIHA_0_C.TXT", 0x15f7,	0x46fc5 },
	{ "AKIHA_1.HA6", 0x19380,	0x46fc8 },
	{ "AKIHA_1.LST", 0xb208,	0x46ffb },
	{ "AKIHA_1.TXT", 0xa0,	0x47012 },
	{ "AKIHA_1_C.TXT", 0x1494,	0x47013 },
	{ "AKIHA_2.HA6", 0x18ba0,	0x47016 },
	{ "AKIHA_2.LST", 0xb208,	0x47048 },
	{ "AKIHA_2.TXT", 0xa0,	0x4705f },
	{ "AKIHA_2_C.TXT", 0x14aa,	0x47060 },
	{ "AKIHA_3.HA6", 0x1100c,	0x47063 },
	{ "AKIHA_3.TXT", 0xa0,	0x47086 },
	{ "AKIHA_3_C.TXT", 0x12ba,	0x47087 },
	{ "AKIHA_9.HA6", 0x1bae8,	0x4708a },
	{ "AKIHA_9.TXT", 0xa0,	0x470c2 },
	{ "AKIHA_9_C.TXT", 0x1448,	0x470c3 },
	{ "AKIHA_SELIST.TXT", 0xc49,	0x470c6 },
	{ "AOKO.CG2", 0x6db28,	0x470c8 },
	{ "AOKO.CPF", 0x2f36c,	0x471a4 },
	{ "AOKO.DOF", 0xbcbfc,	0x47203 },
	{ "AOKO.HA6", 0x6f048,	0x4737d },
	{ "AOKO.PAT", 0x0,	0x4745c },
	{ "AOKO.TXT", 0xc2,	0x4745c },
	{ "AOKO0.PAL", 0x400,	0x4745d },
	{ "AOKO1.PAL", 0x400,	0x4745e },
	{ "AOKO10.PAL", 0x400,	0x4745f },
	{ "AOKO11.PAL", 0x400,	0x47460 },
	{ "AOKO12.PAL", 0x400,	0x47461 },
	{ "AOKO13.PAL", 0x400,	0x47462 },
	{ "AOKO14.PAL", 0x400,	0x47463 },
	{ "AOKO15.PAL", 0x400,	0x47464 },
	{ "AOKO16.PAL", 0x400,	0x47465 },
	{ "AOKO17.PAL", 0x400,	0x47466 },
	{ "AOKO18.PAL", 0x400,	0x47467 },
	{ "AOKO19.PAL", 0x400,	0x47468 },
	{ "AOKO2.ARX", 0x34ec80,	0x47469 },
	{ "AOKO2.PAL", 0x400,	0x47b07 },
	{ "AOKO20.PAL", 0x400,	0x47b08 },
	{ "AOKO21.PAL", 0x400,	0x47b09 },
	{ "AOKO22.PAL", 0x400,	0x47b0a },
	{ "AOKO23.PAL", 0x400,	0x47b0b },
	{ "AOKO24.PAL", 0x400,	0x47b0c },
	{ "AOKO25.PAL", 0x400,	0x47b0d },
	{ "AOKO26.PAL", 0x400,	0x47b0e },
	{ "AOKO27.PAL", 0x400,	0x47b0f },
	{ "AOKO28.PAL", 0x400,	0x47b10 },
	{ "AOKO29.PAL", 0x400,	0x47b11 },
	{ "AOKO3.PAL", 0x400,	0x47b12 },
	{ "AOKO30.PAL", 0x400,	0x47b13 },
	{ "AOKO31.PAL", 0x400,	0x47b14 },
	{ "AOKO32.PAL", 0x400,	0x47b15 },
	{ "AOKO33.PAL", 0x400,	0x47b16 },
	{ "AOKO34.PAL", 0x400,	0x47b17 },
	{ "AOKO35.PAL", 0x400,	0x47b18 },
	{ "AOKO4.PAL", 0x400,	0x47b19 },
	{ "AOKO5.PAL", 0x400,	0x47b1a },
	{ "AOKO6.PAL", 0x400,	0x47b1b },
	{ "AOKO7.PAL", 0x400,	0x47b1c },
	{ "AOKO8.PAL", 0x400,	0x47b1d },
	{ "AOKO9.PAL", 0x400,	0x47b1e },
	{ "AOKOPAL2.ARX", 0x1c980,	0x47b1f },
	{ "AOKO_0.HA6", 0x1f638,	0x47b59 },
	{ "AOKO_0.TXT", 0xa1,	0x47b98 },
	{ "AOKO_0_C.TXT", 0x19b6,	0x47b99 },
	{ "AOKO_1.HA6", 0x226bc,	0x47b9d },
	{ "AOKO_1.TXT", 0xa4,	0x47be2 },
	{ "AOKO_1_C.TXT", 0x18cc,	0x47be3 },
	{ "AOKO_2.HA6", 0x1afb0,	0x47be7 },
	{ "AOKO_2.TXT", 0xa2,	0x47c1d },
	{ "AOKO_2_C.TXT", 0x1653,	0x47c1e },
	{ "AOKO_3.HA6", 0x1100c,	0x47c21 },
	{ "AOKO_3.TXT", 0xa2,	0x47c44 },
	{ "AOKO_3_C.TXT", 0x1388,	0x47c45 },
	{ "AOKO_9.HA6", 0x4a680,	0x47c48 },
	{ "AOKO_9.TXT", 0xa2,	0x47cdd },
	{ "AOKO_9_C.TXT", 0x140b,	0x47cde },
	{ "AOKO_C.TXT", 0x1794,	0x47ce1 },
	{ "AOKO_SELIST.TXT", 0x11cc,	0x47ce4 },
	{ "AOKO_WIN.TXT", 0x7e1,	0x47ce7 },
	{ "ARC.CG2", 0x805a8,	0x47ce8 },
	{ "ARC.CPF", 0x2f36c,	0x47de9 },
	{ "ARC.DOF", 0xde248,	0x47e48 },
	{ "ARC.HA6", 0x7fb6c,	0x48005 },
	{ "ARC.PAT", 0x144c8,	0x48105 },
	{ "ARC.TXT", 0xcc,	0x4812e },
	{ "ARC0.PAL", 0x400,	0x4812f },
	{ "ARC1.PAL", 0x400,	0x48130 },
	{ "ARC10.PAL", 0x400,	0x48131 },
	{ "ARC11.PAL", 0x400,	0x48132 },
	{ "ARC12.PAL", 0x400,	0x48133 },
	{ "ARC13.PAL", 0x400,	0x48134 },
	{ "ARC14.PAL", 0x400,	0x48135 },
	{ "ARC15.PAL", 0x400,	0x48136 },
	{ "ARC16.PAL", 0x400,	0x48137 },
	{ "ARC17.PAL", 0x400,	0x48138 },
	{ "ARC18.PAL", 0x400,	0x48139 },
	{ "ARC19.PAL", 0x400,	0x4813a },
	{ "ARC2.ARX", 0x397d40,	0x4813b },
	{ "ARC2.PAL", 0x400,	0x4886b },
	{ "ARC20.PAL", 0x400,	0x4886c },
	{ "ARC21.PAL", 0x400,	0x4886d },
	{ "ARC22.PAL", 0x400,	0x4886e },
	{ "ARC23.PAL", 0x400,	0x4886f },
	{ "ARC24.PAL", 0x400,	0x48870 },
	{ "ARC25.PAL", 0x400,	0x48871 },
	{ "ARC26.PAL", 0x400,	0x48872 },
	{ "ARC27.PAL", 0x400,	0x48873 },
	{ "ARC28.PAL", 0x400,	0x48874 },
	{ "ARC29.PAL", 0x400,	0x48875 },
	{ "ARC3.PAL", 0x400,	0x48876 },
	{ "ARC30.PAL", 0x400,	0x48877 },
	{ "ARC31.PAL", 0x400,	0x48878 },
	{ "ARC32.PAL", 0x400,	0x48879 },
	{ "ARC33.PAL", 0x400,	0x4887a },
	{ "ARC34.PAL", 0x400,	0x4887b },
	{ "ARC35.PAL", 0x400,	0x4887c },
	{ "ARC4.PAL", 0x400,	0x4887d },
	{ "ARC5.PAL", 0x400,	0x4887e },
	{ "ARC6.PAL", 0x400,	0x4887f },
	{ "ARC7.PAL", 0x400,	0x48880 },
	{ "ARC8.PAL", 0x400,	0x48881 },
	{ "ARC9.PAL", 0x400,	0x48882 },
	{ "ARCPAL2.ARX", 0x1f300,	0x48883 },
	{ "ARC_0.HA6", 0x1a8dc,	0x488c2 },
	{ "ARC_0.TXT", 0xb1,	0x488f8 },
	{ "ARC_0_C.TXT", 0x1cec,	0x488f9 },
	{ "ARC_1.HA6", 0x1aa38,	0x488fd },
	{ "ARC_1.TXT", 0xb4,	0x48933 },
	{ "ARC_1_C.TXT", 0x1abe,	0x48934 },
	{ "ARC_2.HA6", 0xd050,	0x48938 },
	{ "ARC_2.TXT", 0xb4,	0x48953 },
	{ "ARC_2_C.TXT", 0x1b0d,	0x48954 },
	{ "ARC_3.HA6", 0x1100c,	0x48958 },
	{ "ARC_3.TXT", 0xb4,	0x4897b },
	{ "ARC_3_C.TXT", 0x196e,	0x4897c },
	{ "ARC_9.HA6", 0x29794,	0x48980 },
	{ "ARC_9.TXT", 0xb4,	0x489d3 },
	{ "ARC_9_C.TXT", 0x1a0a,	0x489d4 },
	{ "ARC_SELIST.TXT", 0x143c,	0x489d8 },
	{ "ARC_WIN.TXT", 0xaba,	0x489db },
	{ "B_AKIHA.CG2", 0x74e18,	0x48a5e },
	{ "B_AKIHA.CPF", 0x2f36c,	0x48b48 },
	{ "B_AKIHA.HA6", 0x88024,	0x48ba7 },
	{ "B_AKIHA.LST", 0xb208,	0x48cb8 },
	{ "B_AKIHA0.PAL", 0x400,	0x48ccf },
	{ "B_AKIHA1.PAL", 0x400,	0x48cd0 },
	{ "B_AKIHA10.PAL", 0x400,	0x48cd1 },
	{ "B_AKIHA11.PAL", 0x400,	0x48cd2 },
	{ "B_AKIHA12.PAL", 0x400,	0x48cd3 },
	{ "B_AKIHA13.PAL", 0x400,	0x48cd4 },
	{ "B_AKIHA14.PAL", 0x400,	0x48cd5 },
	{ "B_AKIHA15.PAL", 0x400,	0x48cd6 },
	{ "B_AKIHA16.PAL", 0x400,	0x48cd7 },
	{ "B_AKIHA17.PAL", 0x400,	0x48cd8 },
	{ "B_AKIHA18.PAL", 0x400,	0x48cd9 },
	{ "B_AKIHA19.PAL", 0x400,	0x48cda },
	{ "B_AKIHA2.ARX", 0x345f40,	0x48cdb },
	{ "B_AKIHA2.PAL", 0x400,	0x49367 },
	{ "B_AKIHA20.PAL", 0x400,	0x49368 },
	{ "B_AKIHA21.PAL", 0x400,	0x49369 },
	{ "B_AKIHA22.PAL", 0x400,	0x4936a },
	{ "B_AKIHA23.PAL", 0x400,	0x4936b },
	{ "B_AKIHA24.PAL", 0x400,	0x4936c },
	{ "B_AKIHA25.PAL", 0x400,	0x4936d },
	{ "B_AKIHA26.PAL", 0x400,	0x4936e },
	{ "B_AKIHA27.PAL", 0x400,	0x4936f },
	{ "B_AKIHA28.PAL", 0x400,	0x49370 },
	{ "B_AKIHA29.PAL", 0x400,	0x49371 },
	{ "B_AKIHA3.PAL", 0x400,	0x49372 },
	{ "B_AKIHA30.PAL", 0x400,	0x49373 },
	{ "B_AKIHA31.PAL", 0x400,	0x49374 },
	{ "B_AKIHA32.PAL", 0x400,	0x49375 },
	{ "B_AKIHA33.PAL", 0x400,	0x49376 },
	{ "B_AKIHA34.PAL", 0x400,	0x49377 },
	{ "B_AKIHA35.PAL", 0x400,	0x49378 },
	{ "B_AKIHA4.PAL", 0x400,	0x49379 },
	{ "B_AKIHA5.PAL", 0x400,	0x4937a },
	{ "B_AKIHA6.PAL", 0x400,	0x4937b },
	{ "B_AKIHA7.PAL", 0x400,	0x4937c },
	{ "B_AKIHA8.PAL", 0x400,	0x4937d },
	{ "B_AKIHA9.PAL", 0x400,	0x4937e },
	{ "B_AKIHAPAL2.ARX", 0x1d1c0,	0x4937f },
	{ "B_AKIHA_0.TXT", 0x90,	0x493ba },
	{ "B_AKIHA_0_C.TXT", 0x1498,	0x493bb },
	{ "B_AKIHA_8.TXT", 0x90,	0x493be },
	{ "B_AKIHA_8_C.TXT", 0x14e5,	0x493bf },
	{ "B_AKIHA_9.TXT", 0x90,	0x493c2 },
	{ "B_AKIHA_9_C.TXT", 0x14e5,	0x493c3 },
	{ "B_AKIHA_SELIST.TXT", 0xc49,	0x493c6 },
	{ "B_AOKO.CG2", 0x6d978,	0x493c8 },
	{ "B_AOKO.CPF", 0x2f36c,	0x494a4 },
	{ "B_AOKO.HA6", 0x78cdc,	0x49503 },
	{ "B_AOKO0.PAL", 0x400,	0x495f5 },
	{ "B_AOKO1.PAL", 0x400,	0x495f6 },
	{ "B_AOKO10.PAL", 0x400,	0x495f7 },
	{ "B_AOKO11.PAL", 0x400,	0x495f8 },
	{ "B_AOKO12.PAL", 0x400,	0x495f9 },
	{ "B_AOKO13.PAL", 0x400,	0x495fa },
	{ "B_AOKO14.PAL", 0x400,	0x495fb },
	{ "B_AOKO15.PAL", 0x400,	0x495fc },
	{ "B_AOKO16.PAL", 0x400,	0x495fd },
	{ "B_AOKO17.PAL", 0x400,	0x495fe },
	{ "B_AOKO18.PAL", 0x400,	0x495ff },
	{ "B_AOKO19.PAL", 0x400,	0x49600 },
	{ "B_AOKO2.ARX", 0x34ec00,	0x49601 },
	{ "B_AOKO2.PAL", 0x400,	0x49c9f },
	{ "B_AOKO20.PAL", 0x400,	0x49ca0 },
	{ "B_AOKO21.PAL", 0x400,	0x49ca1 },
	{ "B_AOKO22.PAL", 0x400,	0x49ca2 },
	{ "B_AOKO23.PAL", 0x400,	0x49ca3 },
	{ "B_AOKO24.PAL", 0x400,	0x49ca4 },
	{ "B_AOKO25.PAL", 0x400,	0x49ca5 },
	{ "B_AOKO26.PAL", 0x400,	0x49ca6 },
	{ "B_AOKO27.PAL", 0x400,	0x49ca7 },
	{ "B_AOKO28.PAL", 0x400,	0x49ca8 },
	{ "B_AOKO29.PAL", 0x400,	0x49ca9 },
	{ "B_AOKO3.PAL", 0x400,	0x49caa },
	{ "B_AOKO30.PAL", 0x400,	0x49cab },
	{ "B_AOKO31.PAL", 0x400,	0x49cac },
	{ "B_AOKO32.PAL", 0x400,	0x49cad },
	{ "B_AOKO33.PAL", 0x400,	0x49cae },
	{ "B_AOKO34.PAL", 0x400,	0x49caf },
	{ "B_AOKO35.PAL", 0x400,	0x49cb0 },
	{ "B_AOKO4.PAL", 0x400,	0x49cb1 },
	{ "B_AOKO5.PAL", 0x400,	0x49cb2 },
	{ "B_AOKO6.PAL", 0x400,	0x49cb3 },
	{ "B_AOKO7.PAL", 0x400,	0x49cb4 },
	{ "B_AOKO8.PAL", 0x400,	0x49cb5 },
	{ "B_AOKO9.PAL", 0x400,	0x49cb6 },
	{ "B_AOKOPAL2.ARX", 0x1c980,	0x49cb7 },
	{ "B_AOKO_0.TXT", 0x92,	0x49cf1 },
	{ "B_AOKO_0_C.TXT", 0x16c4,	0x49cf2 },
	{ "B_AOKO_9.TXT", 0x92,	0x49cf5 },
	{ "B_AOKO_9_C.TXT", 0x1724,	0x49cf6 },
	{ "B_AOKO_SELIST.TXT", 0x11cc,	0x49cf9 },
	{ "B_ARC.CG2", 0x8e1b0,	0x49cfc },
	{ "B_ARC.CPF", 0x2f36c,	0x49e19 },
	{ "B_ARC.HA6", 0xb12c4,	0x49e78 },
	{ "B_ARC0.PAL", 0x400,	0x49fdb },
	{ "B_ARC1.PAL", 0x400,	0x49fdc },
	{ "B_ARC10.PAL", 0x400,	0x49fdd },
	{ "B_ARC11.PAL", 0x400,	0x49fde },
	{ "B_ARC12.PAL", 0x400,	0x49fdf },
	{ "B_ARC13.PAL", 0x400,	0x49fe0 },
	{ "B_ARC14.PAL", 0x400,	0x49fe1 },
	{ "B_ARC15.PAL", 0x400,	0x49fe2 },
	{ "B_ARC16.PAL", 0x400,	0x49fe3 },
	{ "B_ARC17.PAL", 0x400,	0x49fe4 },
	{ "B_ARC18.PAL", 0x400,	0x49fe5 },
	{ "B_ARC19.PAL", 0x400,	0x49fe6 },
	{ "B_ARC2.ARX", 0x4140c0,	0x49fe7 },
	{ "B_ARC2.PAL", 0x400,	0x4a810 },
	{ "B_ARC20.PAL", 0x400,	0x4a811 },
	{ "B_ARC21.PAL", 0x400,	0x4a812 },
	{ "B_ARC22.PAL", 0x400,	0x4a813 },
	{ "B_ARC23.PAL", 0x400,	0x4a814 },
	{ "B_ARC24.PAL", 0x400,	0x4a815 },
	{ "B_ARC25.PAL", 0x400,	0x4a816 },
	{ "B_ARC26.PAL", 0x400,	0x4a817 },
	{ "B_ARC27.PAL", 0x400,	0x4a818 },
	{ "B_ARC28.PAL", 0x400,	0x4a819 },
	{ "B_ARC29.PAL", 0x400,	0x4a81a },
	{ "B_ARC3.PAL", 0x400,	0x4a81b },
	{ "B_ARC30.PAL", 0x400,	0x4a81c },
	{ "B_ARC31.PAL", 0x400,	0x4a81d },
	{ "B_ARC32.PAL", 0x400,	0x4a81e },
	{ "B_ARC33.PAL", 0x400,	0x4a81f },
	{ "B_ARC34.PAL", 0x400,	0x4a820 },
	{ "B_ARC35.PAL", 0x400,	0x4a821 },
	{ "B_ARC4.PAL", 0x400,	0x4a822 },
	{ "B_ARC5.PAL", 0x400,	0x4a823 },
	{ "B_ARC6.PAL", 0x400,	0x4a824 },
	{ "B_ARC7.PAL", 0x400,	0x4a825 },
	{ "B_ARC8.PAL", 0x400,	0x4a826 },
	{ "B_ARC9.PAL", 0x400,	0x4a827 },
	{ "B_ARCPAL2.ARX", 0x24200,	0x4a828 },
	{ "B_ARC_0.TXT", 0xa2,	0x4a871 },
	{ "B_ARC_0_C.TXT", 0x1da4,	0x4a872 },
	{ "B_ARC_9.TXT", 0xa4,	0x4a876 },
	{ "B_ARC_9_C.TXT", 0x127b,	0x4a877 },
	{ "B_ARC_SELIST.TXT", 0xa07,	0x4a87a },
	{ "B_HERMES.CPF", 0x2f36c,	0x4a87c },
	{ "B_HERMES0.PAL", 0x400,	0x4a8db },
	{ "B_HERMES1.PAL", 0x400,	0x4a8dc },
	{ "B_HERMES10.PAL", 0x400,	0x4a8dd },
	{ "B_HERMES11.PAL", 0x400,	0x4a8de },
	{ "B_HERMES12.PAL", 0x400,	0x4a8df },
	{ "B_HERMES13.PAL", 0x400,	0x4a8e0 },
	{ "B_HERMES14.PAL", 0x400,	0x4a8e1 },
	{ "B_HERMES15.PAL", 0x400,	0x4a8e2 },
	{ "B_HERMES16.PAL", 0x400,	0x4a8e3 },
	{ "B_HERMES17.PAL", 0x400,	0x4a8e4 },
	{ "B_HERMES18.PAL", 0x400,	0x4a8e5 },
	{ "B_HERMES19.PAL", 0x400,	0x4a8e6 },
	{ "B_HERMES2.ARX", 0x40,	0x4a8e7 },
	{ "B_HERMES2.PAL", 0x400,	0x4a8e8 },
	{ "B_HERMES20.PAL", 0x400,	0x4a8e9 },
	{ "B_HERMES21.PAL", 0x400,	0x4a8ea },
	{ "B_HERMES22.PAL", 0x400,	0x4a8eb },
	{ "B_HERMES23.PAL", 0x400,	0x4a8ec },
	{ "B_HERMES24.PAL", 0x400,	0x4a8ed },
	{ "B_HERMES25.PAL", 0x400,	0x4a8ee },
	{ "B_HERMES26.PAL", 0x400,	0x4a8ef },
	{ "B_HERMES27.PAL", 0x400,	0x4a8f0 },
	{ "B_HERMES28.PAL", 0x400,	0x4a8f1 },
	{ "B_HERMES29.PAL", 0x400,	0x4a8f2 },
	{ "B_HERMES3.PAL", 0x400,	0x4a8f3 },
	{ "B_HERMES30.PAL", 0x400,	0x4a8f4 },
	{ "B_HERMES31.PAL", 0x400,	0x4a8f5 },
	{ "B_HERMES32.PAL", 0x400,	0x4a8f6 },
	{ "B_HERMES33.PAL", 0x400,	0x4a8f7 },
	{ "B_HERMES34.PAL", 0x400,	0x4a8f8 },
	{ "B_HERMES35.PAL", 0x400,	0x4a8f9 },
	{ "B_HERMES4.PAL", 0x400,	0x4a8fa },
	{ "B_HERMES5.PAL", 0x400,	0x4a8fb },
	{ "B_HERMES6.PAL", 0x400,	0x4a8fc },
	{ "B_HERMES7.PAL", 0x400,	0x4a8fd },
	{ "B_HERMES8.PAL", 0x400,	0x4a8fe },
	{ "B_HERMES9.PAL", 0x400,	0x4a8ff },
	{ "B_HERMESPAL2.ARX", 0x40,	0x4a900 },
	{ "B_HERMES_9.TXT", 0xa8,	0x4a901 },
	{ "B_HERMES_9_C.TXT", 0x5b9,	0x4a902 },
	{ "B_HERMES_SELIST.TXT", 0x6fb,	0x4a903 },
	{ "B_KOHAKU.CPF", 0x2f36c,	0x4a904 },
	{ "B_KOHAKU_M.CG2", 0x45868,	0x4a963 },
	{ "B_KOHAKU_M.CPF", 0x2f36c,	0x4a9ef },
	{ "B_KOHAKU_M.HA6", 0x42f20,	0x4aa4e },
	{ "B_KOHAKU_M0.PAL", 0x400,	0x4aad4 },
	{ "B_KOHAKU_M1.PAL", 0x400,	0x4aad5 },
	{ "B_KOHAKU_M10.PAL", 0x400,	0x4aad6 },
	{ "B_KOHAKU_M11.PAL", 0x400,	0x4aad7 },
	{ "B_KOHAKU_M12.PAL", 0x400,	0x4aad8 },
	{ "B_KOHAKU_M13.PAL", 0x400,	0x4aad9 },
	{ "B_KOHAKU_M14.PAL", 0x400,	0x4aada },
	{ "B_KOHAKU_M15.PAL", 0x400,	0x4aadb },
	{ "B_KOHAKU_M16.PAL", 0x400,	0x4aadc },
	{ "B_KOHAKU_M17.PAL", 0x400,	0x4aadd },
	{ "B_KOHAKU_M18.PAL", 0x400,	0x4aade },
	{ "B_KOHAKU_M19.PAL", 0x400,	0x4aadf },
	{ "B_KOHAKU_M2.ARX", 0x178500,	0x4aae0 },
	{ "B_KOHAKU_M2.PAL", 0x400,	0x4add1 },
	{ "B_KOHAKU_M20.PAL", 0x400,	0x4add2 },
	{ "B_KOHAKU_M21.PAL", 0x400,	0x4add3 },
	{ "B_KOHAKU_M22.PAL", 0x400,	0x4add4 },
	{ "B_KOHAKU_M23.PAL", 0x400,	0x4add5 },
	{ "B_KOHAKU_M24.PAL", 0x400,	0x4add6 },
	{ "B_KOHAKU_M25.PAL", 0x400,	0x4add7 },
	{ "B_KOHAKU_M26.PAL", 0x400,	0x4add8 },
	{ "B_KOHAKU_M27.PAL", 0x400,	0x4add9 },
	{ "B_KOHAKU_M28.PAL", 0x400,	0x4adda },
	{ "B_KOHAKU_M29.PAL", 0x400,	0x4addb },
	{ "B_KOHAKU_M3.PAL", 0x400,	0x4addc },
	{ "B_KOHAKU_M30.PAL", 0x400,	0x4addd },
	{ "B_KOHAKU_M31.PAL", 0x400,	0x4adde },
	{ "B_KOHAKU_M32.PAL", 0x400,	0x4addf },
	{ "B_KOHAKU_M33.PAL", 0x400,	0x4ade0 },
	{ "B_KOHAKU_M34.PAL", 0x400,	0x4ade1 },
	{ "B_KOHAKU_M35.PAL", 0x400,	0x4ade2 },
	{ "B_KOHAKU_M4.PAL", 0x400,	0x4ade3 },
	{ "B_KOHAKU_M5.PAL", 0x400,	0x4ade4 },
	{ "B_KOHAKU_M6.PAL", 0x400,	0x4ade5 },
	{ "B_KOHAKU_M7.PAL", 0x400,	0x4ade6 },
	{ "B_KOHAKU_M8.PAL", 0x400,	0x4ade7 },
	{ "B_KOHAKU_M9.PAL", 0x400,	0x4ade8 },
	{ "B_KOHAKU_MPAL2.ARX", 0x3e80,	0x4ade9 },
	{ "B_KOHAKU_M_0.TXT", 0xac,	0x4adf1 },
	{ "B_KOHAKU_M_0_C.TXT", 0x1280,	0x4adf2 },
	{ "B_KOHAKU_M_9.TXT", 0xac,	0x4adf5 },
	{ "B_KOHAKU_M_9_C.TXT", 0x12ba,	0x4adf6 },
	{ "B_KOHAKU_M_SELIST.TXT", 0xc99,	0x4adf9 },
	{ "B_MIYAKO.CG2", 0x3ca48,	0x4adfb },
	{ "B_MIYAKO.CPF", 0x2f36c,	0x4ae75 },
	{ "B_MIYAKO.HA6", 0x4d568,	0x4aed4 },
	{ "B_MIYAKO0.PAL", 0x400,	0x4af6f },
	{ "B_MIYAKO1.PAL", 0x400,	0x4af70 },
	{ "B_MIYAKO10.PAL", 0x400,	0x4af71 },
	{ "B_MIYAKO11.PAL", 0x400,	0x4af72 },
	{ "B_MIYAKO12.PAL", 0x400,	0x4af73 },
	{ "B_MIYAKO13.PAL", 0x400,	0x4af74 },
	{ "B_MIYAKO14.PAL", 0x400,	0x4af75 },
	{ "B_MIYAKO15.PAL", 0x400,	0x4af76 },
	{ "B_MIYAKO16.PAL", 0x400,	0x4af77 },
	{ "B_MIYAKO17.PAL", 0x400,	0x4af78 },
	{ "B_MIYAKO18.PAL", 0x400,	0x4af79 },
	{ "B_MIYAKO19.PAL", 0x400,	0x4af7a },
	{ "B_MIYAKO2.ARX", 0x1746c0,	0x4af7b },
	{ "B_MIYAKO2.PAL", 0x400,	0x4b264 },
	{ "B_MIYAKO20.PAL", 0x400,	0x4b265 },
	{ "B_MIYAKO21.PAL", 0x400,	0x4b266 },
	{ "B_MIYAKO22.PAL", 0x400,	0x4b267 },
	{ "B_MIYAKO23.PAL", 0x400,	0x4b268 },
	{ "B_MIYAKO24.PAL", 0x400,	0x4b269 },
	{ "B_MIYAKO25.PAL", 0x400,	0x4b26a },
	{ "B_MIYAKO26.PAL", 0x400,	0x4b26b },
	{ "B_MIYAKO27.PAL", 0x400,	0x4b26c },
	{ "B_MIYAKO28.PAL", 0x400,	0x4b26d },
	{ "B_MIYAKO29.PAL", 0x400,	0x4b26e },
	{ "B_MIYAKO3.PAL", 0x400,	0x4b26f },
	{ "B_MIYAKO30.PAL", 0x400,	0x4b270 },
	{ "B_MIYAKO31.PAL", 0x400,	0x4b271 },
	{ "B_MIYAKO32.PAL", 0x400,	0x4b272 },
	{ "B_MIYAKO33.PAL", 0x400,	0x4b273 },
	{ "B_MIYAKO34.PAL", 0x400,	0x4b274 },
	{ "B_MIYAKO35.PAL", 0x400,	0x4b275 },
	{ "B_MIYAKO4.PAL", 0x400,	0x4b276 },
	{ "B_MIYAKO5.PAL", 0x400,	0x4b277 },
	{ "B_MIYAKO6.PAL", 0x400,	0x4b278 },
	{ "B_MIYAKO7.PAL", 0x400,	0x4b279 },
	{ "B_MIYAKO8.PAL", 0x400,	0x4b27a },
	{ "B_MIYAKO9.PAL", 0x400,	0x4b27b },
	{ "B_MIYAKOPAL2.ARX", 0x5bc0,	0x4b27c },
	{ "B_MIYAKO_0.TXT", 0xa6,	0x4b288 },
	{ "B_MIYAKO_0_C.TXT", 0x138f,	0x4b289 },
	{ "B_MIYAKO_9.TXT", 0xa6,	0x4b28c },
	{ "B_MIYAKO_9_C.TXT", 0x1411,	0x4b28d },
	{ "B_MIYAKO_SELIST.TXT", 0xbc3,	0x4b290 },
	{ "B_M_HISUI_P.CG2", 0x2e480,	0x4b292 },
	{ "B_M_HISUI_P.CPF", 0x2f36c,	0x4b2ef },
	{ "B_M_HISUI_P.HA6", 0x12fb4,	0x4b34e },
	{ "B_M_HISUI_P0.PAL", 0x400,	0x4b374 },
	{ "B_M_HISUI_P1.PAL", 0x400,	0x4b375 },
	{ "B_M_HISUI_P10.PAL", 0x400,	0x4b376 },
	{ "B_M_HISUI_P11.PAL", 0x400,	0x4b377 },
	{ "B_M_HISUI_P12.PAL", 0x400,	0x4b378 },
	{ "B_M_HISUI_P13.PAL", 0x400,	0x4b379 },
	{ "B_M_HISUI_P14.PAL", 0x400,	0x4b37a },
	{ "B_M_HISUI_P15.PAL", 0x400,	0x4b37b },
	{ "B_M_HISUI_P16.PAL", 0x400,	0x4b37c },
	{ "B_M_HISUI_P17.PAL", 0x400,	0x4b37d },
	{ "B_M_HISUI_P18.PAL", 0x400,	0x4b37e },
	{ "B_M_HISUI_P19.PAL", 0x400,	0x4b37f },
	{ "B_M_HISUI_P2.ARX", 0x87500,	0x4b380 },
	{ "B_M_HISUI_P2.PAL", 0x400,	0x4b48f },
	{ "B_M_HISUI_P20.PAL", 0x400,	0x4b490 },
	{ "B_M_HISUI_P21.PAL", 0x400,	0x4b491 },
	{ "B_M_HISUI_P22.PAL", 0x400,	0x4b492 },
	{ "B_M_HISUI_P23.PAL", 0x400,	0x4b493 },
	{ "B_M_HISUI_P24.PAL", 0x400,	0x4b494 },
	{ "B_M_HISUI_P25.PAL", 0x400,	0x4b495 },
	{ "B_M_HISUI_P26.PAL", 0x400,	0x4b496 },
	{ "B_M_HISUI_P27.PAL", 0x400,	0x4b497 },
	{ "B_M_HISUI_P28.PAL", 0x400,	0x4b498 },
	{ "B_M_HISUI_P29.PAL", 0x400,	0x4b499 },
	{ "B_M_HISUI_P3.PAL", 0x400,	0x4b49a },
	{ "B_M_HISUI_P30.PAL", 0x400,	0x4b49b },
	{ "B_M_HISUI_P31.PAL", 0x400,	0x4b49c },
	{ "B_M_HISUI_P32.PAL", 0x400,	0x4b49d },
	{ "B_M_HISUI_P33.PAL", 0x400,	0x4b49e },
	{ "B_M_HISUI_P34.PAL", 0x400,	0x4b49f },
	{ "B_M_HISUI_P35.PAL", 0x400,	0x4b4a0 },
	{ "B_M_HISUI_P4.PAL", 0x400,	0x4b4a1 },
	{ "B_M_HISUI_P5.PAL", 0x400,	0x4b4a2 },
	{ "B_M_HISUI_P6.PAL", 0x400,	0x4b4a3 },
	{ "B_M_HISUI_P7.PAL", 0x400,	0x4b4a4 },
	{ "B_M_HISUI_P8.PAL", 0x400,	0x4b4a5 },
	{ "B_M_HISUI_P9.PAL", 0x400,	0x4b4a6 },
	{ "B_M_HISUI_PPAL2.ARX", 0x2180,	0x4b4a7 },
	{ "B_M_HISUI_P_0.TXT", 0xb2,	0x4b4ac },
	{ "B_M_HISUI_P_0_C.TXT", 0x317,	0x4b4ad },
	{ "B_M_HISUI_P_9.TXT", 0xb2,	0x4b4ae },
	{ "B_M_HISUI_P_9_C.TXT", 0x317,	0x4b4af },
	{ "B_M_HISUI_P_SELIST.TXT", 0x2ea,	0x4b4b0 },
	{ "B_RYOUGI.CG2", 0x7ac08,	0x4b4b1 },
	{ "B_RYOUGI.CPF", 0x2f36c,	0x4b5a7 },
	{ "B_RYOUGI.HA6", 0xb18e8,	0x4b606 },
	{ "B_RYOUGI0.PAL", 0x400,	0x4b76a },
	{ "B_RYOUGI1.PAL", 0x400,	0x4b76b },
	{ "B_RYOUGI10.PAL", 0x400,	0x4b76c },
	{ "B_RYOUGI11.PAL", 0x400,	0x4b76d },
	{ "B_RYOUGI12.PAL", 0x400,	0x4b76e },
	{ "B_RYOUGI13.PAL", 0x400,	0x4b76f },
	{ "B_RYOUGI14.PAL", 0x400,	0x4b770 },
	{ "B_RYOUGI15.PAL", 0x400,	0x4b771 },
	{ "B_RYOUGI16.PAL", 0x400,	0x4b772 },
	{ "B_RYOUGI17.PAL", 0x400,	0x4b773 },
	{ "B_RYOUGI18.PAL", 0x400,	0x4b774 },
	{ "B_RYOUGI19.PAL", 0x400,	0x4b775 },
	{ "B_RYOUGI2.ARX", 0x2f8f40,	0x4b776 },
	{ "B_RYOUGI2.PAL", 0x400,	0x4bd68 },
	{ "B_RYOUGI20.PAL", 0x400,	0x4bd69 },
	{ "B_RYOUGI21.PAL", 0x400,	0x4bd6a },
	{ "B_RYOUGI22.PAL", 0x400,	0x4bd6b },
	{ "B_RYOUGI23.PAL", 0x400,	0x4bd6c },
	{ "B_RYOUGI24.PAL", 0x400,	0x4bd6d },
	{ "B_RYOUGI25.PAL", 0x400,	0x4bd6e },
	{ "B_RYOUGI26.PAL", 0x400,	0x4bd6f },
	{ "B_RYOUGI27.PAL", 0x400,	0x4bd70 },
	{ "B_RYOUGI28.PAL", 0x400,	0x4bd71 },
	{ "B_RYOUGI29.PAL", 0x400,	0x4bd72 },
	{ "B_RYOUGI3.PAL", 0x400,	0x4bd73 },
	{ "B_RYOUGI30.PAL", 0x400,	0x4bd74 },
	{ "B_RYOUGI31.PAL", 0x400,	0x4bd75 },
	{ "B_RYOUGI32.PAL", 0x400,	0x4bd76 },
	{ "B_RYOUGI33.PAL", 0x400,	0x4bd77 },
	{ "B_RYOUGI34.PAL", 0x400,	0x4bd78 },
	{ "B_RYOUGI35.PAL", 0x400,	0x4bd79 },
	{ "B_RYOUGI4.PAL", 0x400,	0x4bd7a },
	{ "B_RYOUGI5.PAL", 0x400,	0x4bd7b },
	{ "B_RYOUGI6.PAL", 0x400,	0x4bd7c },
	{ "B_RYOUGI7.PAL", 0x400,	0x4bd7d },
	{ "B_RYOUGI8.PAL", 0x400,	0x4bd7e },
	{ "B_RYOUGI9.PAL", 0x400,	0x4bd7f },
	{ "B_RYOUGIPAL2.ARX", 0x13800,	0x4bd80 },
	{ "B_RYOUGI_9.HA6", 0x1b680,	0x4bda7 },
	{ "B_RYOUGI_9.TXT", 0xc8,	0x4bdde },
	{ "B_RYOUGI_9_C.TXT", 0x1f6f,	0x4bddf },
	{ "B_RYOUGI_SELIST.TXT", 0x1e59,	0x4bde3 },
	{ "B_WARAKIA.CG2", 0x72e38,	0x4bde7 },
	{ "B_WARAKIA.CPF", 0x2f36c,	0x4becd },
	{ "B_WARAKIA.HA6", 0x68fa0,	0x4bf2c },
	{ "B_WARAKIA0.PAL", 0x400,	0x4bffe },
	{ "B_WARAKIA1.PAL", 0x400,	0x4bfff },
	{ "B_WARAKIA10.PAL", 0x400,	0x4c000 },
	{ "B_WARAKIA11.PAL", 0x400,	0x4c001 },
	{ "B_WARAKIA12.PAL", 0x400,	0x4c002 },
	{ "B_WARAKIA13.PAL", 0x400,	0x4c003 },
	{ "B_WARAKIA14.PAL", 0x400,	0x4c004 },
	{ "B_WARAKIA15.PAL", 0x400,	0x4c005 },
	{ "B_WARAKIA16.PAL", 0x400,	0x4c006 },
	{ "B_WARAKIA17.PAL", 0x400,	0x4c007 },
	{ "B_WARAKIA18.PAL", 0x400,	0x4c008 },
	{ "B_WARAKIA19.PAL", 0x400,	0x4c009 },
	{ "B_WARAKIA2.ARX", 0x339cc0,	0x4c00a },
	{ "B_WARAKIA2.PAL", 0x400,	0x4c67e },
	{ "B_WARAKIA20.PAL", 0x400,	0x4c67f },
	{ "B_WARAKIA21.PAL", 0x400,	0x4c680 },
	{ "B_WARAKIA22.PAL", 0x400,	0x4c681 },
	{ "B_WARAKIA23.PAL", 0x400,	0x4c682 },
	{ "B_WARAKIA24.PAL", 0x400,	0x4c683 },
	{ "B_WARAKIA25.PAL", 0x400,	0x4c684 },
	{ "B_WARAKIA26.PAL", 0x400,	0x4c685 },
	{ "B_WARAKIA27.PAL", 0x400,	0x4c686 },
	{ "B_WARAKIA28.PAL", 0x400,	0x4c687 },
	{ "B_WARAKIA29.PAL", 0x400,	0x4c688 },
	{ "B_WARAKIA3.PAL", 0x400,	0x4c689 },
	{ "B_WARAKIA30.PAL", 0x400,	0x4c68a },
	{ "B_WARAKIA31.PAL", 0x400,	0x4c68b },
	{ "B_WARAKIA32.PAL", 0x400,	0x4c68c },
	{ "B_WARAKIA33.PAL", 0x400,	0x4c68d },
	{ "B_WARAKIA34.PAL", 0x400,	0x4c68e },
	{ "B_WARAKIA35.PAL", 0x400,	0x4c68f },
	{ "B_WARAKIA4.PAL", 0x400,	0x4c690 },
	{ "B_WARAKIA5.PAL", 0x400,	0x4c691 },
	{ "B_WARAKIA6.PAL", 0x400,	0x4c692 },
	{ "B_WARAKIA7.PAL", 0x400,	0x4c693 },
	{ "B_WARAKIA8.PAL", 0x400,	0x4c694 },
	{ "B_WARAKIA9.PAL", 0x400,	0x4c695 },
	{ "B_WARAKIAPAL2.ARX", 0x1c140,	0x4c696 },
	{ "B_WARAKIA_0.TXT", 0xaa,	0x4c6cf },
	{ "B_WARAKIA_0_C.TXT", 0x10c2,	0x4c6d0 },
	{ "B_WARAKIA_9.TXT", 0xaa,	0x4c6d3 },
	{ "B_WARAKIA_9_C.TXT", 0x1100,	0x4c6d4 },
	{ "B_WARAKIA_SELIST.TXT", 0xce0,	0x4c6d7 },
	{ "B_WLEN.CG2", 0x88000,	0x4c6d9 },
	{ "B_WLEN.CPF", 0x2f36c,	0x4c7e9 },
	{ "B_WLEN.HA6", 0x7741c,	0x4c848 },
	{ "B_WLEN0.PAL", 0x400,	0x4c937 },
	{ "B_WLEN1.PAL", 0x400,	0x4c938 },
	{ "B_WLEN10.PAL", 0x400,	0x4c939 },
	{ "B_WLEN11.PAL", 0x400,	0x4c93a },
	{ "B_WLEN12.PAL", 0x400,	0x4c93b },
	{ "B_WLEN13.PAL", 0x400,	0x4c93c },
	{ "B_WLEN14.PAL", 0x400,	0x4c93d },
	{ "B_WLEN15.PAL", 0x400,	0x4c93e },
	{ "B_WLEN16.PAL", 0x400,	0x4c93f },
	{ "B_WLEN17.PAL", 0x400,	0x4c940 },
	{ "B_WLEN18.PAL", 0x400,	0x4c941 },
	{ "B_WLEN19.PAL", 0x400,	0x4c942 },
	{ "B_WLEN2.ARX", 0x423500,	0x4c943 },
	{ "B_WLEN2.PAL", 0x400,	0x4d18a },
	{ "B_WLEN20.PAL", 0x400,	0x4d18b },
	{ "B_WLEN21.PAL", 0x400,	0x4d18c },
	{ "B_WLEN22.PAL", 0x400,	0x4d18d },
	{ "B_WLEN23.PAL", 0x400,	0x4d18e },
	{ "B_WLEN24.PAL", 0x400,	0x4d18f },
	{ "B_WLEN25.PAL", 0x400,	0x4d190 },
	{ "B_WLEN26.PAL", 0x400,	0x4d191 },
	{ "B_WLEN27.PAL", 0x400,	0x4d192 },
	{ "B_WLEN28.PAL", 0x400,	0x4d193 },
	{ "B_WLEN29.PAL", 0x400,	0x4d194 },
	{ "B_WLEN3.PAL", 0x400,	0x4d195 },
	{ "B_WLEN30.PAL", 0x400,	0x4d196 },
	{ "B_WLEN31.PAL", 0x400,	0x4d197 },
	{ "B_WLEN32.PAL", 0x400,	0x4d198 },
	{ "B_WLEN33.PAL", 0x400,	0x4d199 },
	{ "B_WLEN34.PAL", 0x400,	0x4d19a },
	{ "B_WLEN35.PAL", 0x400,	0x4d19b },
	{ "B_WLEN4.PAL", 0x400,	0x4d19c },
	{ "B_WLEN5.PAL", 0x400,	0x4d19d },
	{ "B_WLEN6.PAL", 0x400,	0x4d19e },
	{ "B_WLEN7.PAL", 0x400,	0x4d19f },
	{ "B_WLEN8.PAL", 0x400,	0x4d1a0 },
	{ "B_WLEN9.PAL", 0x400,	0x4d1a1 },
	{ "B_WLENPAL2.ARX", 0x2bec0,	0x4d1a2 },
	{ "B_WLEN_0.TXT", 0x8d,	0x4d1fa },
	{ "B_WLEN_0_C.TXT", 0x1341,	0x4d1fb },
	{ "B_WLEN_9.TXT", 0x8d,	0x4d1fe },
	{ "B_WLEN_9_C.TXT", 0x137b,	0x4d1ff },
	{ "B_WLEN_SELIST.TXT", 0x950,	0x4d202 },
	{ "CHARASELECT.TXT", 0x19d2,	0x4d204 },
	{ "CIEL.CG2", 0x70900,	0x4d208 },
	{ "CIEL.CPF", 0x2f36c,	0x4d2ea },
	{ "CIEL.DOF", 0xf0abc,	0x4d349 },
	{ "CIEL.HA6", 0x857e4,	0x4d52b },
	{ "CIEL.PAT", 0x144c8,	0x4d636 },
	{ "CIEL.TXT", 0xc2,	0x4d65f },
	{ "CIEL0.PAL", 0x400,	0x4d660 },
	{ "CIEL1.PAL", 0x400,	0x4d661 },
	{ "CIEL10.PAL", 0x400,	0x4d662 },
	{ "CIEL11.PAL", 0x400,	0x4d663 },
	{ "CIEL12.PAL", 0x400,	0x4d664 },
	{ "CIEL13.PAL", 0x400,	0x4d665 },
	{ "CIEL14.PAL", 0x400,	0x4d666 },
	{ "CIEL15.PAL", 0x400,	0x4d667 },
	{ "CIEL16.PAL", 0x400,	0x4d668 },
	{ "CIEL17.PAL", 0x400,	0x4d669 },
	{ "CIEL18.PAL", 0x400,	0x4d66a },
	{ "CIEL19.PAL", 0x400,	0x4d66b },
	{ "CIEL2.ARX", 0x267680,	0x4d66c },
	{ "CIEL2.PAL", 0x400,	0x4db3b },
	{ "CIEL20.PAL", 0x400,	0x4db3c },
	{ "CIEL21.PAL", 0x400,	0x4db3d },
	{ "CIEL22.PAL", 0x400,	0x4db3e },
	{ "CIEL23.PAL", 0x400,	0x4db3f },
	{ "CIEL24.PAL", 0x400,	0x4db40 },
	{ "CIEL25.PAL", 0x400,	0x4db41 },
	{ "CIEL26.PAL", 0x400,	0x4db42 },
	{ "CIEL27.PAL", 0x400,	0x4db43 },
	{ "CIEL28.PAL", 0x400,	0x4db44 },
	{ "CIEL29.PAL", 0x400,	0x4db45 },
	{ "CIEL3.PAL", 0x400,	0x4db46 },
	{ "CIEL30.PAL", 0x400,	0x4db47 },
	{ "CIEL31.PAL", 0x400,	0x4db48 },
	{ "CIEL32.PAL", 0x400,	0x4db49 },
	{ "CIEL33.PAL", 0x400,	0x4db4a },
	{ "CIEL34.PAL", 0x400,	0x4db4b },
	{ "CIEL35.PAL", 0x400,	0x4db4c },
	{ "CIEL4.PAL", 0x400,	0x4db4d },
	{ "CIEL5.PAL", 0x400,	0x4db4e },
	{ "CIEL6.PAL", 0x400,	0x4db4f },
	{ "CIEL7.PAL", 0x400,	0x4db50 },
	{ "CIEL8.PAL", 0x400,	0x4db51 },
	{ "CIEL9.PAL", 0x400,	0x4db52 },
	{ "CIELPAL2.ARX", 0xc7c0,	0x4db53 },
	{ "CIEL_0.HA6", 0x17fdc,	0x4db6c },
	{ "CIEL_0.TXT", 0xa7,	0x4db9c },
	{ "CIEL_0_C.TXT", 0x2360,	0x4db9d },
	{ "CIEL_1.HA6", 0x1b4c8,	0x4dba2 },
	{ "CIEL_1.TXT", 0xa5,	0x4dbd9 },
	{ "CIEL_1_C.TXT", 0x1670,	0x4dbda },
	{ "CIEL_2.HA6", 0x1c2c8,	0x4dbdd },
	{ "CIEL_2.TXT", 0x9a,	0x4dc16 },
	{ "CIEL_2_C.TXT", 0x1c95,	0x4dc17 },
	{ "CIEL_3.HA6", 0x1100c,	0x4dc1b },
	{ "CIEL_3.TXT", 0x9c,	0x4dc3e },
	{ "CIEL_3_C.TXT", 0x1a84,	0x4dc3f },
	{ "CIEL_9.HA6", 0x2fb40,	0x4dc43 },
	{ "CIEL_9.TXT", 0x9c,	0x4dca3 },
	{ "CIEL_9_C.TXT", 0x19e4,	0x4dca4 },
	{ "CIEL_C.TXT", 0x1c49,	0x4dca8 },
	{ "CIEL_SELIST.TXT", 0x1020,	0x4dcac },
	{ "CIEL_WIN.TXT", 0x910,	0x4dcaf },
	{ "CROUCH.CPF", 0x2f36c,	0x4dcb1 },
	{ "DAMIEN_C.TXT", 0x1484,	0x4dd91 },
	{ "DEKU.CPF", 0x2f36c,	0x4dd94 },
	{ "EFFECT.CG2", 0x2f560,	0x4de35 },
	{ "EFFECT.HA6", 0x1c60c,	0x4de94 },
	{ "EFFECT.PAT", 0x128e4,	0x4decd },
	{ "EFFECT.TXT", 0xba,	0x4def3 },
	{ "EFFECT2.ARX", 0x153880,	0x4def4 },
	{ "EFFECTPAL2.ARX", 0x13c40,	0x4e19c },
	{ "F_CIEL.ARX", 0xff640,	0x4e1c4 },
	{ "F_CIEL.CG2", 0x20578,	0x4e3c3 },
	{ "F_CIEL.CPF", 0x2f36c,	0x4e404 },
	{ "F_CIEL_C.TXT", 0x1df,	0x4e463 },
	{ "GAKIHA.CPF", 0x2f36c,	0x4e464 },
	{ "GUARD.CPF", 0x2f36c,	0x4e7ab },
	{ "G_AKIHA.CPF", 0x2f36c,	0x4e80a },
	{ "G_AKIHA.HA6", 0x12f8c,	0x4e869 },
	{ "G_AKIHA.PAT", 0x2640f,	0x4e88f },
	{ "G_AKIHA_0.HA6", 0x354c,	0x4e8dc },
	{ "G_AKIHA_0.TXT", 0x9a,	0x4e8e3 },
	{ "G_AKIHA_0_C.TXT", 0x4f2,	0x4e8e4 },
	{ "G_AKIHA_9.TXT", 0x9a,	0x4e8e5 },
	{ "G_AKIHA_9_C.TXT", 0x7e6,	0x4e8e6 },
	{ "G_AKIHA_SELIST.TXT", 0x67c,	0x4e8e7 },
	{ "G_CHAOS_0.HA6", 0x2eb94,	0x4e8e8 },
	{ "G_CHAOS_0.TXT", 0x81,	0x4e946 },
	{ "G_CHAOS_0_C.TXT", 0xc23,	0x4e947 },
	{ "HERMES.CG2", 0x1c618,	0x4e949 },
	{ "HERMES.CPF", 0x2f36c,	0x4e982 },
	{ "HERMES.HA6", 0x15d30,	0x4e9e1 },
	{ "HERMES.PAT", 0x5e269,	0x4ea0d },
	{ "HERMES0.PAL", 0x400,	0x4eaca },
	{ "HERMES1.PAL", 0x400,	0x4eacb },
	{ "HERMES10.PAL", 0x400,	0x4eacc },
	{ "HERMES11.PAL", 0x400,	0x4eacd },
	{ "HERMES12.PAL", 0x400,	0x4eace },
	{ "HERMES13.PAL", 0x400,	0x4eacf },
	{ "HERMES14.PAL", 0x400,	0x4ead0 },
	{ "HERMES15.PAL", 0x400,	0x4ead1 },
	{ "HERMES16.PAL", 0x400,	0x4ead2 },
	{ "HERMES17.PAL", 0x400,	0x4ead3 },
	{ "HERMES18.PAL", 0x400,	0x4ead4 },
	{ "HERMES19.PAL", 0x400,	0x4ead5 },
	{ "HERMES2.ARX", 0x146bc0,	0x4ead6 },
	{ "HERMES2.PAL", 0x400,	0x4ed64 },
	{ "HERMES20.PAL", 0x400,	0x4ed65 },
	{ "HERMES21.PAL", 0x400,	0x4ed66 },
	{ "HERMES22.PAL", 0x400,	0x4ed67 },
	{ "HERMES23.PAL", 0x400,	0x4ed68 },
	{ "HERMES24.PAL", 0x400,	0x4ed69 },
	{ "HERMES25.PAL", 0x400,	0x4ed6a },
	{ "HERMES26.PAL", 0x400,	0x4ed6b },
	{ "HERMES27.PAL", 0x400,	0x4ed6c },
	{ "HERMES28.PAL", 0x400,	0x4ed6d },
	{ "HERMES29.PAL", 0x400,	0x4ed6e },
	{ "HERMES3.PAL", 0x400,	0x4ed6f },
	{ "HERMES30.PAL", 0x400,	0x4ed70 },
	{ "HERMES31.PAL", 0x400,	0x4ed71 },
	{ "HERMES32.PAL", 0x400,	0x4ed72 },
	{ "HERMES33.PAL", 0x400,	0x4ed73 },
	{ "HERMES34.PAL", 0x400,	0x4ed74 },
	{ "HERMES35.PAL", 0x400,	0x4ed75 },
	{ "HERMES4.PAL", 0x400,	0x4ed76 },
	{ "HERMES5.PAL", 0x400,	0x4ed77 },
	{ "HERMES6.PAL", 0x400,	0x4ed78 },
	{ "HERMES7.PAL", 0x400,	0x4ed79 },
	{ "HERMES8.PAL", 0x400,	0x4ed7a },
	{ "HERMES9.PAL", 0x400,	0x4ed7b },
	{ "HERMESPAL2.ARX", 0x133c0,	0x4ed7c },
	{ "HERMES_0.TXT", 0xa6,	0x4eda3 },
	{ "HERMES_0_C.TXT", 0x5f6,	0x4eda4 },
	{ "HERMES_SELIST.TXT", 0x6fb,	0x4eda5 },
	{ "HISUI.CG2", 0x52b70,	0x4ee09 },
	{ "HISUI.CPF", 0x2f36c,	0x4eeaf },
	{ "HISUI.DOF", 0xcde08,	0x4ef0e },
	{ "HISUI.HA6", 0x9c544,	0x4f0aa },
	{ "HISUI.LST", 0x11728,	0x4f1e3 },
	{ "HISUI.PAT", 0xd36a,	0x4f206 },
	{ "HISUI.TXT", 0xdc,	0x4f221 },
	{ "HISUI0.PAL", 0x400,	0x4f222 },
	{ "HISUI1.PAL", 0x400,	0x4f223 },
	{ "HISUI10.PAL", 0x400,	0x4f224 },
	{ "HISUI11.PAL", 0x400,	0x4f225 },
	{ "HISUI12.PAL", 0x400,	0x4f226 },
	{ "HISUI13.PAL", 0x400,	0x4f227 },
	{ "HISUI14.PAL", 0x400,	0x4f228 },
	{ "HISUI15.PAL", 0x400,	0x4f229 },
	{ "HISUI16.PAL", 0x400,	0x4f22a },
	{ "HISUI17.PAL", 0x400,	0x4f22b },
	{ "HISUI18.PAL", 0x400,	0x4f22c },
	{ "HISUI19.PAL", 0x400,	0x4f22d },
	{ "HISUI2.ARX", 0x1d8f80,	0x4f22e },
	{ "HISUI2.PAL", 0x400,	0x4f5e0 },
	{ "HISUI20.PAL", 0x400,	0x4f5e1 },
	{ "HISUI21.PAL", 0x400,	0x4f5e2 },
	{ "HISUI22.PAL", 0x400,	0x4f5e3 },
	{ "HISUI23.PAL", 0x400,	0x4f5e4 },
	{ "HISUI24.PAL", 0x400,	0x4f5e5 },
	{ "HISUI25.PAL", 0x400,	0x4f5e6 },
	{ "HISUI26.PAL", 0x400,	0x4f5e7 },
	{ "HISUI27.PAL", 0x400,	0x4f5e8 },
	{ "HISUI28.PAL", 0x400,	0x4f5e9 },
	{ "HISUI29.PAL", 0x400,	0x4f5ea },
	{ "HISUI3.PAL", 0x400,	0x4f5eb },
	{ "HISUI30.PAL", 0x400,	0x4f5ec },
	{ "HISUI31.PAL", 0x400,	0x4f5ed },
	{ "HISUI32.PAL", 0x400,	0x4f5ee },
	{ "HISUI33.PAL", 0x400,	0x4f5ef },
	{ "HISUI34.PAL", 0x400,	0x4f5f0 },
	{ "HISUI35.PAL", 0x400,	0x4f5f1 },
	{ "HISUI4.PAL", 0x400,	0x4f5f2 },
	{ "HISUI5.PAL", 0x400,	0x4f5f3 },
	{ "HISUI6.PAL", 0x400,	0x4f5f4 },
	{ "HISUI7.PAL", 0x400,	0x4f5f5 },
	{ "HISUI8.PAL", 0x400,	0x4f5f6 },
	{ "HISUI9.PAL", 0x400,	0x4f5f7 },
	{ "HISUIPAL2.ARX", 0x7080,	0x4f5f8 },
	{ "HISUI_0.HA6", 0x15ef8,	0x4f607 },
	{ "HISUI_0.LST", 0x11728,	0x4f633 },
	{ "HISUI_0.TXT", 0xb8,	0x4f656 },
	{ "HISUI_0_C.TXT", 0x1945,	0x4f657 },
	{ "HISUI_1.HA6", 0x2f760,	0x4f65b },
	{ "HISUI_1.LST", 0x11728,	0x4f6ba },
	{ "HISUI_1.TXT", 0xba,	0x4f6dd },
	{ "HISUI_1_C.TXT", 0x190d,	0x4f6de },
	{ "HISUI_2.HA6", 0x1d6fc,	0x4f6e2 },
	{ "HISUI_2.LST", 0x11728,	0x4f71d },
	{ "HISUI_2.TXT", 0xb2,	0x4f740 },
	{ "HISUI_2_C.TXT", 0x1890,	0x4f741 },
	{ "HISUI_3.HA6", 0x1100c,	0x4f745 },
	{ "HISUI_3.TXT", 0xb2,	0x4f768 },
	{ "HISUI_3_C.TXT", 0x14bf,	0x4f769 },
	{ "HISUI_9.HA6", 0x2d18c,	0x4f76c },
	{ "HISUI_9.TXT", 0xb2,	0x4f7c7 },
	{ "HISUI_9_C.TXT", 0x15f3,	0x4f7c8 },
	{ "HISUI_C.TXT", 0x1618,	0x4f7cb },
	{ "HISUI_SELIST.TXT", 0x14c1,	0x4f7ce },
	{ "HISUI_WIN.TXT", 0x6d8,	0x4f7d1 },
	{ "JUMP.CPF", 0x2f36c,	0x4f7d2 },
	{ "JUMPG.CPF", 0x2f36c,	0x4f831 },
	{ "KISHIMA.CG2", 0x77da0,	0x4f890 },
	{ "KISHIMA.CPF", 0x2f36c,	0x4f980 },
	{ "KISHIMA.DOF", 0x7cbf8,	0x4f9df },
	{ "KISHIMA.HA6", 0x55fa8,	0x4fad9 },
	{ "KISHIMA.PAT", 0x0,	0x4fb85 },
	{ "KISHIMA.TXT", 0xd4,	0x4fb85 },
	{ "KISHIMA0.PAL", 0x400,	0x4fb86 },
	{ "KISHIMA1.PAL", 0x400,	0x4fb87 },
	{ "KISHIMA10.PAL", 0x400,	0x4fb88 },
	{ "KISHIMA11.PAL", 0x400,	0x4fb89 },
	{ "KISHIMA12.PAL", 0x400,	0x4fb8a },
	{ "KISHIMA13.PAL", 0x400,	0x4fb8b },
	{ "KISHIMA14.PAL", 0x400,	0x4fb8c },
	{ "KISHIMA15.PAL", 0x400,	0x4fb8d },
	{ "KISHIMA16.PAL", 0x400,	0x4fb8e },
	{ "KISHIMA17.PAL", 0x400,	0x4fb8f },
	{ "KISHIMA18.PAL", 0x400,	0x4fb90 },
	{ "KISHIMA19.PAL", 0x400,	0x4fb91 },
	{ "KISHIMA2.ARX", 0x39a000,	0x4fb92 },
	{ "KISHIMA2.PAL", 0x400,	0x502c6 },
	{ "KISHIMA20.PAL", 0x400,	0x502c7 },
	{ "KISHIMA21.PAL", 0x400,	0x502c8 },
	{ "KISHIMA22.PAL", 0x400,	0x502c9 },
	{ "KISHIMA23.PAL", 0x400,	0x502ca },
	{ "KISHIMA24.PAL", 0x400,	0x502cb },
	{ "KISHIMA25.PAL", 0x400,	0x502cc },
	{ "KISHIMA26.PAL", 0x400,	0x502cd },
	{ "KISHIMA27.PAL", 0x400,	0x502ce },
	{ "KISHIMA28.PAL", 0x400,	0x502cf },
	{ "KISHIMA29.PAL", 0x400,	0x502d0 },
	{ "KISHIMA3.PAL", 0x400,	0x502d1 },
	{ "KISHIMA30.PAL", 0x400,	0x502d2 },
	{ "KISHIMA31.PAL", 0x400,	0x502d3 },
	{ "KISHIMA32.PAL", 0x400,	0x502d4 },
	{ "KISHIMA33.PAL", 0x400,	0x502d5 },
	{ "KISHIMA34.PAL", 0x400,	0x502d6 },
	{ "KISHIMA35.PAL", 0x400,	0x502d7 },
	{ "KISHIMA4.PAL", 0x400,	0x502d8 },
	{ "KISHIMA5.PAL", 0x400,	0x502d9 },
	{ "KISHIMA6.PAL", 0x400,	0x502da },
	{ "KISHIMA7.PAL", 0x400,	0x502db },
	{ "KISHIMA8.PAL", 0x400,	0x502dc },
	{ "KISHIMA9.PAL", 0x400,	0x502dd },
	{ "KISHIMAPAL2.ARX", 0x21c80,	0x502de },
	{ "KISHIMA_0.HA6", 0x1082c,	0x50322 },
	{ "KISHIMA_0.TXT", 0xac,	0x50344 },
	{ "KISHIMA_0_C.TXT", 0x1759,	0x50345 },
	{ "KISHIMA_1.HA6", 0x157f8,	0x50348 },
	{ "KISHIMA_1.TXT", 0xac,	0x50373 },
	{ "KISHIMA_1_C.TXT", 0x1675,	0x50374 },
	{ "KISHIMA_2.HA6", 0xe29c,	0x50377 },
	{ "KISHIMA_2.TXT", 0xac,	0x50394 },
	{ "KISHIMA_2_C.TXT", 0x12a6,	0x50395 },
	{ "KISHIMA_3.HA6", 0x1100c,	0x50398 },
	{ "KISHIMA_3.TXT", 0xac,	0x503bb },
	{ "KISHIMA_3_C.TXT", 0x1034,	0x503bc },
	{ "KISHIMA_9.HA6", 0x16f70,	0x503bf },
	{ "KISHIMA_9.TXT", 0xac,	0x503ed },
	{ "KISHIMA_9_C.TXT", 0x109a,	0x503ee },
	{ "KISHIMA_SELIST.TXT", 0x11ec,	0x503f1 },
	{ "KISHIMA_WIN.TXT", 0x712,	0x503f4 },
	{ "KOHAKU.CG2", 0x62068,	0x503f5 },
	{ "KOHAKU.CPF", 0x2f36c,	0x504ba },
	{ "KOHAKU.DOF", 0xe0d10,	0x50519 },
	{ "KOHAKU.HA6", 0x97c90,	0x506db },
	{ "KOHAKU.PAT", 0xcb37,	0x5080b },
	{ "KOHAKU.TXT", 0xe1,	0x50825 },
	{ "KOHAKU0.PAL", 0x400,	0x50826 },
	{ "KOHAKU1.PAL", 0x400,	0x50827 },
	{ "KOHAKU10.PAL", 0x400,	0x50828 },
	{ "KOHAKU11.PAL", 0x400,	0x50829 },
	{ "KOHAKU12.PAL", 0x400,	0x5082a },
	{ "KOHAKU13.PAL", 0x400,	0x5082b },
	{ "KOHAKU14.PAL", 0x400,	0x5082c },
	{ "KOHAKU15.PAL", 0x400,	0x5082d },
	{ "KOHAKU16.PAL", 0x400,	0x5082e },
	{ "KOHAKU17.PAL", 0x400,	0x5082f },
	{ "KOHAKU18.PAL", 0x400,	0x50830 },
	{ "KOHAKU19.PAL", 0x400,	0x50831 },
	{ "KOHAKU2.ARX", 0x229f40,	0x50832 },
	{ "KOHAKU2.PAL", 0x400,	0x50c86 },
	{ "KOHAKU20.PAL", 0x400,	0x50c87 },
	{ "KOHAKU21.PAL", 0x400,	0x50c88 },
	{ "KOHAKU22.PAL", 0x400,	0x50c89 },
	{ "KOHAKU23.PAL", 0x400,	0x50c8a },
	{ "KOHAKU24.PAL", 0x400,	0x50c8b },
	{ "KOHAKU25.PAL", 0x400,	0x50c8c },
	{ "KOHAKU26.PAL", 0x400,	0x50c8d },
	{ "KOHAKU27.PAL", 0x400,	0x50c8e },
	{ "KOHAKU28.PAL", 0x400,	0x50c8f },
	{ "KOHAKU29.PAL", 0x400,	0x50c90 },
	{ "KOHAKU3.PAL", 0x400,	0x50c91 },
	{ "KOHAKU30.PAL", 0x400,	0x50c92 },
	{ "KOHAKU31.PAL", 0x400,	0x50c93 },
	{ "KOHAKU32.PAL", 0x400,	0x50c94 },
	{ "KOHAKU33.PAL", 0x400,	0x50c95 },
	{ "KOHAKU34.PAL", 0x400,	0x50c96 },
	{ "KOHAKU35.PAL", 0x400,	0x50c97 },
	{ "KOHAKU4.PAL", 0x400,	0x50c98 },
	{ "KOHAKU5.PAL", 0x400,	0x50c99 },
	{ "KOHAKU6.PAL", 0x400,	0x50c9a },
	{ "KOHAKU7.PAL", 0x400,	0x50c9b },
	{ "KOHAKU8.PAL", 0x400,	0x50c9c },
	{ "KOHAKU9.PAL", 0x400,	0x50c9d },
	{ "KOHAKUPAL2.ARX", 0x8540,	0x50c9e },
	{ "KOHAKU_0.HA6", 0x1fa38,	0x50caf },
	{ "KOHAKU_0.TXT", 0xb5,	0x50cef },
	{ "KOHAKU_0_C.TXT", 0x1495,	0x50cf0 },
	{ "KOHAKU_1.HA6", 0x3c9f0,	0x50cf3 },
	{ "KOHAKU_1.TXT", 0xb7,	0x50d6d },
	{ "KOHAKU_1_C.TXT", 0x1d14,	0x50d6e },
	{ "KOHAKU_2.HA6", 0x24d68,	0x50d72 },
	{ "KOHAKU_2.TXT", 0xb7,	0x50dbc },
	{ "KOHAKU_2_C.TXT", 0x162d,	0x50dbd },
	{ "KOHAKU_3.HA6", 0x1100c,	0x50dc0 },
	{ "KOHAKU_3.TXT", 0xb7,	0x50de3 },
	{ "KOHAKU_3_C.TXT", 0x1165,	0x50de4 },
	{ "KOHAKU_9.HA6", 0x30430,	0x50de7 },
	{ "KOHAKU_9.TXT", 0xb7,	0x50e48 },
	{ "KOHAKU_9_C.TXT", 0x1209,	0x50e49 },
	{ "KOHAKU_C.TXT", 0x12f8,	0x50e4c },
	{ "KOHAKU_M.CG2", 0x50ad0,	0x50e4f },
	{ "KOHAKU_M.CPF", 0x2f36c,	0x50ef1 },
	{ "KOHAKU_M.HA6", 0x61818,	0x50f50 },
	{ "KOHAKU_M.TXT", 0xa6,	0x51014 },
	{ "KOHAKU_M0.PAL", 0x400,	0x51015 },
	{ "KOHAKU_M1.PAL", 0x400,	0x51016 },
	{ "KOHAKU_M10.PAL", 0x400,	0x51017 },
	{ "KOHAKU_M11.PAL", 0x400,	0x51018 },
	{ "KOHAKU_M12.PAL", 0x400,	0x51019 },
	{ "KOHAKU_M13.PAL", 0x400,	0x5101a },
	{ "KOHAKU_M14.PAL", 0x400,	0x5101b },
	{ "KOHAKU_M15.PAL", 0x400,	0x5101c },
	{ "KOHAKU_M16.PAL", 0x400,	0x5101d },
	{ "KOHAKU_M17.PAL", 0x400,	0x5101e },
	{ "KOHAKU_M18.PAL", 0x400,	0x5101f },
	{ "KOHAKU_M19.PAL", 0x400,	0x51020 },
	{ "KOHAKU_M2.ARX", 0x1c91c0,	0x51021 },
	{ "KOHAKU_M2.PAL", 0x400,	0x513b4 },
	{ "KOHAKU_M20.PAL", 0x400,	0x513b5 },
	{ "KOHAKU_M21.PAL", 0x400,	0x513b6 },
	{ "KOHAKU_M22.PAL", 0x400,	0x513b7 },
	{ "KOHAKU_M23.PAL", 0x400,	0x513b8 },
	{ "KOHAKU_M24.PAL", 0x400,	0x513b9 },
	{ "KOHAKU_M25.PAL", 0x400,	0x513ba },
	{ "KOHAKU_M26.PAL", 0x400,	0x513bb },
	{ "KOHAKU_M27.PAL", 0x400,	0x513bc },
	{ "KOHAKU_M28.PAL", 0x400,	0x513bd },
	{ "KOHAKU_M29.PAL", 0x400,	0x513be },
	{ "KOHAKU_M3.PAL", 0x400,	0x513bf },
	{ "KOHAKU_M30.PAL", 0x400,	0x513c0 },
	{ "KOHAKU_M31.PAL", 0x400,	0x513c1 },
	{ "KOHAKU_M32.PAL", 0x400,	0x513c2 },
	{ "KOHAKU_M33.PAL", 0x400,	0x513c3 },
	{ "KOHAKU_M34.PAL", 0x400,	0x513c4 },
	{ "KOHAKU_M35.PAL", 0x400,	0x513c5 },
	{ "KOHAKU_M4.PAL", 0x400,	0x513c6 },
	{ "KOHAKU_M5.PAL", 0x400,	0x513c7 },
	{ "KOHAKU_M6.PAL", 0x400,	0x513c8 },
	{ "KOHAKU_M7.PAL", 0x400,	0x513c9 },
	{ "KOHAKU_M8.PAL", 0x400,	0x513ca },
	{ "KOHAKU_M9.PAL", 0x400,	0x513cb },
	{ "KOHAKU_MPAL2.ARX", 0x5fc0,	0x513cc },
	{ "KOHAKU_M_0.HA6", 0x49a8,	0x513d8 },
	{ "KOHAKU_M_0.TXT", 0xbf,	0x513e2 },
	{ "KOHAKU_M_0_C.TXT", 0x1529,	0x513e3 },
	{ "KOHAKU_M_1.HA6", 0x12984,	0x513e6 },
	{ "KOHAKU_M_1.TXT", 0xbd,	0x5140c },
	{ "KOHAKU_M_1_C.TXT", 0x1487,	0x5140d },
	{ "KOHAKU_M_2.HA6", 0x155cc,	0x51410 },
	{ "KOHAKU_M_2.TXT", 0xbd,	0x5143b },
	{ "KOHAKU_M_2_C.TXT", 0x1006,	0x5143c },
	{ "KOHAKU_M_9_C.TXT", 0x1280,	0x5143f },
	{ "KOHAKU_M_C.TXT", 0x1600,	0x51442 },
	{ "KOHAKU_M_SELIST.TXT", 0xff2,	0x51445 },
	{ "KOHAKU_M_WIN.TXT", 0x75c,	0x51447 },
	{ "KOHAKU_SELIST.TXT", 0x1347,	0x51448 },
	{ "KOHAKU_WIN.TXT", 0x75c,	0x5144b },
	{ "LEN.CG2", 0x5a5b0,	0x5148e },
	{ "LEN.CPF", 0x2f36c,	0x51543 },
	{ "LEN.DOF", 0x8ba80,	0x515a2 },
	{ "LEN.HA6", 0x520a8,	0x516ba },
	{ "LEN.PAT", 0xc150,	0x5175f },
	{ "LEN.TXT", 0xcc,	0x51778 },
	{ "LEN0.PAL", 0x400,	0x51779 },
	{ "LEN1.PAL", 0x400,	0x5177a },
	{ "LEN10.PAL", 0x400,	0x5177b },
	{ "LEN11.PAL", 0x400,	0x5177c },
	{ "LEN12.PAL", 0x400,	0x5177d },
	{ "LEN13.PAL", 0x400,	0x5177e },
	{ "LEN14.PAL", 0x400,	0x5177f },
	{ "LEN15.PAL", 0x400,	0x51780 },
	{ "LEN16.PAL", 0x400,	0x51781 },
	{ "LEN17.PAL", 0x400,	0x51782 },
	{ "LEN18.PAL", 0x400,	0x51783 },
	{ "LEN19.PAL", 0x400,	0x51784 },
	{ "LEN2.ARX", 0x201000,	0x51785 },
	{ "LEN2.PAL", 0x400,	0x51b87 },
	{ "LEN20.PAL", 0x400,	0x51b88 },
	{ "LEN21.PAL", 0x400,	0x51b89 },
	{ "LEN22.PAL", 0x400,	0x51b8a },
	{ "LEN23.PAL", 0x400,	0x51b8b },
	{ "LEN24.PAL", 0x400,	0x51b8c },
	{ "LEN25.PAL", 0x400,	0x51b8d },
	{ "LEN26.PAL", 0x400,	0x51b8e },
	{ "LEN27.PAL", 0x400,	0x51b8f },
	{ "LEN28.PAL", 0x400,	0x51b90 },
	{ "LEN29.PAL", 0x400,	0x51b91 },
	{ "LEN3.PAL", 0x400,	0x51b92 },
	{ "LEN30.PAL", 0x400,	0x51b93 },
	{ "LEN31.PAL", 0x400,	0x51b94 },
	{ "LEN32.PAL", 0x400,	0x51b95 },
	{ "LEN33.PAL", 0x400,	0x51b96 },
	{ "LEN34.PAL", 0x400,	0x51b97 },
	{ "LEN35.PAL", 0x400,	0x51b98 },
	{ "LEN4.PAL", 0x400,	0x51b99 },
	{ "LEN5.PAL", 0x400,	0x51b9a },
	{ "LEN6.PAL", 0x400,	0x51b9b },
	{ "LEN7.PAL", 0x400,	0x51b9c },
	{ "LEN8.PAL", 0x400,	0x51b9d },
	{ "LEN9.PAL", 0x400,	0x51b9e },
	{ "LENPAL2.ARX", 0xf580,	0x51b9f },
	{ "LEN_0.HA6", 0x5428,	0x51bbe },
	{ "LEN_0.TXT", 0xa8,	0x51bc9 },
	{ "LEN_0_C.TXT", 0x10cc,	0x51bca },
	{ "LEN_1.HA6", 0x173b4,	0x51bcd },
	{ "LEN_1.TXT", 0xa8,	0x51bfc },
	{ "LEN_1_C.TXT", 0x124e,	0x51bfd },
	{ "LEN_2.HA6", 0xa288,	0x51c00 },
	{ "LEN_2.TXT", 0xa8,	0x51c15 },
	{ "LEN_2_C.TXT", 0x1115,	0x51c16 },
	{ "LEN_3.HA6", 0x1100c,	0x51c19 },
	{ "LEN_3.TXT", 0xa8,	0x51c3c },
	{ "LEN_3_C.TXT", 0xf34,	0x51c3d },
	{ "LEN_9.HA6", 0x15298,	0x51c3f },
	{ "LEN_9.TXT", 0xa8,	0x51c6a },
	{ "LEN_9_C.TXT", 0xfd0,	0x51c6b },
	{ "LEN_C.TXT", 0x142f,	0x51c6d },
	{ "LEN_SELIST.TXT", 0x380,	0x51c70 },
	{ "MIYAKO.CG2", 0x3d4e0,	0x51c92 },
	{ "MIYAKO.CPF", 0x2f36c,	0x51d0d },
	{ "MIYAKO.DOF", 0x636a4,	0x51d6c },
	{ "MIYAKO.HA6", 0x4bed8,	0x51e33 },
	{ "MIYAKO.PAT", 0xbe94,	0x51ecb },
	{ "MIYAKO.TXT", 0xe1,	0x51ee3 },
	{ "MIYAKO0.PAL", 0x400,	0x51ee4 },
	{ "MIYAKO1.PAL", 0x400,	0x51ee5 },
	{ "MIYAKO10.PAL", 0x400,	0x51ee6 },
	{ "MIYAKO11.PAL", 0x400,	0x51ee7 },
	{ "MIYAKO12.PAL", 0x400,	0x51ee8 },
	{ "MIYAKO13.PAL", 0x400,	0x51ee9 },
	{ "MIYAKO14.PAL", 0x400,	0x51eea },
	{ "MIYAKO15.PAL", 0x400,	0x51eeb },
	{ "MIYAKO16.PAL", 0x400,	0x51eec },
	{ "MIYAKO17.PAL", 0x400,	0x51eed },
	{ "MIYAKO18.PAL", 0x400,	0x51eee },
	{ "MIYAKO19.PAL", 0x400,	0x51eef },
	{ "MIYAKO2.ARX", 0x177d80,	0x51ef0 },
	{ "MIYAKO2.PAL", 0x400,	0x521e0 },
	{ "MIYAKO20.PAL", 0x400,	0x521e1 },
	{ "MIYAKO21.PAL", 0x400,	0x521e2 },
	{ "MIYAKO22.PAL", 0x400,	0x521e3 },
	{ "MIYAKO23.PAL", 0x400,	0x521e4 },
	{ "MIYAKO24.PAL", 0x400,	0x521e5 },
	{ "MIYAKO25.PAL", 0x400,	0x521e6 },
	{ "MIYAKO26.PAL", 0x400,	0x521e7 },
	{ "MIYAKO27.PAL", 0x400,	0x521e8 },
	{ "MIYAKO28.PAL", 0x400,	0x521e9 },
	{ "MIYAKO29.PAL", 0x400,	0x521ea },
	{ "MIYAKO3.PAL", 0x400,	0x521eb },
	{ "MIYAKO30.PAL", 0x400,	0x521ec },
	{ "MIYAKO31.PAL", 0x400,	0x521ed },
	{ "MIYAKO32.PAL", 0x400,	0x521ee },
	{ "MIYAKO33.PAL", 0x400,	0x521ef },
	{ "MIYAKO34.PAL", 0x400,	0x521f0 },
	{ "MIYAKO35.PAL", 0x400,	0x521f1 },
	{ "MIYAKO4.PAL", 0x400,	0x521f2 },
	{ "MIYAKO5.PAL", 0x400,	0x521f3 },
	{ "MIYAKO6.PAL", 0x400,	0x521f4 },
	{ "MIYAKO7.PAL", 0x400,	0x521f5 },
	{ "MIYAKO8.PAL", 0x400,	0x521f6 },
	{ "MIYAKO9.PAL", 0x400,	0x521f7 },
	{ "MIYAKOPAL2.ARX", 0x5bc0,	0x521f8 },
	{ "MIYAKO_0.HA6", 0x723c,	0x52204 },
	{ "MIYAKO_0.TXT", 0xb7,	0x52213 },
	{ "MIYAKO_0_C.TXT", 0x14ff,	0x52214 },
	{ "MIYAKO_1.HA6", 0x12b1c,	0x52217 },
	{ "MIYAKO_1.TXT", 0xb7,	0x5223d },
	{ "MIYAKO_1_C.TXT", 0x16c2,	0x5223e },
	{ "MIYAKO_2.HA6", 0xd5ac,	0x52241 },
	{ "MIYAKO_2.TXT", 0xb7,	0x5225c },
	{ "MIYAKO_2_C.TXT", 0x1347,	0x5225d },
	{ "MIYAKO_3.HA6", 0x1100c,	0x52260 },
	{ "MIYAKO_3.TXT", 0xb7,	0x52283 },
	{ "MIYAKO_3_C.TXT", 0x10d9,	0x52284 },
	{ "MIYAKO_9.HA6", 0x193e4,	0x52287 },
	{ "MIYAKO_9.TXT", 0xb7,	0x522ba },
	{ "MIYAKO_9_C.TXT", 0x12f7,	0x522bb },
	{ "MIYAKO_SELIST.TXT", 0xd16,	0x522be },
	{ "MIYAKO_WIN.TXT", 0x776,	0x522c0 },
	{ "M_HISUI.CG2", 0x7d1e8,	0x522c1 },
	{ "M_HISUI.CPF", 0x2f36c,	0x523bc },
	{ "M_HISUI.DOF", 0xb74e8,	0x5241b },
	{ "M_HISUI.HA6", 0xa853c,	0x5258a },
	{ "M_HISUI.PAT", 0x12030,	0x526db },
	{ "M_HISUI.TXT", 0xe8,	0x52700 },
	{ "M_HISUI0.PAL", 0x400,	0x52701 },
	{ "M_HISUI1.PAL", 0x400,	0x52723 },
	{ "M_HISUI10.PAL", 0x400,	0x52724 },
	{ "M_HISUI11.PAL", 0x400,	0x52725 },
	{ "M_HISUI12.PAL", 0x400,	0x52726 },
	{ "M_HISUI13.PAL", 0x400,	0x52727 },
	{ "M_HISUI14.PAL", 0x400,	0x52728 },
	{ "M_HISUI15.PAL", 0x400,	0x52729 },
	{ "M_HISUI16.PAL", 0x400,	0x5272a },
	{ "M_HISUI17.PAL", 0x400,	0x5272b },
	{ "M_HISUI18.PAL", 0x400,	0x5272c },
	{ "M_HISUI19.PAL", 0x400,	0x5272d },
	{ "M_HISUI2.ARX", 0x2d9bc0,	0x5272e },
	{ "M_HISUI2.PAL", 0x400,	0x52ce2 },
	{ "M_HISUI20.PAL", 0x400,	0x52ce3 },
	{ "M_HISUI21.PAL", 0x400,	0x52ce4 },
	{ "M_HISUI22.PAL", 0x400,	0x52ce5 },
	{ "M_HISUI23.PAL", 0x400,	0x52ce6 },
	{ "M_HISUI24.PAL", 0x400,	0x52ce7 },
	{ "M_HISUI25.PAL", 0x400,	0x52ce8 },
	{ "M_HISUI26.PAL", 0x400,	0x52ce9 },
	{ "M_HISUI27.PAL", 0x400,	0x52cea },
	{ "M_HISUI28.PAL", 0x400,	0x52ceb },
	{ "M_HISUI29.PAL", 0x400,	0x52cec },
	{ "M_HISUI3.PAL", 0x400,	0x52ced },
	{ "M_HISUI30.PAL", 0x400,	0x52cee },
	{ "M_HISUI31.PAL", 0x400,	0x52cef },
	{ "M_HISUI32.PAL", 0x400,	0x52cf0 },
	{ "M_HISUI33.PAL", 0x400,	0x52cf1 },
	{ "M_HISUI34.PAL", 0x400,	0x52cf2 },
	{ "M_HISUI35.PAL", 0x400,	0x52cf3 },
	{ "M_HISUI4.PAL", 0x400,	0x52cf4 },
	{ "M_HISUI5.PAL", 0x400,	0x52cf5 },
	{ "M_HISUI6.PAL", 0x400,	0x52cf6 },
	{ "M_HISUI7.PAL", 0x400,	0x52cf7 },
	{ "M_HISUI8.PAL", 0x400,	0x52cf8 },
	{ "M_HISUI9.PAL", 0x400,	0x52cf9 },
	{ "M_HISUIPAL2.ARX", 0x165c0,	0x52cfa },
	{ "M_HISUI_0.HA6", 0x34f4,	0x52d27 },
	{ "M_HISUI_0.TXT", 0xbe,	0x52d2e },
	{ "M_HISUI_0_C.TXT", 0x1a46,	0x52d2f },
	{ "M_HISUI_1.HA6", 0x6350,	0x52d33 },
	{ "M_HISUI_1.TXT", 0xbc,	0x52d40 },
	{ "M_HISUI_1_C.TXT", 0x1087,	0x52d41 },
	{ "M_HISUI_2.HA6", 0xe648,	0x52d44 },
	{ "M_HISUI_2.TXT", 0xbc,	0x52d61 },
	{ "M_HISUI_2_C.TXT", 0x13a6,	0x52d62 },
	{ "M_HISUI_3.HA6", 0x1100c,	0x52d65 },
	{ "M_HISUI_9.HA6", 0x17120,	0x52d88 },
	{ "M_HISUI_9.TXT", 0xbc,	0x52db7 },
	{ "M_HISUI_9_C.TXT", 0x1232,	0x52db8 },
	{ "M_HISUI_M.CG2", 0x53c68,	0x52dbb },
	{ "M_HISUI_M.CPF", 0x2f36c,	0x52e63 },
	{ "M_HISUI_M.HA6", 0x4a510,	0x52ec2 },
	{ "M_HISUI_M.TXT", 0xaa,	0x52f57 },
	{ "M_HISUI_M0.PAL", 0x400,	0x52f58 },
	{ "M_HISUI_M1.PAL", 0x400,	0x52f59 },
	{ "M_HISUI_M10.PAL", 0x400,	0x52f5a },
	{ "M_HISUI_M11.PAL", 0x400,	0x52f5b },
	{ "M_HISUI_M12.PAL", 0x400,	0x52f5c },
	{ "M_HISUI_M13.PAL", 0x400,	0x52f5d },
	{ "M_HISUI_M14.PAL", 0x400,	0x52f5e },
	{ "M_HISUI_M15.PAL", 0x400,	0x52f5f },
	{ "M_HISUI_M16.PAL", 0x400,	0x52f60 },
	{ "M_HISUI_M17.PAL", 0x400,	0x52f61 },
	{ "M_HISUI_M18.PAL", 0x400,	0x52f62 },
	{ "M_HISUI_M19.PAL", 0x400,	0x52f63 },
	{ "M_HISUI_M2.ARX", 0x1d1000,	0x52f64 },
	{ "M_HISUI_M2.PAL", 0x400,	0x53306 },
	{ "M_HISUI_M20.PAL", 0x400,	0x53307 },
	{ "M_HISUI_M21.PAL", 0x400,	0x53308 },
	{ "M_HISUI_M22.PAL", 0x400,	0x53309 },
	{ "M_HISUI_M23.PAL", 0x400,	0x5330a },
	{ "M_HISUI_M24.PAL", 0x400,	0x5330b },
	{ "M_HISUI_M25.PAL", 0x400,	0x5330c },
	{ "M_HISUI_M26.PAL", 0x400,	0x5330d },
	{ "M_HISUI_M27.PAL", 0x400,	0x5330e },
	{ "M_HISUI_M28.PAL", 0x400,	0x5330f },
	{ "M_HISUI_M29.PAL", 0x400,	0x53310 },
	{ "M_HISUI_M3.PAL", 0x400,	0x53311 },
	{ "M_HISUI_M30.PAL", 0x400,	0x53312 },
	{ "M_HISUI_M31.PAL", 0x400,	0x53313 },
	{ "M_HISUI_M32.PAL", 0x400,	0x53314 },
	{ "M_HISUI_M33.PAL", 0x400,	0x53315 },
	{ "M_HISUI_M34.PAL", 0x400,	0x53316 },
	{ "M_HISUI_M35.PAL", 0x400,	0x53317 },
	{ "M_HISUI_M4.PAL", 0x400,	0x53318 },
	{ "M_HISUI_M5.PAL", 0x400,	0x53319 },
	{ "M_HISUI_M6.PAL", 0x400,	0x5331a },
	{ "M_HISUI_M7.PAL", 0x400,	0x5331b },
	{ "M_HISUI_M8.PAL", 0x400,	0x5331c },
	{ "M_HISUI_M9.PAL", 0x400,	0x5331d },
	{ "M_HISUI_MPAL2.ARX", 0xc380,	0x5331e },
	{ "M_HISUI_M_0.HA6", 0x4a84,	0x53337 },
	{ "M_HISUI_M_0.TXT", 0xc4,	0x53341 },
	{ "M_HISUI_M_0_C.TXT", 0xdf3,	0x53342 },
	{ "M_HISUI_M_1.HA6", 0xca14,	0x53344 },
	{ "M_HISUI_M_1.TXT", 0xc4,	0x5335e },
	{ "M_HISUI_M_1_C.TXT", 0xd00,	0x5335f },
	{ "M_HISUI_M_2.HA6", 0x115e4,	0x53361 },
	{ "M_HISUI_M_2.TXT", 0xc4,	0x53384 },
	{ "M_HISUI_M_2_C.TXT", 0xddd,	0x53385 },
	{ "M_HISUI_M_9.HA6", 0x354c,	0x53387 },
	{ "M_HISUI_M_SELIST.TXT", 0x998,	0x5338e },
	{ "M_HISUI_P.CG2", 0x5c638,	0x53390 },
	{ "M_HISUI_P.CPF", 0x2f36c,	0x53449 },
	{ "M_HISUI_P.HA6", 0x73288,	0x534a8 },
	{ "M_HISUI_P.TXT", 0xaa,	0x5358f },
	{ "M_HISUI_P0.PAL", 0x400,	0x53590 },
	{ "M_HISUI_P1.PAL", 0x400,	0x53591 },
	{ "M_HISUI_P10.PAL", 0x400,	0x53592 },
	{ "M_HISUI_P11.PAL", 0x400,	0x53593 },
	{ "M_HISUI_P12.PAL", 0x400,	0x53594 },
	{ "M_HISUI_P13.PAL", 0x400,	0x53595 },
	{ "M_HISUI_P14.PAL", 0x400,	0x53596 },
	{ "M_HISUI_P15.PAL", 0x400,	0x53597 },
	{ "M_HISUI_P16.PAL", 0x400,	0x53598 },
	{ "M_HISUI_P17.PAL", 0x400,	0x53599 },
	{ "M_HISUI_P18.PAL", 0x400,	0x5359a },
	{ "M_HISUI_P19.PAL", 0x400,	0x5359b },
	{ "M_HISUI_P2.ARX", 0x22c200,	0x5359c },
	{ "M_HISUI_P2.PAL", 0x400,	0x539f5 },
	{ "M_HISUI_P20.PAL", 0x400,	0x539f6 },
	{ "M_HISUI_P21.PAL", 0x400,	0x539f7 },
	{ "M_HISUI_P22.PAL", 0x400,	0x539f8 },
	{ "M_HISUI_P23.PAL", 0x400,	0x539f9 },
	{ "M_HISUI_P24.PAL", 0x400,	0x539fa },
	{ "M_HISUI_P25.PAL", 0x400,	0x539fb },
	{ "M_HISUI_P26.PAL", 0x400,	0x539fc },
	{ "M_HISUI_P27.PAL", 0x400,	0x539fd },
	{ "M_HISUI_P28.PAL", 0x400,	0x539fe },
	{ "M_HISUI_P29.PAL", 0x400,	0x539ff },
	{ "M_HISUI_P3.PAL", 0x400,	0x53a00 },
	{ "M_HISUI_P30.PAL", 0x400,	0x53a01 },
	{ "M_HISUI_P31.PAL", 0x400,	0x53a02 },
	{ "M_HISUI_P32.PAL", 0x400,	0x53a03 },
	{ "M_HISUI_P33.PAL", 0x400,	0x53a04 },
	{ "M_HISUI_P34.PAL", 0x400,	0x53a05 },
	{ "M_HISUI_P35.PAL", 0x400,	0x53a06 },
	{ "M_HISUI_P4.PAL", 0x400,	0x53a07 },
	{ "M_HISUI_P5.PAL", 0x400,	0x53a08 },
	{ "M_HISUI_P6.PAL", 0x400,	0x53a09 },
	{ "M_HISUI_P7.PAL", 0x400,	0x53a0a },
	{ "M_HISUI_P8.PAL", 0x400,	0x53a0b },
	{ "M_HISUI_P9.PAL", 0x400,	0x53a0c },
	{ "M_HISUI_PPAL2.ARX", 0x11f00,	0x53a0d },
	{ "M_HISUI_P_0.HA6", 0xd6f4,	0x53a31 },
	{ "M_HISUI_P_0.TXT", 0xc8,	0x53a4c },
	{ "M_HISUI_P_0_C.TXT", 0x31c,	0x53a4d },
	{ "M_HISUI_P_1.HA6", 0xca2c,	0x53a4e },
	{ "M_HISUI_P_1.TXT", 0xc4,	0x53a68 },
	{ "M_HISUI_P_1_C.TXT", 0x31c,	0x53a69 },
	{ "M_HISUI_P_2.HA6", 0xfe14,	0x53a6a },
	{ "M_HISUI_P_2.TXT", 0xc4,	0x53a8a },
	{ "M_HISUI_P_2_C.TXT", 0x31c,	0x53a8b },
	{ "M_HISUI_P_9.HA6", 0x354c,	0x53a8c },
	{ "M_HISUI_P_9_C.TXT", 0x160e,	0x53a93 },
	{ "M_HISUI_P_SELIST.TXT", 0x84c,	0x53a96 },
	{ "M_HISUI_SELIST.TXT", 0xd1e,	0x53a98 },
	{ "NANAYA.CG2", 0x57eb0,	0x53a9a },
	{ "NANAYA.CPF", 0x2f36c,	0x53b4a },
	{ "NANAYA.DOF", 0x918d4,	0x53ba9 },
	{ "NANAYA.HA6", 0x6b754,	0x53ccd },
	{ "NANAYA.PAT", 0x0,	0x53da4 },
	{ "NANAYA.TXT", 0xce,	0x53da4 },
	{ "NANAYA0.PAL", 0x400,	0x53da5 },
	{ "NANAYA1.PAL", 0x400,	0x53da6 },
	{ "NANAYA10.PAL", 0x400,	0x53da7 },
	{ "NANAYA11.PAL", 0x400,	0x53da8 },
	{ "NANAYA12.PAL", 0x400,	0x53da9 },
	{ "NANAYA13.PAL", 0x400,	0x53daa },
	{ "NANAYA14.PAL", 0x400,	0x53dab },
	{ "NANAYA15.PAL", 0x400,	0x53dac },
	{ "NANAYA16.PAL", 0x400,	0x53dad },
	{ "NANAYA17.PAL", 0x400,	0x53dae },
	{ "NANAYA18.PAL", 0x400,	0x53daf },
	{ "NANAYA19.PAL", 0x400,	0x53db0 },
	{ "NANAYA2.ARX", 0x1c5080,	0x53db1 },
	{ "NANAYA2.PAL", 0x400,	0x5413c },
	{ "NANAYA20.PAL", 0x400,	0x5413d },
	{ "NANAYA21.PAL", 0x400,	0x5413e },
	{ "NANAYA22.PAL", 0x400,	0x5413f },
	{ "NANAYA23.PAL", 0x400,	0x54140 },
	{ "NANAYA24.PAL", 0x400,	0x54141 },
	{ "NANAYA25.PAL", 0x400,	0x54142 },
	{ "NANAYA26.PAL", 0x400,	0x54143 },
	{ "NANAYA27.PAL", 0x400,	0x54144 },
	{ "NANAYA28.PAL", 0x400,	0x54145 },
	{ "NANAYA29.PAL", 0x400,	0x54146 },
	{ "NANAYA3.PAL", 0x400,	0x54147 },
	{ "NANAYA30.PAL", 0x400,	0x54148 },
	{ "NANAYA31.PAL", 0x400,	0x54149 },
	{ "NANAYA32.PAL", 0x400,	0x5414a },
	{ "NANAYA33.PAL", 0x400,	0x5414b },
	{ "NANAYA34.PAL", 0x400,	0x5414c },
	{ "NANAYA35.PAL", 0x400,	0x5414d },
	{ "NANAYA4.PAL", 0x400,	0x5414e },
	{ "NANAYA5.PAL", 0x400,	0x5414f },
	{ "NANAYA6.PAL", 0x400,	0x54150 },
	{ "NANAYA7.PAL", 0x400,	0x54151 },
	{ "NANAYA8.PAL", 0x400,	0x54152 },
	{ "NANAYA9.PAL", 0x400,	0x54153 },
	{ "NANAYAPAL2.ARX", 0x91c0,	0x54154 },
	{ "NANAYA_0.HA6", 0xbab8,	0x54167 },
	{ "NANAYA_0.TXT", 0xa8,	0x5417f },
	{ "NANAYA_0_C.TXT", 0x1428,	0x54180 },
	{ "NANAYA_1.HA6", 0x175f8,	0x54183 },
	{ "NANAYA_1.TXT", 0xa4,	0x541b2 },
	{ "NANAYA_1_C.TXT", 0x187e,	0x541b3 },
	{ "NANAYA_2.HA6", 0x15d78,	0x541b7 },
	{ "NANAYA_2.TXT", 0xa4,	0x541e3 },
	{ "NANAYA_2_C.TXT", 0x1251,	0x541e4 },
	{ "NANAYA_3.HA6", 0x1100c,	0x541e7 },
	{ "NANAYA_3.TXT", 0xa4,	0x5420a },
	{ "NANAYA_3_C.TXT", 0xe14,	0x5420b },
	{ "NANAYA_9.HA6", 0x16a88,	0x5420d },
	{ "NANAYA_9.TXT", 0xa4,	0x5423b },
	{ "NANAYA_9_C.TXT", 0xd7e,	0x5423c },
	{ "NANAYA_C.TXT", 0xf9d,	0x5423e },
	{ "NANAYA_SELIST.TXT", 0xf36,	0x54240 },
	{ "NANAYA_WIN.TXT", 0x712,	0x54242 },
	{ "NECHAOS.CG2", 0x5fde8,	0x54243 },
	{ "NECHAOS.CPF", 0x2f36c,	0x54303 },
	{ "NECHAOS.DOF", 0x6f67c,	0x54362 },
	{ "NECHAOS.HA6", 0x796d8,	0x54441 },
	{ "NECHAOS.PAT", 0x310b6,	0x54534 },
	{ "NECHAOS.TXT", 0xec,	0x54597 },
	{ "NECHAOS0.PAL", 0x400,	0x54598 },
	{ "NECHAOS1.PAL", 0x400,	0x54599 },
	{ "NECHAOS10.PAL", 0x400,	0x5459a },
	{ "NECHAOS11.PAL", 0x400,	0x5459b },
	{ "NECHAOS12.PAL", 0x400,	0x5459c },
	{ "NECHAOS13.PAL", 0x400,	0x5459d },
	{ "NECHAOS14.PAL", 0x400,	0x5459e },
	{ "NECHAOS15.PAL", 0x400,	0x5459f },
	{ "NECHAOS16.PAL", 0x400,	0x545a0 },
	{ "NECHAOS17.PAL", 0x400,	0x545a1 },
	{ "NECHAOS18.PAL", 0x400,	0x545a2 },
	{ "NECHAOS19.PAL", 0x400,	0x545a3 },
	{ "NECHAOS2.ARX", 0x343bc0,	0x545a4 },
	{ "NECHAOS2.PAL", 0x400,	0x54c2c },
	{ "NECHAOS20.PAL", 0x400,	0x54c2d },
	{ "NECHAOS21.PAL", 0x400,	0x54c2e },
	{ "NECHAOS22.PAL", 0x400,	0x54c2f },
	{ "NECHAOS23.PAL", 0x400,	0x54c30 },
	{ "NECHAOS24.PAL", 0x400,	0x54c31 },
	{ "NECHAOS25.PAL", 0x400,	0x54c32 },
	{ "NECHAOS26.PAL", 0x400,	0x54c33 },
	{ "NECHAOS27.PAL", 0x400,	0x54c34 },
	{ "NECHAOS28.PAL", 0x400,	0x54c35 },
	{ "NECHAOS29.PAL", 0x400,	0x54c36 },
	{ "NECHAOS3.PAL", 0x400,	0x54c37 },
	{ "NECHAOS30.PAL", 0x400,	0x54c38 },
	{ "NECHAOS31.PAL", 0x400,	0x54c39 },
	{ "NECHAOS32.PAL", 0x400,	0x54c3a },
	{ "NECHAOS33.PAL", 0x400,	0x54c3b },
	{ "NECHAOS34.PAL", 0x400,	0x54c3c },
	{ "NECHAOS35.PAL", 0x400,	0x54c3d },
	{ "NECHAOS4.PAL", 0x400,	0x54c3e },
	{ "NECHAOS5.PAL", 0x400,	0x54c3f },
	{ "NECHAOS6.PAL", 0x400,	0x54c40 },
	{ "NECHAOS7.PAL", 0x400,	0x54c41 },
	{ "NECHAOS8.PAL", 0x400,	0x54c42 },
	{ "NECHAOS9.PAL", 0x400,	0x54c43 },
	{ "NECHAOSPAL2.ARX", 0x26b80,	0x54c44 },
	{ "NECHAOS_0.HA6", 0x3564,	0x54c92 },
	{ "NECHAOS_0.TXT", 0xc2,	0x54c99 },
	{ "NECHAOS_0_C.TXT", 0x1240,	0x54c9a },
	{ "NECHAOS_1.HA6", 0x59a0,	0x54c9d },
	{ "NECHAOS_1.TXT", 0xc2,	0x54ca9 },
	{ "NECHAOS_1_C.TXT", 0x1112,	0x54caa },
	{ "NECHAOS_2.HA6", 0x7bdc,	0x54cad },
	{ "NECHAOS_2.TXT", 0xc2,	0x54cbd },
	{ "NECHAOS_2_C.TXT", 0x12e5,	0x54cbe },
	{ "NECHAOS_SELIST.TXT", 0xf5b,	0x54cc1 },
	{ "NECO.CG2", 0x47c08,	0x54cc3 },
	{ "NECO.CPF", 0x2f36c,	0x54d53 },
	{ "NECO.DOF", 0x66304,	0x54db2 },
	{ "NECO.HA6", 0x5cf48,	0x54e7f },
	{ "NECO.PAT", 0x310b6,	0x54f39 },
	{ "NECO.TXT", 0xd3,	0x54f9c },
	{ "NECO0.PAL", 0x400,	0x54f9d },
	{ "NECO1.PAL", 0x400,	0x54f9e },
	{ "NECO10.PAL", 0x400,	0x54f9f },
	{ "NECO11.PAL", 0x400,	0x54fa0 },
	{ "NECO12.PAL", 0x400,	0x54fa1 },
	{ "NECO13.PAL", 0x400,	0x54fa2 },
	{ "NECO14.PAL", 0x400,	0x54fa3 },
	{ "NECO15.PAL", 0x400,	0x54fa4 },
	{ "NECO16.PAL", 0x400,	0x54fa5 },
	{ "NECO17.PAL", 0x400,	0x54fa6 },
	{ "NECO18.PAL", 0x400,	0x54fa7 },
	{ "NECO19.PAL", 0x400,	0x54fa8 },
	{ "NECO2.ARX", 0x2d0c80,	0x54fa9 },
	{ "NECO2.PAL", 0x400,	0x5554b },
	{ "NECO20.PAL", 0x400,	0x5554c },
	{ "NECO21.PAL", 0x400,	0x5554d },
	{ "NECO22.PAL", 0x400,	0x5554e },
	{ "NECO23.PAL", 0x400,	0x5554f },
	{ "NECO24.PAL", 0x400,	0x55550 },
	{ "NECO25.PAL", 0x400,	0x55551 },
	{ "NECO26.PAL", 0x400,	0x55552 },
	{ "NECO27.PAL", 0x400,	0x55553 },
	{ "NECO28.PAL", 0x400,	0x55554 },
	{ "NECO29.PAL", 0x400,	0x55555 },
	{ "NECO3.PAL", 0x400,	0x55556 },
	{ "NECO30.PAL", 0x400,	0x55557 },
	{ "NECO31.PAL", 0x400,	0x55558 },
	{ "NECO32.PAL", 0x400,	0x55559 },
	{ "NECO33.PAL", 0x400,	0x5555a },
	{ "NECO34.PAL", 0x400,	0x5555b },
	{ "NECO35.PAL", 0x400,	0x5555c },
	{ "NECO4.PAL", 0x400,	0x5555d },
	{ "NECO5.PAL", 0x400,	0x5555e },
	{ "NECO6.PAL", 0x400,	0x5555f },
	{ "NECO7.PAL", 0x400,	0x55560 },
	{ "NECO8.PAL", 0x400,	0x55561 },
	{ "NECO9.PAL", 0x400,	0x55562 },
	{ "NECOPAL2.ARX", 0x25280,	0x55563 },
	{ "NECO_0.HA6", 0x34fc,	0x555ae },
	{ "NECO_0.TXT", 0xab,	0x555b5 },
	{ "NECO_0_C.TXT", 0x1a9f,	0x555b6 },
	{ "NECO_1.HA6", 0x90cc,	0x555ba },
	{ "NECO_1.TXT", 0xab,	0x555cd },
	{ "NECO_1_C.TXT", 0x1732,	0x555ce },
	{ "NECO_2.HA6", 0x5ee4,	0x555d1 },
	{ "NECO_2.TXT", 0xab,	0x555dd },
	{ "NECO_2_C.TXT", 0x16d0,	0x555de },
	{ "NECO_3.HA6", 0x1100c,	0x555e1 },
	{ "NECO_9.HA6", 0x13480,	0x55604 },
	{ "NECO_9.TXT", 0xab,	0x5562b },
	{ "NECO_9_C.TXT", 0x165f,	0x5562c },
	{ "NECO_P.CG2", 0x24838,	0x55671 },
	{ "NECO_P.CPF", 0x2f36c,	0x556bb },
	{ "NECO_P.HA6", 0x5a00c,	0x5571a },
	{ "NECO_P.TXT", 0x9e,	0x557cf },
	{ "NECO_P0.PAL", 0x400,	0x557d0 },
	{ "NECO_P1.PAL", 0x400,	0x557d1 },
	{ "NECO_P10.PAL", 0x400,	0x557d2 },
	{ "NECO_P11.PAL", 0x400,	0x557d3 },
	{ "NECO_P12.PAL", 0x400,	0x557d4 },
	{ "NECO_P13.PAL", 0x400,	0x557d5 },
	{ "NECO_P14.PAL", 0x400,	0x557d6 },
	{ "NECO_P15.PAL", 0x400,	0x557d7 },
	{ "NECO_P16.PAL", 0x400,	0x557d8 },
	{ "NECO_P17.PAL", 0x400,	0x557d9 },
	{ "NECO_P18.PAL", 0x400,	0x557da },
	{ "NECO_P19.PAL", 0x400,	0x557db },
	{ "NECO_P2.ARX", 0xbe580,	0x557dc },
	{ "NECO_P2.PAL", 0x400,	0x55959 },
	{ "NECO_P20.PAL", 0x400,	0x5595a },
	{ "NECO_P21.PAL", 0x400,	0x5595b },
	{ "NECO_P22.PAL", 0x400,	0x5595c },
	{ "NECO_P23.PAL", 0x400,	0x5595d },
	{ "NECO_P24.PAL", 0x400,	0x5595e },
	{ "NECO_P25.PAL", 0x400,	0x5595f },
	{ "NECO_P26.PAL", 0x400,	0x55960 },
	{ "NECO_P27.PAL", 0x400,	0x55961 },
	{ "NECO_P28.PAL", 0x400,	0x55962 },
	{ "NECO_P29.PAL", 0x400,	0x55963 },
	{ "NECO_P3.PAL", 0x400,	0x55964 },
	{ "NECO_P30.PAL", 0x400,	0x55965 },
	{ "NECO_P31.PAL", 0x400,	0x55966 },
	{ "NECO_P32.PAL", 0x400,	0x55967 },
	{ "NECO_P33.PAL", 0x400,	0x55968 },
	{ "NECO_P34.PAL", 0x400,	0x55969 },
	{ "NECO_P35.PAL", 0x400,	0x5596a },
	{ "NECO_P4.PAL", 0x400,	0x5596b },
	{ "NECO_P5.PAL", 0x400,	0x5596c },
	{ "NECO_P6.PAL", 0x400,	0x5596d },
	{ "NECO_P7.PAL", 0x400,	0x5596e },
	{ "NECO_P8.PAL", 0x400,	0x5596f },
	{ "NECO_P9.PAL", 0x400,	0x55970 },
	{ "NECO_PPAL2.ARX", 0x9600,	0x55971 },
	{ "NECO_P_0.HA6", 0x354c,	0x55984 },
	{ "NECO_P_0.TXT", 0xb3,	0x5598b },
	{ "NECO_P_0_C.TXT", 0x17b2,	0x5598c },
	{ "NECO_P_1.HA6", 0x94d8,	0x5598f },
	{ "NECO_P_1.TXT", 0xb3,	0x559a2 },
	{ "NECO_P_1_C.TXT", 0x18a2,	0x559a3 },
	{ "NECO_P_2.HA6", 0x86f8,	0x559a7 },
	{ "NECO_P_2.TXT", 0xb3,	0x559b8 },
	{ "NECO_P_2_C.TXT", 0x17b4,	0x559b9 },
	{ "NECO_P_9.HA6", 0x2c,	0x559bc },
	{ "NECO_P_SELIST.TXT", 0xd39,	0x559bd },
	{ "NECO_SELIST.TXT", 0xdaa,	0x559bf },
	{ "NERO.CG2", 0x8c308,	0x559c1 },
	{ "NERO.CPF", 0x2f36c,	0x55ada },
	{ "NERO.DOF", 0x9eaa0,	0x55b39 },
	{ "NERO.HA6", 0x75774,	0x55c77 },
	{ "NERO.PAT", 0x0,	0x55d62 },
	{ "NERO.TXT", 0xc2,	0x55d62 },
	{ "NERO0.PAL", 0x400,	0x55d63 },
	{ "NERO1.PAL", 0x400,	0x55d64 },
	{ "NERO10.PAL", 0x400,	0x55d65 },
	{ "NERO11.PAL", 0x400,	0x55d66 },
	{ "NERO12.PAL", 0x400,	0x55d67 },
	{ "NERO13.PAL", 0x400,	0x55d68 },
	{ "NERO14.PAL", 0x400,	0x55d69 },
	{ "NERO15.PAL", 0x400,	0x55d6a },
	{ "NERO16.PAL", 0x400,	0x55d6b },
	{ "NERO17.PAL", 0x400,	0x55d6c },
	{ "NERO18.PAL", 0x400,	0x55d6d },
	{ "NERO19.PAL", 0x400,	0x55d6e },
	{ "NERO2.ARX", 0x37f700,	0x55d6f },
	{ "NERO2.PAL", 0x400,	0x5646e },
	{ "NERO20.PAL", 0x400,	0x5646f },
	{ "NERO21.PAL", 0x400,	0x56470 },
	{ "NERO22.PAL", 0x400,	0x56471 },
	{ "NERO23.PAL", 0x400,	0x56472 },
	{ "NERO24.PAL", 0x400,	0x56473 },
	{ "NERO25.PAL", 0x400,	0x56474 },
	{ "NERO26.PAL", 0x400,	0x56475 },
	{ "NERO27.PAL", 0x400,	0x56476 },
	{ "NERO28.PAL", 0x400,	0x56477 },
	{ "NERO29.PAL", 0x400,	0x56478 },
	{ "NERO3.PAL", 0x400,	0x56479 },
	{ "NERO30.PAL", 0x400,	0x5647a },
	{ "NERO31.PAL", 0x400,	0x5647b },
	{ "NERO32.PAL", 0x400,	0x5647c },
	{ "NERO33.PAL", 0x400,	0x5647d },
	{ "NERO34.PAL", 0x400,	0x5647e },
	{ "NERO35.PAL", 0x400,	0x5647f },
	{ "NERO4.PAL", 0x400,	0x56480 },
	{ "NERO5.PAL", 0x400,	0x56481 },
	{ "NERO6.PAL", 0x400,	0x56482 },
	{ "NERO7.PAL", 0x400,	0x56483 },
	{ "NERO8.PAL", 0x400,	0x56484 },
	{ "NERO9.PAL", 0x400,	0x56485 },
	{ "NEROPAL2.ARX", 0x116c0,	0x56486 },
	{ "NERO_0.HA6", 0xb8e4,	0x564a9 },
	{ "NERO_0.TXT", 0x9c,	0x564c1 },
	{ "NERO_0_C.TXT", 0xfaf,	0x564c2 },
	{ "NERO_1.HA6", 0x172cc,	0x564c4 },
	{ "NERO_1.TXT", 0x9c,	0x564f3 },
	{ "NERO_1_C.TXT", 0xe79,	0x564f4 },
	{ "NERO_2.HA6", 0x14808,	0x564f6 },
	{ "NERO_2.TXT", 0x9c,	0x56520 },
	{ "NERO_2_C.TXT", 0xfbe,	0x56521 },
	{ "NERO_3.HA6", 0x1100c,	0x56523 },
	{ "NERO_9.HA6", 0x1fdec,	0x56546 },
	{ "NERO_9.TXT", 0x9c,	0x56586 },
	{ "NERO_9_C.TXT", 0xdc4,	0x56587 },
	{ "NERO_SELIST.TXT", 0xb41,	0x56589 },
	{ "NERO_WIN.TXT", 0x72c,	0x5658b },
	{ "RIES.CG2", 0x7d998,	0x56751 },
	{ "RIES.CPF", 0x2f36c,	0x5684d },
	{ "RIES.HA6", 0x310c8,	0x568ac },
	{ "RIES.PAT", 0xd120,	0x5690f },
	{ "RIES.TXT", 0xd3,	0x5692a },
	{ "RIES0.PAL", 0x400,	0x5692b },
	{ "RIES1.PAL", 0x400,	0x5692c },
	{ "RIES10.PAL", 0x400,	0x5692d },
	{ "RIES11.PAL", 0x400,	0x5692e },
	{ "RIES12.PAL", 0x400,	0x5692f },
	{ "RIES13.PAL", 0x400,	0x56930 },
	{ "RIES14.PAL", 0x400,	0x56931 },
	{ "RIES15.PAL", 0x400,	0x56932 },
	{ "RIES16.PAL", 0x400,	0x56933 },
	{ "RIES17.PAL", 0x400,	0x56934 },
	{ "RIES18.PAL", 0x400,	0x56935 },
	{ "RIES19.PAL", 0x400,	0x56936 },
	{ "RIES2.ARX", 0x39f980,	0x56937 },
	{ "RIES2.PAL", 0x400,	0x57077 },
	{ "RIES20.PAL", 0x400,	0x57078 },
	{ "RIES21.PAL", 0x400,	0x57079 },
	{ "RIES22.PAL", 0x400,	0x5707a },
	{ "RIES23.PAL", 0x400,	0x5707b },
	{ "RIES24.PAL", 0x400,	0x5707c },
	{ "RIES25.PAL", 0x400,	0x5707d },
	{ "RIES26.PAL", 0x400,	0x5707e },
	{ "RIES27.PAL", 0x400,	0x5707f },
	{ "RIES28.PAL", 0x400,	0x57080 },
	{ "RIES29.PAL", 0x400,	0x57081 },
	{ "RIES3.PAL", 0x400,	0x57082 },
	{ "RIES30.PAL", 0x400,	0x57083 },
	{ "RIES31.PAL", 0x400,	0x57084 },
	{ "RIES32.PAL", 0x400,	0x57085 },
	{ "RIES33.PAL", 0x400,	0x57086 },
	{ "RIES34.PAL", 0x400,	0x57087 },
	{ "RIES35.PAL", 0x400,	0x57088 },
	{ "RIES4.PAL", 0x400,	0x57089 },
	{ "RIES5.PAL", 0x400,	0x5708a },
	{ "RIES6.PAL", 0x400,	0x5708b },
	{ "RIES7.PAL", 0x400,	0x5708c },
	{ "RIES8.PAL", 0x400,	0x5708d },
	{ "RIES9.PAL", 0x400,	0x5708e },
	{ "RIESPAL2.ARX", 0x19780,	0x5708f },
	{ "RIES_0.HA6", 0x12fc0,	0x570c2 },
	{ "RIES_0.TXT", 0xb5,	0x570e8 },
	{ "RIES_0_C.TXT", 0x15ab,	0x570e9 },
	{ "RIES_1.HA6", 0x14cd8,	0x570ec },
	{ "RIES_1.TXT", 0xaf,	0x57116 },
	{ "RIES_1_C.TXT", 0xf3c,	0x57117 },
	{ "RIES_2.HA6", 0x13e80,	0x57119 },
	{ "RIES_2.TXT", 0xad,	0x57141 },
	{ "RIES_2_C.TXT", 0xf10,	0x57142 },
	{ "RIES_3.HA6", 0x1100c,	0x57144 },
	{ "RIES_3.TXT", 0xad,	0x57167 },
	{ "RIES_3_C.TXT", 0xb71,	0x57168 },
	{ "RIES_9.HA6", 0x2b9c8,	0x5716a },
	{ "RIES_9.TXT", 0xad,	0x571c2 },
	{ "RIES_9_C.TXT", 0xa18,	0x571c3 },
	{ "RIES_SELIST.TXT", 0xd9f,	0x571c5 },
	{ "ROA.CG2", 0x60550,	0x571e8 },
	{ "ROA.CPF", 0x2f36c,	0x572a9 },
	{ "ROA.HA6", 0x6f0a0,	0x57308 },
	{ "ROA.PAT", 0x0,	0x573e7 },
	{ "ROA.TXT", 0xbc,	0x573e7 },
	{ "ROA0.PAL", 0x400,	0x573e8 },
	{ "ROA1.PAL", 0x400,	0x573e9 },
	{ "ROA10.PAL", 0x400,	0x573ea },
	{ "ROA11.PAL", 0x400,	0x573eb },
	{ "ROA12.PAL", 0x400,	0x573ec },
	{ "ROA13.PAL", 0x400,	0x573ed },
	{ "ROA14.PAL", 0x400,	0x573ee },
	{ "ROA15.PAL", 0x400,	0x573ef },
	{ "ROA16.PAL", 0x400,	0x573f0 },
	{ "ROA17.PAL", 0x400,	0x573f1 },
	{ "ROA18.PAL", 0x400,	0x573f2 },
	{ "ROA19.PAL", 0x400,	0x573f3 },
	{ "ROA2.ARX", 0x38acc0,	0x573f4 },
	{ "ROA2.PAL", 0x400,	0x57b0a },
	{ "ROA20.PAL", 0x400,	0x57b0b },
	{ "ROA21.PAL", 0x400,	0x57b0c },
	{ "ROA22.PAL", 0x400,	0x57b0d },
	{ "ROA23.PAL", 0x400,	0x57b0e },
	{ "ROA24.PAL", 0x400,	0x57b0f },
	{ "ROA25.PAL", 0x400,	0x57b10 },
	{ "ROA26.PAL", 0x400,	0x57b11 },
	{ "ROA27.PAL", 0x400,	0x57b12 },
	{ "ROA28.PAL", 0x400,	0x57b13 },
	{ "ROA29.PAL", 0x400,	0x57b14 },
	{ "ROA3.PAL", 0x400,	0x57b15 },
	{ "ROA30.PAL", 0x400,	0x57b16 },
	{ "ROA31.PAL", 0x400,	0x57b17 },
	{ "ROA32.PAL", 0x400,	0x57b18 },
	{ "ROA33.PAL", 0x400,	0x57b19 },
	{ "ROA34.PAL", 0x400,	0x57b1a },
	{ "ROA35.PAL", 0x400,	0x57b1b },
	{ "ROA4.PAL", 0x400,	0x57b1c },
	{ "ROA5.PAL", 0x400,	0x57b1d },
	{ "ROA6.PAL", 0x400,	0x57b1e },
	{ "ROA7.PAL", 0x400,	0x57b1f },
	{ "ROA8.PAL", 0x400,	0x57b20 },
	{ "ROA9.PAL", 0x400,	0x57b21 },
	{ "ROAPAL2.ARX", 0x239c0,	0x57b22 },
	{ "ROA_0.HA6", 0x3e5c,	0x57b6a },
	{ "ROA_0.TXT", 0x98,	0x57b72 },
	{ "ROA_0_C.TXT", 0x1547,	0x57b73 },
	{ "ROA_1.HA6", 0x9320,	0x57b76 },
	{ "ROA_1.TXT", 0x98,	0x57b89 },
	{ "ROA_1_C.TXT", 0x109e,	0x57b8a },
	{ "ROA_2.HA6", 0x593c,	0x57b8d },
	{ "ROA_2.TXT", 0x98,	0x57b99 },
	{ "ROA_2_C.TXT", 0x1094,	0x57b9a },
	{ "ROA_3.HA6", 0x1100c,	0x57b9d },
	{ "ROA_9.HA6", 0x15548,	0x57bc0 },
	{ "ROA_9.TXT", 0x98,	0x57beb },
	{ "ROA_9_C.TXT", 0xf88,	0x57bec },
	{ "ROA_SELIST.TXT", 0xc76,	0x57bee },
	{ "RYOUGI.CG2", 0x7abc0,	0x57bf0 },
	{ "RYOUGI.CPF", 0x2f36c,	0x57ce6 },
	{ "RYOUGI.HA6", 0xb4ff4,	0x57d45 },
	{ "RYOUGI.PAT", 0xd506,	0x57eaf },
	{ "RYOUGI0.PAL", 0x400,	0x57eca },
	{ "RYOUGI1.PAL", 0x400,	0x57ecb },
	{ "RYOUGI10.PAL", 0x400,	0x57ecc },
	{ "RYOUGI11.PAL", 0x400,	0x57ecd },
	{ "RYOUGI12.PAL", 0x400,	0x57ece },
	{ "RYOUGI13.PAL", 0x400,	0x57ecf },
	{ "RYOUGI14.PAL", 0x400,	0x57ed0 },
	{ "RYOUGI15.PAL", 0x400,	0x57ed1 },
	{ "RYOUGI16.PAL", 0x400,	0x57ed2 },
	{ "RYOUGI17.PAL", 0x400,	0x57ed3 },
	{ "RYOUGI18.PAL", 0x400,	0x57ed4 },
	{ "RYOUGI19.PAL", 0x400,	0x57ed5 },
	{ "RYOUGI2.ARX", 0x2f94c0,	0x57ed6 },
	{ "RYOUGI2.PAL", 0x400,	0x584c9 },
	{ "RYOUGI20.PAL", 0x400,	0x584ca },
	{ "RYOUGI21.PAL", 0x400,	0x584cb },
	{ "RYOUGI22.PAL", 0x400,	0x584cc },
	{ "RYOUGI23.PAL", 0x400,	0x584cd },
	{ "RYOUGI24.PAL", 0x400,	0x584ce },
	{ "RYOUGI25.PAL", 0x400,	0x584cf },
	{ "RYOUGI26.PAL", 0x400,	0x584d0 },
	{ "RYOUGI27.PAL", 0x400,	0x584d1 },
	{ "RYOUGI28.PAL", 0x400,	0x584d2 },
	{ "RYOUGI29.PAL", 0x400,	0x584d3 },
	{ "RYOUGI3.PAL", 0x400,	0x584d4 },
	{ "RYOUGI30.PAL", 0x400,	0x584d5 },
	{ "RYOUGI31.PAL", 0x400,	0x584d6 },
	{ "RYOUGI32.PAL", 0x400,	0x584d7 },
	{ "RYOUGI33.PAL", 0x400,	0x584d8 },
	{ "RYOUGI34.PAL", 0x400,	0x584d9 },
	{ "RYOUGI35.PAL", 0x400,	0x584da },
	{ "RYOUGI4.PAL", 0x400,	0x584db },
	{ "RYOUGI5.PAL", 0x400,	0x584dc },
	{ "RYOUGI6.PAL", 0x400,	0x584dd },
	{ "RYOUGI7.PAL", 0x400,	0x584de },
	{ "RYOUGI8.PAL", 0x400,	0x584df },
	{ "RYOUGI9.PAL", 0x400,	0x584e0 },
	{ "RYOUGIPAL2.ARX", 0x13800,	0x584e1 },
	{ "RYOUGI_0.HA6", 0x59a8,	0x58508 },
	{ "RYOUGI_0.TXT", 0xc1,	0x58514 },
	{ "RYOUGI_0_C.TXT", 0x2a6b,	0x58515 },
	{ "RYOUGI_1.HA6", 0x7a08,	0x5851b },
	{ "RYOUGI_1.TXT", 0xbf,	0x5852b },
	{ "RYOUGI_1_C.TXT", 0x1f25,	0x5852c },
	{ "RYOUGI_2.HA6", 0x96dc,	0x58530 },
	{ "RYOUGI_2.TXT", 0xc1,	0x58543 },
	{ "RYOUGI_2_C.TXT", 0x21cd,	0x58544 },
	{ "RYOUGI_9.HA6", 0x19594,	0x58549 },
	{ "RYOUGI_9.TXT", 0xc1,	0x5857c },
	{ "RYOUGI_9_C.TXT", 0x1f6f,	0x5857d },
	{ "RYOUGI_SELIST.TXT", 0x1e59,	0x58581 },
	{ "SAMPLE_C.TXT", 0xa9c,	0x58585 },
	{ "SATSUKI.CG2", 0x5f4b8,	0x58587 },
	{ "SATSUKI.CPF", 0x2f36c,	0x58646 },
	{ "SATSUKI.DOF", 0x6f4d8,	0x586a5 },
	{ "SATSUKI.HA6", 0x646c0,	0x58784 },
	{ "SATSUKI.LST", 0x8b68,	0x5884d },
	{ "SATSUKI.PAT", 0x0,	0x5885f },
	{ "SATSUKI.TXT", 0xd4,	0x5885f },
	{ "SATSUKI0.PAL", 0x400,	0x58860 },
	{ "SATSUKI1.PAL", 0x400,	0x58861 },
	{ "SATSUKI10.PAL", 0x400,	0x58862 },
	{ "SATSUKI11.PAL", 0x400,	0x58863 },
	{ "SATSUKI12.PAL", 0x400,	0x58864 },
	{ "SATSUKI13.PAL", 0x400,	0x58865 },
	{ "SATSUKI14.PAL", 0x400,	0x58866 },
	{ "SATSUKI15.PAL", 0x400,	0x58867 },
	{ "SATSUKI16.PAL", 0x400,	0x58868 },
	{ "SATSUKI17.PAL", 0x400,	0x58869 },
	{ "SATSUKI18.PAL", 0x400,	0x5886a },
	{ "SATSUKI19.PAL", 0x400,	0x5886b },
	{ "SATSUKI2.ARX", 0x3266c0,	0x5886c },
	{ "SATSUKI2.PAL", 0x400,	0x58eb9 },
	{ "SATSUKI20.PAL", 0x400,	0x58eba },
	{ "SATSUKI21.PAL", 0x400,	0x58ebb },
	{ "SATSUKI22.PAL", 0x400,	0x58ebc },
	{ "SATSUKI23.PAL", 0x400,	0x58ebd },
	{ "SATSUKI24.PAL", 0x400,	0x58ebe },
	{ "SATSUKI25.PAL", 0x400,	0x58ebf },
	{ "SATSUKI26.PAL", 0x400,	0x58ec0 },
	{ "SATSUKI27.PAL", 0x400,	0x58ec1 },
	{ "SATSUKI28.PAL", 0x400,	0x58ec2 },
	{ "SATSUKI29.PAL", 0x400,	0x58ec3 },
	{ "SATSUKI3.PAL", 0x400,	0x58ec4 },
	{ "SATSUKI30.PAL", 0x400,	0x58ec5 },
	{ "SATSUKI31.PAL", 0x400,	0x58ec6 },
	{ "SATSUKI32.PAL", 0x400,	0x58ec7 },
	{ "SATSUKI33.PAL", 0x400,	0x58ec8 },
	{ "SATSUKI34.PAL", 0x400,	0x58ec9 },
	{ "SATSUKI35.PAL", 0x400,	0x58eca },
	{ "SATSUKI4.PAL", 0x400,	0x58ecb },
	{ "SATSUKI5.PAL", 0x400,	0x58ecc },
	{ "SATSUKI6.PAL", 0x400,	0x58ecd },
	{ "SATSUKI7.PAL", 0x400,	0x58ece },
	{ "SATSUKI8.PAL", 0x400,	0x58ecf },
	{ "SATSUKI9.PAL", 0x400,	0x58ed0 },
	{ "SATSUKIPAL2.ARX", 0x21440,	0x58ed1 },
	{ "SATSUKI_0.HA6", 0x817c,	0x58f14 },
	{ "SATSUKI_0.LST", 0x8b68,	0x58f25 },
	{ "SATSUKI_0.TXT", 0xaa,	0x58f37 },
	{ "SATSUKI_0_C.TXT", 0x1299,	0x58f38 },
	{ "SATSUKI_1.HA6", 0xae68,	0x58f3b },
	{ "SATSUKI_1.LST", 0x8b68,	0x58f51 },
	{ "SATSUKI_1.TXT", 0xaa,	0x58f63 },
	{ "SATSUKI_1_C.TXT", 0x107d,	0x58f64 },
	{ "SATSUKI_2.HA6", 0xe7b4,	0x58f67 },
	{ "SATSUKI_2.LST", 0x8b68,	0x58f84 },
	{ "SATSUKI_2.TXT", 0xa8,	0x58f96 },
	{ "SATSUKI_2_C.TXT", 0xfdc,	0x58f97 },
	{ "SATSUKI_3.HA6", 0x1100c,	0x58f99 },
	{ "SATSUKI_3.TXT", 0xa8,	0x58fbc },
	{ "SATSUKI_3_C.TXT", 0xd8c,	0x58fbd },
	{ "SATSUKI_9.HA6", 0x1a160,	0x58fbf },
	{ "SATSUKI_9.TXT", 0xa8,	0x58ff4 },
	{ "SATSUKI_9_C.TXT", 0xe91,	0x58ff5 },
	{ "SATSUKI_C.TXT", 0xfc9,	0x58ff7 },
	{ "SATSUKI_SELIST.TXT", 0xf14,	0x58ff9 },
	{ "SATSUKI_WIN.TXT", 0x92c,	0x58ffb },
	{ "SHIKI.CG2", 0x5a7a8,	0x58ffd },
	{ "SHIKI.CPF", 0x2f36c,	0x590b2 },
	{ "SHIKI.HA6", 0x79d78,	0x59111 },
	{ "SHIKI.PAT", 0x0,	0x59205 },
	{ "SHIKI.TXT", 0xca,	0x59205 },
	{ "SHIKI0.PAL", 0x400,	0x59206 },
	{ "SHIKI1.PAL", 0x400,	0x59207 },
	{ "SHIKI10.PAL", 0x400,	0x59208 },
	{ "SHIKI11.PAL", 0x400,	0x59209 },
	{ "SHIKI12.PAL", 0x400,	0x5920a },
	{ "SHIKI13.PAL", 0x400,	0x5920b },
	{ "SHIKI14.PAL", 0x400,	0x5920c },
	{ "SHIKI15.PAL", 0x400,	0x5920d },
	{ "SHIKI16.PAL", 0x400,	0x5920e },
	{ "SHIKI17.PAL", 0x400,	0x5920f },
	{ "SHIKI18.PAL", 0x400,	0x59210 },
	{ "SHIKI19.PAL", 0x400,	0x59211 },
	{ "SHIKI2.ARX", 0x1d4640,	0x59212 },
	{ "SHIKI2.PAL", 0x400,	0x595bb },
	{ "SHIKI20.PAL", 0x400,	0x595bc },
	{ "SHIKI21.PAL", 0x400,	0x595bd },
	{ "SHIKI22.PAL", 0x400,	0x595be },
	{ "SHIKI23.PAL", 0x400,	0x595bf },
	{ "SHIKI24.PAL", 0x400,	0x595c0 },
	{ "SHIKI25.PAL", 0x400,	0x595c1 },
	{ "SHIKI26.PAL", 0x400,	0x595c2 },
	{ "SHIKI27.PAL", 0x400,	0x595c3 },
	{ "SHIKI28.PAL", 0x400,	0x595c4 },
	{ "SHIKI29.PAL", 0x400,	0x595c5 },
	{ "SHIKI3.PAL", 0x400,	0x595c6 },
	{ "SHIKI30.PAL", 0x400,	0x595c7 },
	{ "SHIKI31.PAL", 0x400,	0x595c8 },
	{ "SHIKI32.PAL", 0x400,	0x595c9 },
	{ "SHIKI33.PAL", 0x400,	0x595ca },
	{ "SHIKI34.PAL", 0x400,	0x595cb },
	{ "SHIKI35.PAL", 0x400,	0x595cc },
	{ "SHIKI4.PAL", 0x400,	0x595cd },
	{ "SHIKI5.PAL", 0x400,	0x595ce },
	{ "SHIKI6.PAL", 0x400,	0x595cf },
	{ "SHIKI7.PAL", 0x400,	0x595d0 },
	{ "SHIKI8.PAL", 0x400,	0x595d1 },
	{ "SHIKI9.PAL", 0x400,	0x595d2 },
	{ "SHIKIPAL2.ARX", 0x8d80,	0x595d3 },
	{ "SHIKI_0.HA6", 0x28d88,	0x595e5 },
	{ "SHIKI_0.TXT", 0xa6,	0x59637 },
	{ "SHIKI_0_C.TXT", 0x1397,	0x59638 },
	{ "SHIKI_1.HA6", 0xea70,	0x5963b },
	{ "SHIKI_1.TXT", 0xa6,	0x59659 },
	{ "SHIKI_1_C.TXT", 0x10d0,	0x5965a },
	{ "SHIKI_2.HA6", 0x16e7c,	0x5965d },
	{ "SHIKI_2.TXT", 0xa6,	0x5968b },
	{ "SHIKI_2_C.TXT", 0xfee,	0x5968c },
	{ "SHIKI_3.HA6", 0x1100c,	0x5968e },
	{ "SHIKI_3.TXT", 0xa6,	0x596b1 },
	{ "SHIKI_3_C.TXT", 0xe65,	0x596b2 },
	{ "SHIKI_9.HA6", 0x2425c,	0x596b4 },
	{ "SHIKI_9.TXT", 0xa6,	0x596fd },
	{ "SHIKI_9_C.TXT", 0xe81,	0x596fe },
	{ "SHIKI_C.TXT", 0x1017,	0x59700 },
	{ "SHIKI_SELIST.TXT", 0xfed,	0x59703 },
	{ "SION.CG2", 0x5f278,	0x59705 },
	{ "SION.CPF", 0x2f36c,	0x597c4 },
	{ "SION.HA6", 0x57a24,	0x59823 },
	{ "SION.PAT", 0x0,	0x598d3 },
	{ "SION.TXT", 0xd8,	0x598d3 },
	{ "SION0.PAL", 0x400,	0x598d4 },
	{ "SION1.PAL", 0x400,	0x598d5 },
	{ "SION10.PAL", 0x400,	0x598d6 },
	{ "SION11.PAL", 0x400,	0x598d7 },
	{ "SION12.PAL", 0x400,	0x598d8 },
	{ "SION13.PAL", 0x400,	0x598d9 },
	{ "SION14.PAL", 0x400,	0x598da },
	{ "SION15.PAL", 0x400,	0x598db },
	{ "SION16.PAL", 0x400,	0x598dc },
	{ "SION17.PAL", 0x400,	0x598dd },
	{ "SION18.PAL", 0x400,	0x598de },
	{ "SION19.PAL", 0x400,	0x598df },
	{ "SION2.ARX", 0x263300,	0x598e0 },
	{ "SION2.PAL", 0x400,	0x59da7 },
	{ "SION20.PAL", 0x400,	0x59da8 },
	{ "SION21.PAL", 0x400,	0x59da9 },
	{ "SION22.PAL", 0x400,	0x59daa },
	{ "SION23.PAL", 0x400,	0x59dab },
	{ "SION24.PAL", 0x400,	0x59dac },
	{ "SION25.PAL", 0x400,	0x59dad },
	{ "SION26.PAL", 0x400,	0x59dae },
	{ "SION27.PAL", 0x400,	0x59daf },
	{ "SION28.PAL", 0x400,	0x59db0 },
	{ "SION29.PAL", 0x400,	0x59db1 },
	{ "SION3.PAL", 0x400,	0x59db2 },
	{ "SION30.PAL", 0x400,	0x59db3 },
	{ "SION31.PAL", 0x400,	0x59db4 },
	{ "SION32.PAL", 0x400,	0x59db5 },
	{ "SION33.PAL", 0x400,	0x59db6 },
	{ "SION34.PAL", 0x400,	0x59db7 },
	{ "SION35.PAL", 0x400,	0x59db8 },
	{ "SION4.PAL", 0x400,	0x59db9 },
	{ "SION5.PAL", 0x400,	0x59dba },
	{ "SION6.PAL", 0x400,	0x59dbb },
	{ "SION7.PAL", 0x400,	0x59dbc },
	{ "SION8.PAL", 0x400,	0x59dbd },
	{ "SION9.PAL", 0x400,	0x59dbe },
	{ "SIONPAL2.ARX", 0x10a40,	0x59dbf },
	{ "SION_0.HA6", 0xdacc,	0x59de1 },
	{ "SION_0.TXT", 0x98,	0x59dfd },
	{ "SION_0_C.TXT", 0x1571,	0x59dfe },
	{ "SION_1.HA6", 0x121d4,	0x59e01 },
	{ "SION_1.TXT", 0x98,	0x59e26 },
	{ "SION_1_C.TXT", 0x1339,	0x59e27 },
	{ "SION_2.HA6", 0xbf28,	0x59e2a },
	{ "SION_2.TXT", 0x9c,	0x59e42 },
	{ "SION_2_C.TXT", 0x1316,	0x59e43 },
	{ "SION_3.HA6", 0x16678,	0x59e46 },
	{ "SION_3.TXT", 0x9c,	0x59e73 },
	{ "SION_3_C.TXT", 0x1018,	0x59e74 },
	{ "SION_9.HA6", 0x2b634,	0x59e77 },
	{ "SION_9.TXT", 0x9c,	0x59ece },
	{ "SION_9_C.TXT", 0x1118,	0x59ecf },
	{ "SION_C.TXT", 0x118a,	0x59ed2 },
	{ "SION_SELIST.TXT", 0xe7d,	0x59ed5 },
	{ "SION_WIN.TXT", 0x942,	0x59ed7 },
	{ "S_AKIHA.CG2", 0x72a18,	0x59ed9 },
	{ "S_AKIHA.CPF", 0x2f36c,	0x59fbf },
	{ "S_AKIHA.HA6", 0x7ae6c,	0x5a01e },
	{ "S_AKIHA.PAT", 0x0,	0x5a114 },
	{ "S_AKIHA0.PAL", 0x400,	0x5a114 },
	{ "S_AKIHA1.PAL", 0x400,	0x5a115 },
	{ "S_AKIHA10.PAL", 0x400,	0x5a116 },
	{ "S_AKIHA11.PAL", 0x400,	0x5a117 },
	{ "S_AKIHA12.PAL", 0x400,	0x5a118 },
	{ "S_AKIHA13.PAL", 0x400,	0x5a119 },
	{ "S_AKIHA14.PAL", 0x400,	0x5a11a },
	{ "S_AKIHA15.PAL", 0x400,	0x5a11b },
	{ "S_AKIHA16.PAL", 0x400,	0x5a11c },
	{ "S_AKIHA17.PAL", 0x400,	0x5a11d },
	{ "S_AKIHA18.PAL", 0x400,	0x5a11e },
	{ "S_AKIHA19.PAL", 0x400,	0x5a11f },
	{ "S_AKIHA2.ARX", 0x3868c0,	0x5a120 },
	{ "S_AKIHA2.PAL", 0x400,	0x5a82e },
	{ "S_AKIHA20.PAL", 0x400,	0x5a82f },
	{ "S_AKIHA21.PAL", 0x400,	0x5a830 },
	{ "S_AKIHA22.PAL", 0x400,	0x5a831 },
	{ "S_AKIHA23.PAL", 0x400,	0x5a832 },
	{ "S_AKIHA24.PAL", 0x400,	0x5a833 },
	{ "S_AKIHA25.PAL", 0x400,	0x5a834 },
	{ "S_AKIHA26.PAL", 0x400,	0x5a835 },
	{ "S_AKIHA27.PAL", 0x400,	0x5a836 },
	{ "S_AKIHA28.PAL", 0x400,	0x5a837 },
	{ "S_AKIHA29.PAL", 0x400,	0x5a838 },
	{ "S_AKIHA3.PAL", 0x400,	0x5a839 },
	{ "S_AKIHA30.PAL", 0x400,	0x5a83a },
	{ "S_AKIHA31.PAL", 0x400,	0x5a83b },
	{ "S_AKIHA32.PAL", 0x400,	0x5a83c },
	{ "S_AKIHA33.PAL", 0x400,	0x5a83d },
	{ "S_AKIHA34.PAL", 0x400,	0x5a83e },
	{ "S_AKIHA35.PAL", 0x400,	0x5a83f },
	{ "S_AKIHA4.PAL", 0x400,	0x5a840 },
	{ "S_AKIHA5.PAL", 0x400,	0x5a841 },
	{ "S_AKIHA6.PAL", 0x400,	0x5a842 },
	{ "S_AKIHA7.PAL", 0x400,	0x5a843 },
	{ "S_AKIHA8.PAL", 0x400,	0x5a844 },
	{ "S_AKIHA9.PAL", 0x400,	0x5a845 },
	{ "S_AKIHAPAL2.ARX", 0x20c00,	0x5a846 },
	{ "S_AKIHA_0.HA6", 0x4b2c,	0x5a888 },
	{ "S_AKIHA_0.TXT", 0xb6,	0x5a892 },
	{ "S_AKIHA_0_C.TXT", 0x1a88,	0x5a893 },
	{ "S_AKIHA_1.HA6", 0x9d54,	0x5a897 },
	{ "S_AKIHA_1.TXT", 0xb6,	0x5a8ab },
	{ "S_AKIHA_1_C.TXT", 0x1452,	0x5a8ac },
	{ "S_AKIHA_2.HA6", 0x9bd4,	0x5a8af },
	{ "S_AKIHA_2.TXT", 0xb6,	0x5a8c3 },
	{ "S_AKIHA_2_C.TXT", 0x13f5,	0x5a8c4 },
	{ "S_AKIHA_SELIST.TXT", 0x1338,	0x5a8c7 },
	{ "TEMPLATE.CPF", 0x2f36c,	0x5a8ca },
	{ "VECTOR.TXT", 0x196f,	0x5a94a },
	{ "VECTORLIST.TXT", 0x30d5,	0x5a94e },
	{ "VECTORSAMPLE.TXT", 0x9c0,	0x5a955 },
	{ "V_SION.CG2", 0x7a0e0,	0x5a957 },
	{ "V_SION.CPF", 0x2f36c,	0x5aa4c },
	{ "V_SION.HA6", 0x6961c,	0x5aaab },
	{ "V_SION.LST", 0x137c8,	0x5ab7e },
	{ "V_SION.PAT", 0xbe38,	0x5aba5 },
	{ "V_SION.TXT", 0xe1,	0x5abbd },
	{ "V_SION0.PAL", 0x400,	0x5abbe },
	{ "V_SION1.PAL", 0x400,	0x5abbf },
	{ "V_SION10.PAL", 0x400,	0x5abc0 },
	{ "V_SION11.PAL", 0x400,	0x5abc1 },
	{ "V_SION12.PAL", 0x400,	0x5abc2 },
	{ "V_SION13.PAL", 0x400,	0x5abc3 },
	{ "V_SION14.PAL", 0x400,	0x5abc4 },
	{ "V_SION15.PAL", 0x400,	0x5abc5 },
	{ "V_SION16.PAL", 0x400,	0x5abc6 },
	{ "V_SION17.PAL", 0x400,	0x5abc7 },
	{ "V_SION18.PAL", 0x400,	0x5abc8 },
	{ "V_SION19.PAL", 0x400,	0x5abc9 },
	{ "V_SION2.ARX", 0x3b9d00,	0x5abca },
	{ "V_SION2.PAL", 0x400,	0x5b33e },
	{ "V_SION20.PAL", 0x400,	0x5b33f },
	{ "V_SION21.PAL", 0x400,	0x5b340 },
	{ "V_SION22.PAL", 0x400,	0x5b341 },
	{ "V_SION23.PAL", 0x400,	0x5b342 },
	{ "V_SION24.PAL", 0x400,	0x5b343 },
	{ "V_SION25.PAL", 0x400,	0x5b344 },
	{ "V_SION26.PAL", 0x400,	0x5b345 },
	{ "V_SION27.PAL", 0x400,	0x5b346 },
	{ "V_SION28.PAL", 0x400,	0x5b347 },
	{ "V_SION29.PAL", 0x400,	0x5b348 },
	{ "V_SION3.PAL", 0x400,	0x5b349 },
	{ "V_SION30.PAL", 0x400,	0x5b34a },
	{ "V_SION31.PAL", 0x400,	0x5b34b },
	{ "V_SION32.PAL", 0x400,	0x5b34c },
	{ "V_SION33.PAL", 0x400,	0x5b34d },
	{ "V_SION34.PAL", 0x400,	0x5b34e },
	{ "V_SION35.PAL", 0x400,	0x5b34f },
	{ "V_SION4.PAL", 0x400,	0x5b350 },
	{ "V_SION5.PAL", 0x400,	0x5b351 },
	{ "V_SION6.PAL", 0x400,	0x5b352 },
	{ "V_SION7.PAL", 0x400,	0x5b353 },
	{ "V_SION8.PAL", 0x400,	0x5b354 },
	{ "V_SION9.PAL", 0x400,	0x5b355 },
	{ "V_SIONPAL2.ARX", 0x19bc0,	0x5b356 },
	{ "V_SION_0.HA6", 0xa590,	0x5b38a },
	{ "V_SION_0.LST", 0x137c8,	0x5b39f },
	{ "V_SION_0.TXT", 0xb5,	0x5b3c6 },
	{ "V_SION_0_C.TXT", 0x1692,	0x5b3c7 },
	{ "V_SION_1.HA6", 0x13c94,	0x5b3ca },
	{ "V_SION_1.LST", 0x137c8,	0x5b3f2 },
	{ "V_SION_1.TXT", 0xc2,	0x5b419 },
	{ "V_SION_1_C.TXT", 0x174a,	0x5b41a },
	{ "V_SION_2.HA6", 0x99cc,	0x5b41d },
	{ "V_SION_2.LST", 0x137c8,	0x5b431 },
	{ "V_SION_2.TXT", 0xb7,	0x5b458 },
	{ "V_SION_2_C.TXT", 0x1305,	0x5b459 },
	{ "V_SION_3.HA6", 0x1100c,	0x5b45c },
	{ "V_SION_3.TXT", 0xb7,	0x5b47f },
	{ "V_SION_3_C.TXT", 0x1017,	0x5b480 },
	{ "V_SION_9.HA6", 0x214d8,	0x5b483 },
	{ "V_SION_9.TXT", 0xb7,	0x5b4c6 },
	{ "V_SION_9_C.TXT", 0x1042,	0x5b4c7 },
	{ "V_SION_C.TXT", 0x1200,	0x5b4ca },
	{ "V_SION_SELIST.TXT", 0xf25,	0x5b4cd },
	{ "V_SION_WIN.TXT", 0x556,	0x5b4cf },
	{ "WARAKIA.CG2", 0x72e38,	0x5b4d0 },
	{ "WARAKIA.CPF", 0x2f36c,	0x5b5b6 },
	{ "WARAKIA.HA6", 0x67d24,	0x5b615 },
	{ "WARAKIA.PAT", 0xc39a,	0x5b6e5 },
	{ "WARAKIA.TXT", 0xe8,	0x5b6fe },
	{ "WARAKIA0.PAL", 0x400,	0x5b6ff },
	{ "WARAKIA1.PAL", 0x400,	0x5b700 },
	{ "WARAKIA10.PAL", 0x400,	0x5b701 },
	{ "WARAKIA11.PAL", 0x400,	0x5b702 },
	{ "WARAKIA12.PAL", 0x400,	0x5b703 },
	{ "WARAKIA13.PAL", 0x400,	0x5b704 },
	{ "WARAKIA14.PAL", 0x400,	0x5b705 },
	{ "WARAKIA15.PAL", 0x400,	0x5b706 },
	{ "WARAKIA16.PAL", 0x400,	0x5b707 },
	{ "WARAKIA17.PAL", 0x400,	0x5b708 },
	{ "WARAKIA18.PAL", 0x400,	0x5b709 },
	{ "WARAKIA19.PAL", 0x400,	0x5b70a },
	{ "WARAKIA2.ARX", 0x339cc0,	0x5b70b },
	{ "WARAKIA2.PAL", 0x400,	0x5bd7f },
	{ "WARAKIA20.PAL", 0x400,	0x5bd80 },
	{ "WARAKIA21.PAL", 0x400,	0x5bd81 },
	{ "WARAKIA22.PAL", 0x400,	0x5bd82 },
	{ "WARAKIA23.PAL", 0x400,	0x5bd83 },
	{ "WARAKIA24.PAL", 0x400,	0x5bd84 },
	{ "WARAKIA25.PAL", 0x400,	0x5bd85 },
	{ "WARAKIA26.PAL", 0x400,	0x5bd86 },
	{ "WARAKIA27.PAL", 0x400,	0x5bd87 },
	{ "WARAKIA28.PAL", 0x400,	0x5bd88 },
	{ "WARAKIA29.PAL", 0x400,	0x5bd89 },
	{ "WARAKIA3.PAL", 0x400,	0x5bd8a },
	{ "WARAKIA30.PAL", 0x400,	0x5bd8b },
	{ "WARAKIA31.PAL", 0x400,	0x5bd8c },
	{ "WARAKIA32.PAL", 0x400,	0x5bd8d },
	{ "WARAKIA33.PAL", 0x400,	0x5bd8e },
	{ "WARAKIA34.PAL", 0x400,	0x5bd8f },
	{ "WARAKIA35.PAL", 0x400,	0x5bd90 },
	{ "WARAKIA4.PAL", 0x400,	0x5bd91 },
	{ "WARAKIA5.PAL", 0x400,	0x5bd92 },
	{ "WARAKIA6.PAL", 0x400,	0x5bd93 },
	{ "WARAKIA7.PAL", 0x400,	0x5bd94 },
	{ "WARAKIA8.PAL", 0x400,	0x5bd95 },
	{ "WARAKIA9.PAL", 0x400,	0x5bd96 },
	{ "WARAKIAPAL2.ARX", 0x1c140,	0x5bd97 },
	{ "WARAKIA_0.HA6", 0x9358,	0x5bdd0 },
	{ "WARAKIA_0.TXT", 0xbc,	0x5bde3 },
	{ "WARAKIA_0_C.TXT", 0x1637,	0x5bde4 },
	{ "WARAKIA_1.HA6", 0x163c8,	0x5bde7 },
	{ "WARAKIA_1.TXT", 0xc0,	0x5be14 },
	{ "WARAKIA_1_C.TXT", 0x1257,	0x5be15 },
	{ "WARAKIA_2.HA6", 0xd920,	0x5be18 },
	{ "WARAKIA_2.TXT", 0xbc,	0x5be34 },
	{ "WARAKIA_2_C.TXT", 0x11d4,	0x5be35 },
	{ "WARAKIA_3.HA6", 0x1100c,	0x5be38 },
	{ "WARAKIA_3.TXT", 0xbc,	0x5be5b },
	{ "WARAKIA_3_C.TXT", 0xeed,	0x5be5c },
	{ "WARAKIA_9.HA6", 0x16588,	0x5be5e },
	{ "WARAKIA_9.TXT", 0xbc,	0x5be8b },
	{ "WARAKIA_9_C.TXT", 0x1014,	0x5be8c },
	{ "WARAKIA_C.TXT", 0x10d9,	0x5be8f },
	{ "WARAKIA_SELIST.TXT", 0xdaf,	0x5be92 },
	{ "WARC.CG2", 0x76b28,	0x5beb5 },
	{ "WARC.CPF", 0x2f36c,	0x5bfa3 },
	{ "WARC.HA6", 0x7a6e0,	0x5c002 },
	{ "WARC.PAT", 0xef12,	0x5c0f7 },
	{ "WARC.TXT", 0xd3,	0x5c115 },
	{ "WARC0.PAL", 0x400,	0x5c116 },
	{ "WARC1.PAL", 0x400,	0x5c117 },
	{ "WARC10.PAL", 0x400,	0x5c118 },
	{ "WARC11.PAL", 0x400,	0x5c119 },
	{ "WARC12.PAL", 0x400,	0x5c11a },
	{ "WARC13.PAL", 0x400,	0x5c11b },
	{ "WARC14.PAL", 0x400,	0x5c11c },
	{ "WARC15.PAL", 0x400,	0x5c11d },
	{ "WARC16.PAL", 0x400,	0x5c11e },
	{ "WARC17.PAL", 0x400,	0x5c11f },
	{ "WARC18.PAL", 0x400,	0x5c120 },
	{ "WARC19.PAL", 0x400,	0x5c121 },
	{ "WARC2.ARX", 0x361800,	0x5c122 },
	{ "WARC2.PAL", 0x400,	0x5c7e5 },
	{ "WARC20.PAL", 0x400,	0x5c7e6 },
	{ "WARC21.PAL", 0x400,	0x5c7e7 },
	{ "WARC22.PAL", 0x400,	0x5c7e8 },
	{ "WARC23.PAL", 0x400,	0x5c7e9 },
	{ "WARC24.PAL", 0x400,	0x5c7ea },
	{ "WARC25.PAL", 0x400,	0x5c7eb },
	{ "WARC26.PAL", 0x400,	0x5c7ec },
	{ "WARC27.PAL", 0x400,	0x5c7ed },
	{ "WARC28.PAL", 0x400,	0x5c7ee },
	{ "WARC29.PAL", 0x400,	0x5c7ef },
	{ "WARC3.PAL", 0x400,	0x5c7f0 },
	{ "WARC30.PAL", 0x400,	0x5c7f1 },
	{ "WARC31.PAL", 0x400,	0x5c7f2 },
	{ "WARC32.PAL", 0x400,	0x5c7f3 },
	{ "WARC33.PAL", 0x400,	0x5c7f4 },
	{ "WARC34.PAL", 0x400,	0x5c7f5 },
	{ "WARC35.PAL", 0x400,	0x5c7f6 },
	{ "WARC4.PAL", 0x400,	0x5c7f7 },
	{ "WARC5.PAL", 0x400,	0x5c7f8 },
	{ "WARC6.PAL", 0x400,	0x5c7f9 },
	{ "WARC7.PAL", 0x400,	0x5c7fa },
	{ "WARC8.PAL", 0x400,	0x5c7fb },
	{ "WARC9.PAL", 0x400,	0x5c7fc },
	{ "WARCPAL2.ARX", 0x20c00,	0x5c7fd },
	{ "WARC_0.HA6", 0xaee0,	0x5c83f },
	{ "WARC_0.TXT", 0xab,	0x5c855 },
	{ "WARC_0_C.TXT", 0x15d2,	0x5c856 },
	{ "WARC_1.HA6", 0x11410,	0x5c859 },
	{ "WARC_1.TXT", 0xab,	0x5c87c },
	{ "WARC_1_C.TXT", 0x12b8,	0x5c87d },
	{ "WARC_2.HA6", 0xb97c,	0x5c880 },
	{ "WARC_2.TXT", 0xad,	0x5c898 },
	{ "WARC_2_C.TXT", 0x1481,	0x5c899 },
	{ "WARC_3.HA6", 0x1100c,	0x5c89c },
	{ "WARC_3.TXT", 0xad,	0x5c8bf },
	{ "WARC_3_C.TXT", 0x1292,	0x5c8c0 },
	{ "WARC_9.HA6", 0x46fb8,	0x5c8c3 },
	{ "WARC_9.TXT", 0xad,	0x5c951 },
	{ "WARC_9_C.TXT", 0x1279,	0x5c952 },
	{ "WARC_C.TXT", 0x1706,	0x5c955 },
	{ "WARC_SELIST.TXT", 0xc14,	0x5c958 },
	{ "WLEN.CG2", 0x88000,	0x5c97b },
	{ "WLEN.CPF", 0x2f36c,	0x5ca8b },
	{ "WLEN.HA6", 0x739fc,	0x5caea },
	{ "WLEN.PAT", 0x0,	0x5cbd2 },
	{ "WLEN.TXT", 0xc2,	0x5cbd2 },
	{ "WLEN0.PAL", 0x400,	0x5cbd3 },
	{ "WLEN1.PAL", 0x400,	0x5cbd4 },
	{ "WLEN10.PAL", 0x400,	0x5cbd5 },
	{ "WLEN11.PAL", 0x400,	0x5cbd6 },
	{ "WLEN12.PAL", 0x400,	0x5cbd7 },
	{ "WLEN13.PAL", 0x400,	0x5cbd8 },
	{ "WLEN14.PAL", 0x400,	0x5cbd9 },
	{ "WLEN15.PAL", 0x400,	0x5cbda },
	{ "WLEN16.PAL", 0x400,	0x5cbdb },
	{ "WLEN17.PAL", 0x400,	0x5cbdc },
	{ "WLEN18.PAL", 0x400,	0x5cbdd },
	{ "WLEN19.PAL", 0x400,	0x5cbde },
	{ "WLEN2.ARX", 0x423500,	0x5cbdf },
	{ "WLEN2.PAL", 0x400,	0x5d426 },
	{ "WLEN20.PAL", 0x400,	0x5d427 },
	{ "WLEN21.PAL", 0x400,	0x5d428 },
	{ "WLEN22.PAL", 0x400,	0x5d429 },
	{ "WLEN23.PAL", 0x400,	0x5d42a },
	{ "WLEN24.PAL", 0x400,	0x5d42b },
	{ "WLEN25.PAL", 0x400,	0x5d42c },
	{ "WLEN26.PAL", 0x400,	0x5d42d },
	{ "WLEN27.PAL", 0x400,	0x5d42e },
	{ "WLEN28.PAL", 0x400,	0x5d42f },
	{ "WLEN29.PAL", 0x400,	0x5d430 },
	{ "WLEN3.PAL", 0x400,	0x5d431 },
	{ "WLEN30.PAL", 0x400,	0x5d432 },
	{ "WLEN31.PAL", 0x400,	0x5d433 },
	{ "WLEN32.PAL", 0x400,	0x5d434 },
	{ "WLEN33.PAL", 0x400,	0x5d435 },
	{ "WLEN34.PAL", 0x400,	0x5d436 },
	{ "WLEN35.PAL", 0x400,	0x5d437 },
	{ "WLEN4.PAL", 0x400,	0x5d438 },
	{ "WLEN5.PAL", 0x400,	0x5d439 },
	{ "WLEN6.PAL", 0x400,	0x5d43a },
	{ "WLEN7.PAL", 0x400,	0x5d43b },
	{ "WLEN8.PAL", 0x400,	0x5d43c },
	{ "WLEN9.PAL", 0x400,	0x5d43d },
	{ "WLENPAL2.ARX", 0x2bec0,	0x5d43e },
	{ "WLEN_0.HA6", 0xdbfc,	0x5d496 },
	{ "WLEN_0.TXT", 0x9c,	0x5d4b2 },
	{ "WLEN_0_C.TXT", 0x1788,	0x5d4b3 },
	{ "WLEN_1.HA6", 0x141dc,	0x5d4b6 },
	{ "WLEN_1.TXT", 0x9c,	0x5d4df },
	{ "WLEN_1_C.TXT", 0x133b,	0x5d4e0 },
	{ "WLEN_2.HA6", 0x15eec,	0x5d4e3 },
	{ "WLEN_2.TXT", 0x9c,	0x5d50f },
	{ "WLEN_2_C.TXT", 0x1510,	0x5d510 },
	{ "WLEN_3.HA6", 0x1100c,	0x5d513 },
	{ "WLEN_3.TXT", 0x9c,	0x5d536 },
	{ "WLEN_3_C.TXT", 0x1238,	0x5d537 },
	{ "WLEN_9.HA6", 0x15f68,	0x5d53a },
	{ "WLEN_9.TXT", 0x9c,	0x5d566 },
	{ "WLEN_9_C.TXT", 0x12f2,	0x5d567 },
	{ "WLEN_C.TXT", 0x16d8,	0x5d56a },
	{ "WLEN_SELIST.TXT", 0x1377,	0x5d56d },
};

static const int mbaa_iso_index_count = sizeof(mbaa_iso_index)/sizeof(mbaa_iso_index[0]);

static const int mbaa_iso_cvmfile_start = 0x13d1800;
static const int mbaa_iso_data_start = 0x1800;

#endif
